(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/admin-server.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var ADMIN_TOKEN_EXPIRATION_TIME = 15 * 60 * 1000;                      // 17
var publicAdminSettings = ["google", "github", "ldap", "saml", "emailToken", "splashUrl", "signupDialog", "adminAlert", "adminAlertTime", "adminAlertUrl", "termsUrl", "privacyUrl", "appMarketUrl", "appIndexUrl", "appUpdatesEnabled", "serverTitle", "returnAddress", "ldapNameField", "organizationMembership"];
                                                                       //
var FEATURE_KEY_FIELDS_PUBLISHED_TO_ADMINS = ["customer", "expires", "features", "isElasticBilling", "isTrial", "issued", "userLimit"];
                                                                       //
var PUBLIC_FEATURE_KEY_FIELDS = ["expires", "features"];               // 29
                                                                       //
var Fs = Npm.require("fs");                                            // 33
var SANDSTORM_ADMIN_TOKEN = SANDSTORM_VARDIR + "/adminToken";          // 34
                                                                       //
var tokenIsValid = function (token) {                                  // 36
  if (token && Fs.existsSync(SANDSTORM_ADMIN_TOKEN)) {                 // 37
    var stats = Fs.statSync(SANDSTORM_ADMIN_TOKEN);                    // 38
    var expireTime = new Date(Date.now() - ADMIN_TOKEN_EXPIRATION_TIME);
    if (stats.mtime < expireTime) {                                    // 40
      return false;                                                    // 41
    } else {                                                           //
      return Fs.readFileSync(SANDSTORM_ADMIN_TOKEN, { encoding: "utf8" }) === token;
    }                                                                  //
  } else {                                                             //
    return false;                                                      // 46
  }                                                                    //
};                                                                     //
                                                                       //
var checkAuth = function (token) {                                     // 50
  check(token, Match.OneOf(undefined, null, String));                  // 51
  if (!isAdmin() && !tokenIsValid(token)) {                            // 52
    throw new Meteor.Error(403, "User must be admin or provide a valid token");
  }                                                                    //
};                                                                     //
                                                                       //
var clearAdminToken = function (token) {                               // 57
  if (tokenIsValid(token)) {                                           // 58
    Fs.unlinkSync(SANDSTORM_ADMIN_TOKEN);                              // 59
    console.log("Admin token deleted.");                               // 60
  }                                                                    //
};                                                                     //
                                                                       //
Meteor.methods({                                                       // 64
  setAccountSetting: function (token, serviceName, value) {            // 65
    checkAuth(token);                                                  // 66
    check(serviceName, String);                                        // 67
    check(value, Boolean);                                             // 68
                                                                       //
    // TODO(someday): currently this relies on the fact that an account is tied to a single
    // identity, and thus has only that entry in "services". This will need to be looked at when
    // multiple login methods/identities are allowed for a single account.
    if (!value && !tokenIsValid(token) && serviceName in Meteor.user().services) {
      throw new Meteor.Error(403, "You can not disable the login service that your account uses.");
    }                                                                  //
                                                                       //
    // Only check configurations for OAuth services.                   //
    var oauthServices = ["google", "github"];                          // 79
    if (value && oauthServices.indexOf(serviceName) != -1) {           // 80
      var ServiceConfiguration = Package["service-configuration"].ServiceConfiguration;
      var config = ServiceConfiguration.configurations.findOne({ service: serviceName });
      if (!config) {                                                   // 83
        throw new Meteor.Error(403, "You must configure the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                //
                                                                       //
      if (!config.clientId || !config.secret) {                        // 88
        throw new Meteor.Error(403, "You must provide a non-empty clientId and secret for the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
      }                                                                //
    }                                                                  //
                                                                       //
    Settings.upsert({ _id: serviceName }, { $set: { value: value } });
    if (value) {                                                       // 95
      Settings.update({ _id: serviceName }, { $unset: { automaticallyReset: 1 } });
    }                                                                  //
  },                                                                   //
                                                                       //
  setSmtpConfig: function (token, config) {                            // 100
    checkAuth(token);                                                  // 101
    check(config, {                                                    // 102
      hostname: String,                                                // 103
      port: Number,                                                    // 104
      auth: {                                                          // 105
        user: String,                                                  // 106
        pass: String                                                   // 107
      },                                                               //
      returnAddress: String                                            // 109
    });                                                                //
                                                                       //
    Settings.upsert({ _id: "smtpConfig" }, { $set: { value: config } });
  },                                                                   //
                                                                       //
  setSetting: function (token, name, value) {                          // 115
    checkAuth(token);                                                  // 116
    check(name, String);                                               // 117
    check(value, Match.OneOf(null, String, Date, Boolean));            // 118
                                                                       //
    Settings.upsert({ _id: name }, { $set: { value: value } });        // 120
  },                                                                   //
                                                                       //
  submitFeatureKey: function (token, textBlock) {                      // 123
    checkAuth(token);                                                  // 124
    check(textBlock, Match.OneOf(null, String));                       // 125
                                                                       //
    var db = this.connection.sandstormDb;                              // 127
                                                                       //
    if (!textBlock) {                                                  // 129
      // Delete the feature key.                                       //
      db.collections.featureKey.remove("currentFeatureKey");           // 131
      return;                                                          // 132
    }                                                                  //
                                                                       //
    // textBlock is a base64'd string, possibly with newlines and comment lines starting with "-"
    var featureKeyBase64 = _.chain(textBlock.split("\n")).filter(function (line) {
      return line.length > 0 && line[0] !== "-";                       //
    }).value().join("");                                               //
                                                                       //
    var buf = new Buffer(featureKeyBase64, "base64");                  // 141
    if (buf.length < 64) {                                             // 142
      throw new Meteor.Error(401, "Invalid feature key");              // 143
    }                                                                  //
                                                                       //
    // loadSignedFeatureKey is provided in feature-key.js              //
    var featureKey = loadSignedFeatureKey(buf);                        // 147
    if (!featureKey) {                                                 // 148
      throw new Meteor.Error(401, "Invalid feature key");              // 149
    }                                                                  //
                                                                       //
    // Persist the feature key in the database.                        //
    db.collections.featureKey.upsert("currentFeatureKey", {            // 153
      _id: "currentFeatureKey",                                        // 156
      value: buf                                                       // 157
    });                                                                //
  },                                                                   //
                                                                       //
  saveOrganizationSettings: function (token, params) {                 // 162
    checkAuth(token);                                                  // 163
    check(params, {                                                    // 164
      membership: {                                                    // 165
        emailToken: {                                                  // 166
          enabled: Boolean,                                            // 167
          domain: String                                               // 168
        },                                                             //
        google: {                                                      // 170
          enabled: Boolean,                                            // 171
          domain: String                                               // 172
        },                                                             //
        ldap: {                                                        // 174
          enabled: Boolean                                             // 175
        },                                                             //
        saml: {                                                        // 177
          enabled: Boolean                                             // 178
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    // Disabled until we've actually implemented the feature.          //
    //settings: {                                                      //
    //  publishContacts: Boolean,                                      //
    //},                                                               //
    this.connection.sandstormDb.collections.settings.upsert({ _id: "organizationMembership" }, { value: params.membership });
  },                                                                   //
                                                                       //
  adminConfigureLoginService: function (token, options) {              // 190
    checkAuth(token);                                                  // 191
    check(options, Match.ObjectIncluding({ service: String }));        // 192
                                                                       //
    var ServiceConfiguration = Package["service-configuration"].ServiceConfiguration;
                                                                       //
    ServiceConfiguration.configurations.upsert({ service: options.service }, options);
  },                                                                   //
                                                                       //
  clearResumeTokensForService: function (token, serviceName) {         // 199
    checkAuth(token);                                                  // 200
    check(serviceName, String);                                        // 201
                                                                       //
    var query = {};                                                    // 203
    query["services." + serviceName] = { $exists: true };              // 204
    Meteor.users.find(query).forEach(function (identity) {             // 205
      if (identity.services.resume && identity.services.resume.loginTokens && identity.services.resume.loginTokens.length > 0) {
        Meteor.users.update({ _id: identity._id }, { $set: { "services.resume.loginTokens": [] } });
      }                                                                //
                                                                       //
      Meteor.users.update({ "loginIdentities.id": identity._id }, { $set: { "services.resume.loginTokens": [] } });
    });                                                                //
  },                                                                   //
                                                                       //
  adminUpdateUser: function (token, userInfo) {                        // 216
    checkAuth(token);                                                  // 217
    check(userInfo, {                                                  // 218
      userId: String,                                                  // 219
      signupKey: Boolean,                                              // 220
      isAdmin: Boolean                                                 // 221
    });                                                                //
                                                                       //
    var userId = userInfo.userId;                                      // 224
    if (userId === Meteor.userId() && !userInfo.isAdmin) {             // 225
      throw new Meteor.Error(403, "User cannot remove admin permissions from itself.");
    }                                                                  //
                                                                       //
    Meteor.users.update({ _id: userId }, { $set: _.omit(userInfo, ["_id", "userId"]) });
  },                                                                   //
                                                                       //
  testSend: function (token, smtpConfig, to) {                         // 232
    checkAuth(token);                                                  // 233
    check(smtpConfig, {                                                // 234
      hostname: String,                                                // 235
      port: Number,                                                    // 236
      auth: {                                                          // 237
        user: String,                                                  // 238
        pass: String                                                   // 239
      },                                                               //
      returnAddress: String                                            // 241
    });                                                                //
    check(to, String);                                                 // 243
    var returnAddress = smtpConfig.returnAddress;                      //
    var restConfig = babelHelpers.objectWithoutProperties(smtpConfig, ["returnAddress"]);
                                                                       //
    SandstormEmail.send({                                              // 246
      to: to,                                                          // 247
      from: globalDb.getServerTitle() + " <" + returnAddress + ">",    // 248
      subject: "Testing your Sandstorm's SMTP setting",                // 249
      text: "Success! Your outgoing SMTP is working.",                 // 250
      smtpConfig: restConfig                                           // 251
    });                                                                //
  },                                                                   //
                                                                       //
  createSignupKey: function (token, note, quota) {                     // 255
    checkAuth(token);                                                  // 256
    check(note, String);                                               // 257
    check(quota, Match.OneOf(undefined, null, Number));                // 258
                                                                       //
    var key = Random.id();                                             // 260
    var content = { _id: key, used: false, note: note };               // 261
    if (typeof quota === "number") content.quota = quota;              // 262
    SignupKeys.insert(content);                                        // 263
    return key;                                                        // 264
  },                                                                   //
                                                                       //
  sendInvites: function (token, origin, from, list, subject, message, quota) {
    checkAuth(token);                                                  // 268
    check([origin, from, list, subject, message], [String]);           // 269
    check(quota, Match.OneOf(undefined, null, Number));                // 270
                                                                       //
    if (!from.trim()) {                                                // 272
      throw new Meteor.Error(403, "Must enter 'from' address.");       // 273
    }                                                                  //
                                                                       //
    if (!list.trim()) {                                                // 276
      throw new Meteor.Error(403, "Must enter 'to' addresses.");       // 277
    }                                                                  //
                                                                       //
    this.unblock();                                                    // 280
                                                                       //
    list = list.split("\n");                                           // 282
    for (var i in babelHelpers.sanitizeForInObject(list)) {            // 283
      var email = list[i].trim();                                      // 284
                                                                       //
      if (email) {                                                     // 286
        var key = Random.id();                                         // 287
                                                                       //
        var content = {                                                // 289
          _id: key,                                                    // 290
          used: false,                                                 // 291
          note: "E-mail invite to " + email,                           // 292
          email: email,                                                // 293
          definitelySent: false                                        // 294
        };                                                             //
        if (typeof quota === "number") content.quota = quota;          // 296
        SignupKeys.insert(content);                                    // 297
        SandstormEmail.send({                                          // 298
          to: email,                                                   // 299
          from: from,                                                  // 300
          envelopeFrom: globalDb.getReturnAddress(),                   // 301
          subject: subject,                                            // 302
          text: message.replace(/\$KEY/g, origin + "/signup/" + key)   // 303
        });                                                            //
        SignupKeys.update(key, { $set: { definitelySent: true } });    // 305
      }                                                                //
    }                                                                  //
                                                                       //
    return { sent: true };                                             // 309
  },                                                                   //
                                                                       //
  offerIpNetwork: function (token) {                                   // 312
    checkAuth(token);                                                  // 313
    if (!isAdmin()) {                                                  // 314
      throw new Meteor.Error(403, "Offering IpNetwork is only allowed for logged in users " + "(a token is not sufficient). Please sign in with an admin account");
    }                                                                  //
                                                                       //
    var requirements = [{ userIsAdmin: Meteor.userId() }];             // 319
    var sturdyRef = waitPromise(saveFrontendRef({ ipNetwork: true }, { webkey: null }, requirements)).sturdyRef;
    return ROOT_URL.protocol + "//" + globalDb.makeApiHost(sturdyRef) + "#" + sturdyRef;
  },                                                                   //
                                                                       //
  offerIpInterface: function (token) {                                 // 328
    checkAuth(token);                                                  // 329
    if (!isAdmin()) {                                                  // 330
      throw new Meteor.Error(403, "Offering IpInterface is only allowed for logged in users " + "(a token is not sufficient). Please sign in with an admin account");
    }                                                                  //
                                                                       //
    var requirements = [{ userIsAdmin: Meteor.userId() }];             // 335
    var sturdyRef = waitPromise(saveFrontendRef({ ipInterface: true }, { webkey: null }, requirements)).sturdyRef;
    return ROOT_URL.protocol + "//" + globalDb.makeApiHost(sturdyRef) + "#" + sturdyRef;
  },                                                                   //
                                                                       //
  adminToggleDisableCap: function (token, capId, value) {              // 344
    checkAuth(token);                                                  // 345
    check(capId, String);                                              // 346
    check(value, Boolean);                                             // 347
                                                                       //
    if (value) {                                                       // 349
      ApiTokens.update({ _id: capId }, { $set: { revoked: true } });   // 350
    } else {                                                           //
      ApiTokens.update({ _id: capId }, { $set: { revoked: false } });  // 352
    }                                                                  //
  },                                                                   //
                                                                       //
  updateQuotas: function (token, list, quota) {                        // 356
    checkAuth(token);                                                  // 357
    check(list, String);                                               // 358
    check(quota, Match.OneOf(undefined, null, Number));                // 359
                                                                       //
    if (!list.trim()) {                                                // 361
      throw new Meteor.Error(400, "Must enter addresses.");            // 362
    }                                                                  //
                                                                       //
    var items = list.split("\n");                                      // 365
    var invalid = [];                                                  // 366
    for (var i in babelHelpers.sanitizeForInObject(items)) {           // 367
      var modifier = typeof quota === "number" ? { $set: { quota: quota } } : { $unset: { quota: "" } };
      var n = SignupKeys.update({ email: items[i] }, modifier, { multi: true });
      n += Meteor.users.update({ signupEmail: items[i] }, modifier, { multi: true });
                                                                       //
      if (n < 1) invalid.push(items[i]);                               // 373
    }                                                                  //
                                                                       //
    if (invalid.length > 0) {                                          // 376
      throw new Meteor.Error(404, "These addresses did not map to any user nor invite: " + invalid.join(", "));
    }                                                                  //
  },                                                                   //
                                                                       //
  dismissAdminStatsNotifications: function (token) {                   // 382
    checkAuth(token);                                                  // 383
    globalDb.collections.notifications.remove({ "admin.type": "reportStats" });
  },                                                                   //
                                                                       //
  signUpAsAdmin: function (token) {                                    // 387
    check(token, String);                                              // 388
    checkAuth(token);                                                  // 389
    if (!this.userId) {                                                // 390
      throw new Meteor.Error(403, "Must be logged in to sign up as admin.");
    }                                                                  //
                                                                       //
    if (!Meteor.user().loginIdentities) {                              // 394
      throw new Meteor.Error(403, "Must be logged into an account to sign up as admin.");
    }                                                                  //
                                                                       //
    Meteor.users.update({ _id: this.userId }, { $set: { isAdmin: true, signupKey: "admin" } });
    clearAdminToken(token);                                            // 399
  }                                                                    //
});                                                                    //
                                                                       //
var authorizedAsAdmin = function (token, userId) {                     // 403
  return Match.test(token, Match.OneOf(undefined, null, String)) && (userId && isAdminById(userId) || tokenIsValid(token));
};                                                                     //
                                                                       //
Meteor.publish("admin", function (token) {                             // 408
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 409
  return Settings.find();                                              // 410
});                                                                    //
                                                                       //
Meteor.publish("adminServiceConfiguration", function (token) {         // 413
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 414
  return Package["service-configuration"].ServiceConfiguration.configurations.find();
});                                                                    //
                                                                       //
Meteor.publish("publicAdminSettings", function () {                    // 418
  return Settings.find({ _id: { $in: publicAdminSettings } });         // 419
});                                                                    //
                                                                       //
Meteor.publish("adminToken", function (token) {                        // 422
  check(token, String);                                                // 423
  this.added("adminToken", "adminToken", { tokenIsValid: tokenIsValid(token) });
  this.ready();                                                        // 425
});                                                                    //
                                                                       //
Meteor.publish("allUsers", function (token) {                          // 428
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 429
  return Meteor.users.find();                                          // 430
});                                                                    //
                                                                       //
Meteor.publish("activityStats", function (token) {                     // 433
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 434
  return ActivityStats.find();                                         // 435
});                                                                    //
                                                                       //
Meteor.publish("statsTokens", function (token) {                       // 438
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 439
  return StatsTokens.find();                                           // 440
});                                                                    //
                                                                       //
Meteor.publish("allPackages", function (token) {                       // 443
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 444
  return Packages.find({ manifest: { $exists: true } }, { fields: { appId: 1, "manifest.appVersion": 1,
      "manifest.actions": 1, "manifest.appTitle": 1 } });              // 447
});                                                                    //
                                                                       //
Meteor.publish("realTimeStats", function (token) {                     // 450
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 451
                                                                       //
  // Last five minutes.                                                //
  this.added("realTimeStats", "now", computeStats(new Date(Date.now() - 5 * 60 * 1000)));
                                                                       //
  // Since last sample.                                                //
  var lastSample = ActivityStats.findOne({}, { sort: { timestamp: -1 } });
  var lastSampleTime = lastSample ? lastSample.timestamp : new Date(0);
  this.added("realTimeStats", "today", computeStats(lastSampleTime));  // 459
                                                                       //
  // TODO(someday): Update every few minutes?                          //
                                                                       //
  this.ready();                                                        // 463
});                                                                    //
                                                                       //
Meteor.publish("adminLog", function (token) {                          // 466
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 467
                                                                       //
  var logfile = SANDSTORM_LOGDIR + "/sandstorm.log";                   // 469
                                                                       //
  var fd = Fs.openSync(logfile, "r");                                  // 471
  var startSize = Fs.fstatSync(fd).size;                               // 472
                                                                       //
  // Start tailing at EOF - 8k.                                        //
  var offset = Math.max(0, startSize - 8192);                          // 475
                                                                       //
  var _this = this;                                                    // 477
  function doTail() {                                                  // 478
    for (;;) {                                                         // 479
      var buf = new Buffer(Math.max(1024, startSize - offset));        // 480
      var n = Fs.readSync(fd, buf, 0, buf.length, offset);             // 481
      if (n <= 0) break;                                               // 482
      _this.added("adminLog", offset, { text: buf.toString("utf8", 0, n) });
      offset += n;                                                     // 484
    }                                                                  //
  }                                                                    //
                                                                       //
  // Watch the file for changes.                                       //
  var watcher = Fs.watch(logfile, { persistent: false }, Meteor.bindEnvironment(doTail));
                                                                       //
  // When the subscription stops, stop watching the file.              //
  this.onStop(function () {                                            // 492
    watcher.close();                                                   // 493
    Fs.closeSync(fd);                                                  // 494
  });                                                                  //
                                                                       //
  // Read initial 8k tail data immediately.                            //
  doTail();                                                            // 498
                                                                       //
  // Notify ready.                                                     //
  this.ready();                                                        // 501
});                                                                    //
                                                                       //
Meteor.publish("adminApiTokens", function (token) {                    // 504
  if (!authorizedAsAdmin(token, this.userId)) return [];               // 505
  return ApiTokens.find({                                              // 506
    $or: [{ "frontendRef.ipNetwork": { $exists: true } }, { "frontendRef.ipInterface": { $exists: true } }]
  }, {                                                                 //
    fields: {                                                          // 512
      frontendRef: 1,                                                  // 513
      created: 1,                                                      // 514
      requirements: 1,                                                 // 515
      revoked: 1                                                       // 516
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
Meteor.publish("featureKey", function (forAdmin, token) {              // 521
  var _this2 = this;                                                   //
                                                                       //
  if (forAdmin && !authorizedAsAdmin(token, this.userId)) return [];   // 522
                                                                       //
  var fields = forAdmin ? FEATURE_KEY_FIELDS_PUBLISHED_TO_ADMINS : PUBLIC_FEATURE_KEY_FIELDS;
                                                                       //
  var db = this.connection.sandstormDb;                                // 527
  var featureKeyQuery = db.collections.featureKey.find({ _id: "currentFeatureKey" });
  var observeHandle = featureKeyQuery.observe({                        // 529
    added: function (doc) {                                            // 530
      // Load and verify the signed feature key.                       //
      var buf = new Buffer(doc.value);                                 // 532
      var featureKey = loadSignedFeatureKey(buf);                      // 533
                                                                       //
      if (featureKey) {                                                // 535
        var _ref;                                                      //
                                                                       //
        // If the signature is valid, publish the feature key information.
        var filteredFeatureKey = (_ref = _).pick.apply(_ref, [featureKey].concat(fields));
        _this2.added("featureKey", doc._id, filteredFeatureKey);       // 538
      }                                                                //
    },                                                                 //
                                                                       //
    changed: function (newDoc, oldDoc) {                               // 542
      // Load and reverify the new signed feature key.                 //
      var buf = new Buffer(newDoc.value);                              // 544
      var featureKey = loadSignedFeatureKey(buf);                      // 545
                                                                       //
      if (featureKey) {                                                // 547
        var _ref2;                                                     //
                                                                       //
        // If the signature is valid, call this.changed() with the interesting fields.
        var filteredFeatureKey = (_ref2 = _).pick.apply(_ref2, [featureKey].concat(fields));
        _this2.changed("featureKey", newDoc._id, filteredFeatureKey);  // 550
      } else {                                                         //
        // Otherwise, call this.removed(), since the new feature key is invalid.
        _this2.removed("featureKey", oldDoc._id);                      // 553
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (oldDoc) {                                       // 557
      _this2.removed("featureKey", oldDoc._id);                        // 558
    }                                                                  //
  });                                                                  //
                                                                       //
  this.onStop(function () {                                            // 562
    observeHandle.stop();                                              // 563
  });                                                                  //
                                                                       //
  this.ready();                                                        // 566
});                                                                    //
                                                                       //
function observeOauthService(name) {                                   // 569
  Settings.find({ _id: name, value: true }).observe({                  // 570
    added: function () {                                               // 571
      // Tell the oauth library it should accept login attempts from this service.
      Accounts.oauth.registerService(name);                            // 573
    },                                                                 //
                                                                       //
    removed: function () {                                             // 576
      // Tell the oauth library it should deny login attempts from this service.
      Accounts.oauth.unregisterService(name);                          // 578
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
observeOauthService("github");                                         // 583
observeOauthService("google");                                         // 584
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=admin-server.js.map
