(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/core.js                                                      //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require("crypto");                                    // 17
var Capnp = Npm.require("capnp");                                      // 18
                                                                       //
var PersistentHandle = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentHandle;
var SandstormCore = Capnp.importSystem("sandstorm/supervisor.capnp").SandstormCore;
var SandstormCoreFactory = Capnp.importSystem("sandstorm/backend.capnp").SandstormCoreFactory;
var PersistentOngoingNotification = Capnp.importSystem("sandstorm/supervisor.capnp").PersistentOngoingNotification;
var PersistentUiView = Capnp.importSystem("sandstorm/persistentuiview.capnp").PersistentUiView;
                                                                       //
var SandstormCoreImpl = (function () {                                 //
  function SandstormCoreImpl(grainId) {                                // 27
    babelHelpers.classCallCheck(this, SandstormCoreImpl);              //
                                                                       //
    this.grainId = grainId;                                            // 28
  }                                                                    //
                                                                       //
  SandstormCoreImpl.prototype.restore = (function () {                 // 26
    function restore(sturdyRef, requiredPermissions) {                 // 31
      var _this2 = this;                                               //
                                                                       //
      var _this = this;                                                // 32
      return inMeteor(function () {                                    // 33
        var hashedSturdyRef = hashSturdyRef(sturdyRef);                // 34
        var token = ApiTokens.findOne({                                // 35
          _id: hashedSturdyRef,                                        // 36
          "owner.grain.grainId": _this2.grainId                        // 37
        });                                                            //
                                                                       //
        if (!token) {                                                  // 40
          throw new Error("no such token");                            // 41
        }                                                              //
                                                                       //
        // Honor `requiredPermissions`.                                //
        var requirements = [];                                         // 45
        if (requiredPermissions && token.owner.grain.introducerIdentity) {
          requirements.push({                                          // 47
            permissionsHeld: {                                         // 48
              permissions: requiredPermissions,                        // 49
              identityId: token.owner.grain.introducerIdentity,        // 50
              grainId: _this.grainId                                   // 51
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        return restoreInternal(sturdyRef, { grain: Match.ObjectIncluding({ grainId: _this.grainId }) }, requirements, hashedSturdyRef);
      });                                                              //
    }                                                                  //
                                                                       //
    return restore;                                                    //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.drop = (function () {                    // 26
    function drop(sturdyRef) {                                         // 62
      var _this = this;                                                // 63
      return inMeteor(function () {                                    // 64
        return dropInternal(sturdyRef, { grain: Match.ObjectIncluding({ grainId: _this.grainId }) });
      });                                                              //
    }                                                                  //
                                                                       //
    return drop;                                                       //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.makeToken = (function () {               // 26
    function makeToken(ref, owner, requirements) {                     // 69
      var _this = this;                                                // 70
      return inMeteor(function () {                                    // 71
        var sturdyRef = new Buffer(generateSturdyRef());               // 72
        var hashedSturdyRef = hashSturdyRef(sturdyRef);                // 73
        ApiTokens.insert({                                             // 74
          _id: hashedSturdyRef,                                        // 75
          grainId: _this.grainId,                                      // 76
          objectId: ref,                                               // 77
          owner: owner,                                                // 78
          created: new Date(),                                         // 79
          requirements: requirements                                   // 80
        });                                                            //
                                                                       //
        return {                                                       // 83
          token: sturdyRef                                             // 84
        };                                                             //
      });                                                              //
    }                                                                  //
                                                                       //
    return makeToken;                                                  //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.makeChildToken = (function () {          // 26
    function makeChildToken(parent, owner, requirements) {             // 89
      var _this = this;                                                // 90
      return inMeteor(function () {                                    // 91
        return {                                                       // 92
          token: new Buffer(makeChildTokenInternal(parent, owner, requirements))
        };                                                             //
      });                                                              //
    }                                                                  //
                                                                       //
    return makeChildToken;                                             //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.getOwnerNotificationTarget = (function () {
    function getOwnerNotificationTarget() {                            // 98
      var grainId = this.grainId;                                      // 99
      return {                                                         // 100
        owner: {                                                       // 101
          addOngoing: function (displayInfo, notification) {           // 102
            return inMeteor(function () {                              // 103
              var grain = Grains.findOne({ _id: grainId });            // 104
              if (!grain) {                                            // 105
                throw new Error("Grain not found.");                   // 106
              }                                                        //
                                                                       //
              var castedNotification = notification.castAs(PersistentOngoingNotification);
              var wakelockToken = waitPromise(castedNotification.save()).sturdyRef;
                                                                       //
              // We have to close both the casted cap and the original. Perhaps this should be fixed in
              // node-capnp?                                           //
              castedNotification.close();                              // 114
              notification.close();                                    // 115
              var notificationId = Notifications.insert({              // 116
                ongoing: wakelockToken,                                // 117
                grainId: grainId,                                      // 118
                userId: grain.userId,                                  // 119
                text: displayInfo.caption,                             // 120
                timestamp: new Date(),                                 // 121
                isUnread: true                                         // 122
              });                                                      //
                                                                       //
              var persistentMethods = {                                // 125
                save: function (params) {                              // 126
                  return saveFrontendRef({ notificationHandle: notificationId }, params.sealFor);
                }                                                      //
              };                                                       //
                                                                       //
              return { handle: makeNotificationHandle(notificationId, false, persistentMethods) };
            });                                                        //
          }                                                            //
        }                                                              //
      };                                                               //
    }                                                                  //
                                                                       //
    return getOwnerNotificationTarget;                                 //
  })();                                                                //
                                                                       //
  return SandstormCoreImpl;                                            //
})();                                                                  //
                                                                       //
var makeSandstormCore = function (grainId) {                           // 139
  return new Capnp.Capability(new SandstormCoreImpl(grainId), SandstormCore);
};                                                                     //
                                                                       //
var NotificationHandle = (function () {                                //
  function NotificationHandle(notificationId, saved, persistentMethods) {
    babelHelpers.classCallCheck(this, NotificationHandle);             //
                                                                       //
    this.notificationId = notificationId;                              // 145
    this.saved = saved;                                                // 146
    _.extend(this, persistentMethods);                                 // 147
  }                                                                    //
                                                                       //
  NotificationHandle.prototype.close = (function () {                  // 143
    function close() {                                                 // 150
      var _this = this;                                                // 151
      return inMeteor(function () {                                    // 152
        if (!_this.saved) {                                            // 153
          dismissNotification(_this.notificationId);                   // 154
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return NotificationHandle;                                           //
})();                                                                  //
                                                                       //
var PersistentUiViewImpl = (function () {                              //
  function PersistentUiViewImpl(persistentMethods) {                   // 161
    babelHelpers.classCallCheck(this, PersistentUiViewImpl);           //
                                                                       //
    _.extend(this, persistentMethods);                                 // 162
  }                                                                    //
                                                                       //
  // All other UiView methods are currently unimplemented, which, while not strictly correct,
  // results in the same overall behavior, since users can't call restore() on a PersistentUiView,
  // and grains can't call methods on UiViews because they lack the "is human" pseudopermission.
  return PersistentUiViewImpl;                                         //
})();                                                                  //
                                                                       //
var makePersistentUiView = function (persistentMethods) {              // 170
  return new Capnp.Capability(new PersistentUiViewImpl(persistentMethods), PersistentUiView);
};                                                                     //
                                                                       //
function makeNotificationHandle(notificationId, saved, persistentMethods) {
  return new Capnp.Capability(new NotificationHandle(notificationId, saved, persistentMethods), PersistentHandle);
}                                                                      //
                                                                       //
function dropWakelock(grainId, wakeLockNotificationId) {               // 179
  waitPromise(globalBackend.useGrain(grainId, function (supervisor) {  // 180
    return supervisor.drop({ wakeLockNotification: wakeLockNotificationId });
  }));                                                                 //
}                                                                      //
                                                                       //
function dismissNotification(notificationId, callCancel) {             // 185
  var notification = Notifications.findOne({ _id: notificationId });   // 186
  if (notification) {                                                  // 187
    Notifications.remove({ _id: notificationId });                     // 188
    if (notification.ongoing) {                                        // 189
      // For some reason, Mongo returns an object that looks buffer-like, but isn't a buffer.
      // Only way to fix seems to be to copy it.                       //
      var id = new Buffer(notification.ongoing);                       // 192
                                                                       //
      if (!callCancel) {                                               // 194
        dropInternal(id, { frontend: null });                          // 195
      } else {                                                         //
        var notificationCap = restoreInternal(id, { frontend: null }, []).cap;
        var castedNotification = notificationCap.castAs(PersistentOngoingNotification);
        dropInternal(id, { frontend: null });                          // 199
        try {                                                          // 200
          waitPromise(castedNotification.cancel());                    // 201
          castedNotification.close();                                  // 202
          notificationCap.close();                                     // 203
        } catch (err) {                                                //
          if (err.kjType !== "disconnected") {                         // 205
            // ignore disconnected errors, since cancel may shutdown the grain before the supervisor
            // responds.                                               //
            throw err;                                                 // 208
          }                                                            //
        }                                                              //
      }                                                                //
    } else if (notification.appUpdates) {                              //
      _.forEach(notification.appUpdates, function (app, appId) {       // 213
        deletePackage(app.packageId);                                  // 214
      });                                                              //
    }                                                                  //
  }                                                                    //
}                                                                      //
                                                                       //
hashSturdyRef = function (sturdyRef) {                                 // 220
  return Crypto.createHash("sha256").update(sturdyRef).digest("base64");
};                                                                     //
                                                                       //
function generateSturdyRef() {                                         // 224
  return Random.secret();                                              // 225
}                                                                      //
                                                                       //
Meteor.methods({                                                       // 228
  dismissNotification: function (notificationId) {                     // 229
    // This will remove notifications from the database and from view of the user.
    // For ongoing notifications, it will begin the process of cancelling and dropping them from
    // the app.                                                        //
                                                                       //
    check(notificationId, String);                                     // 234
                                                                       //
    var notification = Notifications.findOne({ _id: notificationId });
    if (!notification) {                                               // 237
      throw new Meteor.Error(404, "Notification id not found.");       // 238
    } else if (notification.userId !== Meteor.userId()) {              //
      throw new Meteor.Error(403, "Notification does not belong to current user.");
    } else {                                                           //
      dismissNotification(notificationId, true);                       // 242
    }                                                                  //
  },                                                                   //
                                                                       //
  readAllNotifications: function () {                                  // 246
    // Marks all notifications as read for the current user.           //
    if (!Meteor.userId()) {                                            // 248
      throw new Meteor.Error(403, "User not logged in.");              // 249
    }                                                                  //
                                                                       //
    Notifications.update({ userId: Meteor.userId() }, { $set: { isUnread: false } }, { multi: true });
  }                                                                    //
});                                                                    //
                                                                       //
saveFrontendRef = function (frontendRef, owner, requirements) {        // 256
  return inMeteor(function () {                                        // 257
    var sturdyRef = new Buffer(generateSturdyRef());                   // 258
    var hashedSturdyRef = hashSturdyRef(sturdyRef);                    // 259
    ApiTokens.insert({                                                 // 260
      _id: hashedSturdyRef,                                            // 261
      frontendRef: frontendRef,                                        // 262
      owner: owner,                                                    // 263
      created: new Date(),                                             // 264
      requirements: requirements                                       // 265
    });                                                                //
    return { sturdyRef: sturdyRef };                                   // 267
  });                                                                  //
};                                                                     //
                                                                       //
checkRequirements = function (requirements) {                          // 271
  if (!requirements) {                                                 // 272
    return true;                                                       // 273
  }                                                                    //
                                                                       //
  for (var i in babelHelpers.sanitizeForInObject(requirements)) {      // 276
    var requirement = requirements[i];                                 // 277
    if (requirement.tokenValid) {                                      // 278
      var token = ApiTokens.findOne({ _id: requirement.tokenValid }, { fields: { requirements: 1 } });
      if (!checkRequirements(token.requirements)) {                    // 280
        return false;                                                  // 281
      }                                                                //
    } else if (requirement.permissionsHeld) {                          //
      var p = requirement.permissionsHeld;                             // 284
      var viewInfo = Grains.findOne(p.grainId, { fields: { cachedViewInfo: 1 } }).cachedViewInfo;
      var currentPermissions = SandstormPermissions.grainPermissions(globalDb, { grain: { _id: p.grainId, identityId: p.identityId } }, viewInfo || {});
      if (!currentPermissions) {                                       // 288
        return false;                                                  // 289
      }                                                                //
                                                                       //
      for (var ii = 0; ii < p.permissions.length; ++ii) {              // 292
        if (p.permissions[ii] && !currentPermissions[ii]) {            // 293
          return false;                                                // 294
        }                                                              //
      }                                                                //
    } else if (requirement.userIsAdmin) {                              //
      if (!isAdminById(requirement.userIsAdmin)) {                     // 298
        return false;                                                  // 299
      }                                                                //
    } else {                                                           //
      throw new Meteor.Error(403, "Unknown requirement");              // 302
    }                                                                  //
  }                                                                    //
                                                                       //
  return true;                                                         // 306
};                                                                     //
                                                                       //
restoreInternal = function (originalToken, ownerPattern, requirements, tokenId) {
  // Restores the token `originalToken`, which is a Buffer.            //
  //                                                                   //
  // `ownerPattern` is a match pattern (i.e. used with check()) that the token's owner must match.
  // This is used to enforce than an entity can't use tokens owned by some other entity.
  //                                                                   //
  // `requirements` is a list of additional MembraneRequirements to add to the returned capability,
  // beyond what's already stored in ApiTokens. This is often an empty list.
  //                                                                   //
  // `tokenId` is optional. If specified, it should be hashSturdyRef(originalToken); only specify
  // it if you happen to have computed this already.                   //
  //                                                                   //
  // (When the token turns out to have a parent, this function will call itself recursively. When
  // it does, `originalToken` stays the same, but `tokenId` is replaced with the parent. This is
  // because the capability we ultimately restore needs to become a child of the token that is
  // being restored.)                                                  //
                                                                       //
  tokenId = tokenId || hashSturdyRef(originalToken);                   // 326
  requirements = requirements || [];                                   // 327
                                                                       //
  var token = ApiTokens.findOne(tokenId);                              // 329
  if (!token) {                                                        // 330
    throw new Meteor.Error(403, "No token found to restore");          // 331
  }                                                                    //
                                                                       //
  if (token.revoked) {                                                 // 334
    throw new Meteor.Error(403, "Token has been revoked");             // 335
  }                                                                    //
                                                                       //
  // The ownerPattern should specify the appropriate user or grain involved, if appropriate.
  check(token.owner, ownerPattern);                                    // 339
                                                                       //
  // Check requirements on the token.                                  //
  if (!checkRequirements(token.requirements)) {                        // 342
    throw new Meteor.Error(403, "Requirements not satisfied.");        // 343
  }                                                                    //
                                                                       //
  // Check expiration.                                                 //
  if (token.expires && token.expires.getTime() <= Date.now()) {        // 347
    throw new Meteor.Error(403, "Authorization token expired");        // 348
  }                                                                    //
                                                                       //
  if (token.expiresIfUnused) {                                         // 351
    if (token.expiresIfUnused.getTime() <= Date.now()) {               // 352
      throw new Meteor.Error(403, "Authorization token expired");      // 353
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(token._id, { $unset: { expiresIfUnused: "" } });
    }                                                                  //
  }                                                                    //
                                                                       //
  // If this token has a parent, go ahead and recurse to it now.       //
  if (token.parentToken) {                                             // 361
    // A token which chains to some parent token.  Restore the parent token (possibly recursively),
    // checking requirements on the way up.                            //
    return restoreInternal(originalToken, Match.Any, requirements, token.parentToken);
  }                                                                    //
                                                                       //
  // Check the passed-in `requirements`.                               //
  if (!checkRequirements(requirements)) {                              // 368
    throw new Meteor.Error(403, "Requirements not satisfied.");        // 369
  }                                                                    //
                                                                       //
  // The capability we restore must implement SystemPersistent, and we already know what the
  // implementation will look like. So, construct it here.             //
  //                                                                   //
  // TODO(cleanup): It would probably be a lot cleaner to have a common base class that these
  //   inherit, and to pass down some common parameters to the constructor.
  var persistentMethods = {                                            // 377
    save: function (params) {                                          // 378
      return inMeteor(function () {                                    // 379
        var sturdyRef = new Buffer(makeChildTokenInternal(originalToken, params.sealFor, requirements, token));
        return { sturdyRef: sturdyRef };                               // 382
      });                                                              //
    }                                                                  //
                                                                       //
  };                                                                   //
                                                                       //
  // TODO(someday): Implement SystemPersistent.addRequirements().      //
  if (token.frontendRef) {                                             // 389
    // A token which represents a capability implemented by a pseudo-driver.
                                                                       //
    if (token.frontendRef.notificationHandle) {                        // 392
      var notificationId = token.frontendRef.notificationHandle;       // 393
      return { cap: makeNotificationHandle(notificationId, true, persistentMethods) };
    } else if (token.frontendRef.ipNetwork) {                          //
      return { cap: makeIpNetwork(persistentMethods) };                // 396
    } else if (token.frontendRef.ipInterface) {                        //
      return { cap: makeIpInterface(persistentMethods) };              // 398
    } else if (token.frontendRef.emailVerifier) {                      //
      return { cap: makeEmailVerifier(persistentMethods, tokenId, token.frontendRef.emailVerifier) };
    } else if (token.frontendRef.verifiedEmail) {                      //
      return { cap: makeVerifiedEmail(persistentMethods) };            // 403
    } else {                                                           //
      throw new Meteor.Error(500, "Unknown frontend token type.");     // 405
    }                                                                  //
  } else if (token.objectId) {                                         //
    // A token which represents a specific capability exported by a grain.
                                                                       //
    // Fix Mongo converting Buffers to Uint8Arrays.                    //
    if (token.objectId.appRef) {                                       // 411
      token.objectId.appRef = new Buffer(token.objectId.appRef);       // 412
    }                                                                  //
                                                                       //
    // Ensure the grain is running, then restore the capability.       //
    return waitPromise(globalBackend.useGrain(token.grainId, function (supervisor) {
      // Note that in this case it is the supervisor's job to implement SystemPersistent, so we
      // discard persistentMethods.                                    //
      return supervisor.restore(token.objectId, requirements, originalToken);
    }));                                                               //
  } else if (token.grainId) {                                          //
    // It's a UiView.                                                  //
                                                                       //
    // If a grain is attempting to restore a UiView, it gets a UiView which filters out all
    // the method calls.  In the future, we may allow grains to restore UiViews that pass along the
    // "is human" pseudopermission (say, to allow an app to proxy all requests to some grain and
    // do some transformation), which will return a different capability.
    return { cap: makePersistentUiView(persistentMethods) };           // 428
  } else {                                                             //
    throw new Meteor.Error(500, "Unknown token type. ID: " + token._id);
  }                                                                    //
};                                                                     //
                                                                       //
function dropInternal(sturdyRef, ownerPattern) {                       // 434
  // Drops `sturdyRef`, checking first that its owner matches `ownerPattern`.
                                                                       //
  var hashedSturdyRef = hashSturdyRef(sturdyRef);                      // 437
  var token = ApiTokens.findOne({ _id: hashedSturdyRef });             // 438
  if (!token) {                                                        // 439
    return;                                                            // 440
  }                                                                    //
                                                                       //
  check(token.owner, ownerPattern);                                    // 443
                                                                       //
  if (token.frontendRef) {                                             // 445
    if (token.frontendRef.notificationHandle) {                        // 446
      var notificationId = token.frontendRef.notificationHandle;       // 447
      globalDb.removeApiTokens({ _id: hashedSturdyRef });              // 448
      var anyToken = ApiTokens.findOne({ "frontendRef.notificationHandle": notificationId });
      if (!anyToken) {                                                 // 450
        // No other tokens referencing this notification exist, so dismiss the notification
        dismissNotification(notificationId);                           // 452
      }                                                                //
    } else {                                                           //
      throw new Error("Unknown frontend token type.");                 // 455
    }                                                                  //
  } else if (token.objectId) {                                         //
    if (token.objectId.wakeLockNotification) {                         // 458
      dropWakelock(token.grainId, token.objectId.wakeLockNotification);
    } else {                                                           //
      throw new Error("Unknown objectId token type.");                 // 461
    }                                                                  //
  } else {                                                             //
    throw new Error("Unknown token type.");                            // 464
  }                                                                    //
}                                                                      //
                                                                       //
function makeChildTokenInternal(rawParentToken, owner, requirements, tokenInfo) {
  var hashedParent = hashSturdyRef(rawParentToken);                    // 469
                                                                       //
  // If we don't have tokenInfo (we were called from SandstormCore.makeChildToken()), look up
  // the parent token now and use that.                                //
  // TODO(someday): I think this makeChildToken() will lose titles because of this.
  //   Option 1: Somehow pass info through the supervisor for it to pass back through
  //       makeChildToken().                                           //
  //   Option 2: Follow the chain of parentTokens here.                //
  //   Option 3: Maybe the title should actually be passed in a PowerboxDescriptor which the app
  //       is expected to thread through to offer(). This is analogous to how file transfers
  //       work: the file name is not normally stored in the content. Our petname system for
  //       grains already matches that model better -- and if we decide to get rid of said petname
  //       system and instead have everyone see the author's title, then we'd presumably stop
  //       storing it in the token `owner` field altogether, so this becomes moot.
  tokenInfo = tokenInfo || ApiTokens.findOne(hashedParent);            // 483
  if (!tokenInfo) {                                                    // 484
    throw new Error("parent token doesn't exist");                     // 485
  }                                                                    //
                                                                       //
  if (tokenInfo.identityId) {                                          // 488
    // This is a UiView capability. We need to use SandstormPermissions to create it in order
    // to properly denormalize fields.                                 //
                                                                       //
    if (owner.user) {                                                  // 492
      // Initialize the grain's title to a copy of the title set by the human user closest in the
      // sharing graph. It turns out that the "root token" is actually the token representing that
      // user, not the grain owner, because user-to-user sharing relationships are not parent-child
      // token relationships.                                          //
      var rootTitle = (((tokenInfo.owner || {}).grain || {}).saveLabel || {}).defaultText;
      if (!owner.user.title && rootTitle) {                            // 498
        owner.user.title = rootTitle;                                  // 499
      }                                                                //
    }                                                                  //
                                                                       //
    return SandstormPermissions.createNewApiToken(globalDb, { rawParentToken: rawParentToken }, tokenInfo.grainId, tokenInfo.petname, { allAccess: null }, owner).token;
  }                                                                    //
                                                                       //
  if (owner.user) {                                                    // 507
    throw new Error("can't save non-UiView with user as owner");       // 508
  }                                                                    //
                                                                       //
  var sturdyRef = generateSturdyRef();                                 // 511
  var hashedSturdyRef = hashSturdyRef(sturdyRef);                      // 512
                                                                       //
  var newTokenInfo = {                                                 // 514
    _id: hashedSturdyRef,                                              // 515
    parentToken: hashedParent,                                         // 516
    owner: owner,                                                      // 517
    created: new Date(),                                               // 518
    requirements: requirements                                         // 519
  };                                                                   //
                                                                       //
  // For non-UiView capabilities, we need not denormalize grainId, identityId, etc. into the child
  // token.                                                            //
                                                                       //
  ApiTokens.insert(newTokenInfo);                                      // 525
                                                                       //
  return sturdyRef;                                                    // 527
};                                                                     //
                                                                       //
function SandstormCoreFactoryImpl() {}                                 // 530
                                                                       //
SandstormCoreFactoryImpl.prototype.getSandstormCore = function (grainId) {
  return { core: makeSandstormCore(grainId) };                         // 534
};                                                                     //
                                                                       //
makeSandstormCoreFactory = function () {                               // 537
  return new Capnp.Capability(new SandstormCoreFactoryImpl(), SandstormCoreFactory);
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=core.js.map
