(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var ECMAScript = Package.ecmascript.ECMAScript;
var SandstormDb = Package['sandstorm-db'].SandstormDb;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var ContactProfiles;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                         //
// packages/sandstorm-contacts/contacts-server.js                                          //
//                                                                                         //
/////////////////////////////////////////////////////////////////////////////////////////////
                                                                                           //
// Sandstorm - Personal Cloud Sandbox                                                      //
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                   //
// All rights reserved.                                                                    //
//                                                                                         //
// Licensed under the Apache License, Version 2.0 (the "License");                         //
// you may not use this file except in compliance with the License.                        //
// You may obtain a copy of the License at                                                 //
//                                                                                         //
//   http://www.apache.org/licenses/LICENSE-2.0                                            //
//                                                                                         //
// Unless required by applicable law or agreed to in writing, software                     //
// distributed under the License is distributed on an "AS IS" BASIS,                       //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                //
// See the License for the specific language governing permissions and                     //
// limitations under the License.                                                          //
                                                                                           //
Meteor.publish("contactProfiles", function () {                                            // 17
  var db = this.connection.sandstormDb;                                                    // 18
  var _this = this;                                                                        // 19
                                                                                           //
  // We maintain a map from identity IDs to live query handles that track profile changes.
  var contactIdentities = {};                                                              // 22
                                                                                           //
  function addIdentityOfContact(contact) {                                                 // 24
    if (!(contact.identityId in contactIdentities)) {                                      // 25
      var user = Meteor.users.findOne({ _id: contact.identityId });                        // 26
      if (user) {                                                                          // 27
        SandstormDb.fillInProfileDefaults(user);                                           // 28
        SandstormDb.fillInIntrinsicName(user);                                             // 29
        var filteredUser = _.pick(user, "_id", "profile");                                 // 30
        _this.added("contactProfiles", user._id, filteredUser);                            // 31
      }                                                                                    //
                                                                                           //
      contactIdentities[contact.identityId] = Meteor.users.find({ _id: contact.identityId }, { fields: { profile: 1 } }).observeChanges({
        changed: function (id, fields) {                                                   // 36
          _this.changed("contactProfiles", id, fields);                                    // 37
        }                                                                                  //
      });                                                                                  //
    }                                                                                      //
  }                                                                                        //
                                                                                           //
  var cursor = db.collections.contacts.find({ ownerId: this.userId });                     // 43
                                                                                           //
  var handle = cursor.observe({                                                            // 45
    added: function (contact) {                                                            // 46
      addIdentityOfContact(contact);                                                       // 47
    },                                                                                     //
                                                                                           //
    changed: function (contact) {                                                          // 50
      addIdentityOfContact(contact);                                                       // 51
    },                                                                                     //
                                                                                           //
    removed: function (contact) {                                                          // 54
      _this.removed("contactProfiles", contact.identityId);                                // 55
      contactIdentities[contact.identityId].stop();                                        // 56
      delete contactIdentities[contact.identityId];                                        // 57
    }                                                                                      //
  });                                                                                      //
  this.ready();                                                                            // 60
                                                                                           //
  this.onStop(function () {                                                                // 62
    handle.stop();                                                                         // 63
    Object.keys(contactIdentities).forEach(function (identityId) {                         // 64
      contactIdentities[identityId].stop();                                                // 65
      delete contactIdentities[identityId];                                                // 66
    });                                                                                    //
  });                                                                                      //
});                                                                                        //
/////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-contacts'] = {
  ContactProfiles: ContactProfiles
};

})();

//# sourceMappingURL=sandstorm-contacts.js.map
