(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var ECMAScript = Package.ecmascript.ECMAScript;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, ApiHosts, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, FeatureKey, successfulReferralsCount, getUserQuota, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, sendReferralProgramNotification, roleAssignmentPattern, SandstormDb, appName, addStaticAsset, userPictureUrl, fetchPicture;

(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/db.js                                                                                        //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                                 //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                              //
// All rights reserved.                                                                                               //
//                                                                                                                    //
// Licensed under the Apache License, Version 2.0 (the "License");                                                    //
// you may not use this file except in compliance with the License.                                                   //
// You may obtain a copy of the License at                                                                            //
//                                                                                                                    //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                       //
//                                                                                                                    //
// Unless required by applicable law or agreed to in writing, software                                                //
// distributed under the License is distributed on an "AS IS" BASIS,                                                  //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                           //
// See the License for the specific language governing permissions and                                                //
// limitations under the License.                                                                                     //
                                                                                                                      //
// This file defines the database schema.                                                                             //
                                                                                                                      //
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every                        //
// query it makes, so you can see if it's doing queries too often, etc.                                               //
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                               // 21
  (function () {                                                                                                      //
    var oldFind = Mongo.Collection.prototype.find;                                                                    // 22
    Mongo.Collection.prototype.find = function () {                                                                   // 23
      console.log(this._prefix, arguments);                                                                           // 24
      return oldFind.apply(this, arguments);                                                                          // 25
    };                                                                                                                //
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
// Helper so that we don't have to if (Meteor.isServer) before declaring indexes.                                     //
if (Meteor.isServer) {                                                                                                // 30
  Mongo.Collection.prototype.ensureIndexOnServer = Mongo.Collection.prototype._ensureIndex;                           // 31
} else {                                                                                                              //
  Mongo.Collection.prototype.ensureIndexOnServer = function () {};                                                    // 33
}                                                                                                                     //
                                                                                                                      //
// TODO(soon): Systematically go through this file and add ensureIndexOnServer() as needed.                           //
                                                                                                                      //
// Users = new Mongo.Collection("users");                                                                             //
// The users collection is special and can be accessed through `Meteor.users`.                                        //
// See https://docs.meteor.com/#/full/meteor_users.                                                                   //
//                                                                                                                    //
// There are two distinct types of entries in the users collection: identities and accounts. An                       //
// identity contains personal profile information and typically includes some intrinsic method for                    //
// authenticating as the owner of that information.                                                                   //
//                                                                                                                    //
// An account is an owner of app actions, grains, contacts, notifications, and payment info.                          //
// Each account can have multiple identities linked to it. To log in as an account you must first                     //
// authenticate as one of its linked identities.                                                                      //
//                                                                                                                    //
// Every user contains the following fields:                                                                          //
//   _id: Unique string ID. For accounts, this is random. For identities, this is the globally                        //
//        stable SHA-256 ID of this identity, hex-encoded.                                                            //
//   createdAt: Date when this entry was added to the collection.                                                     //
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                               //
//   services: Object containing login data used by Meteor authentication services.                                   //
//   expires: Date when this user should be deleted. Only present for demo users.                                     //
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.                       //
//                     TODO(cleanup): Unlike other dates in our database, this is stored as a number                  //
//                     rather than as a Date object. We should fix that.                                              //
//   appDemoId: If present and non-null, then the user is a demo user who arrived via an /appdemo/                    //
//              link. This field contains the app ID of the app that the user started out demoing.                    //
//              Unlike the `expires` field, this field is not cleared when the user upgrades from                     //
//              being a demo user.                                                                                    //
//                                                                                                                    //
// Identity users additionally contain the following fields:                                                          //
//   profile: Object containing the data that will be shared with users and grains that come into                     //
//            contact with this identity. Includes the following fields:                                              //
//       service: String containing the name of this identity's authentication method.                                //
//       name: String containing the chosen display name of the identity.                                             //
//       handle: String containing the identity's preferred handle.                                                   //
//       picture: _id into the StaticAssets table for the identity's picture. If not present,                         //
//                an identicon will be used.                                                                          //
//       pronoun: One of "male", "female", "neutral", or "robot".                                                     //
//   unverifiedEmail: If present, a string containing an email address specified by the user.                         //
//   referredBy: ID of the Account that referred this Identity.                                                       //
//                                                                                                                    //
// Account users additionally contain the following fields:                                                           //
//   loginIdentities: Array of identity objects, each of which may include the following fields.                      //
//       id: The globally-stable SHA-256 ID of this identity, hex-encoded.                                            //
//   nonloginIdentities: Array of identity objects, of the same form as `loginIdentities`. We use                     //
//                       a separate array here so that we can use a Mongo index to enforce the                        //
//                       invariant that an identity only be a login identity for a single account.                    //
//   primaryEmail: String containing this account's primary email address. Must be a verified adress                  //
//                 of one of this account's linked identities. Call SandstormDb.getUserEmails()                       //
//                 to do this checking automatically.                                                                 //
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.                 //
//   signupKey: If this is an invited user, then this field contains their signup key.                                //
//   signupNote: If the user was invited through a link, then this field contains the note that the                   //
//               inviter admin attached to the key.                                                                   //
//   signupEmail: If the user was invited by email, then this field contains the email address that                   //
//                the invite was sent to.                                                                             //
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's                   //
//                       terms of service.                                                                            //
//   plan: _id of an entry in the Plans table which determines the user's quota.                                      //
//   planBonus: {storage, compute, grains} bonus amounts to add to the user's plan. The payments                      //
//              module writes data here; we merely read it. Missing fields should be treated as                       //
//              zeroes. Does not yet include referral bonus, which is calculated separately.                          //
//              TODO(cleanup): Use for referral bonus too.                                                            //
//   storageUsage: Number of bytes this user is currently storing.                                                    //
//   payments: Object defined by payments module, if loaded.                                                          //
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                                //
//   accessRequests: Object containing the following fields; used to limit spam.                                      //
//       count: Number of "request access" emails during sent during the current interval.                            //
//       resetAt: Date when the count should be reset.                                                                //
//   referredByComplete: ID of the Account that referred this Account. If this is set, we                             //
//                        stop writing new referredBy values onto Identities for this account.                        //
//   referredCompleteDate: The Date at which the completed referral occurred.                                         //
//   referredIdentityIds: List of Identity IDs that this Account has referred. This is used for                       //
//                        reliably determining which Identity's names are safe to display.                            //
//   experiments: Object where each field is an experiment that the user is in, and each value                        //
//           is the parameters for that experiment. Typically, the value simply names which                           //
//           experiment group which the user is in, where "control" is one group. If an experiment                    //
//           is not listed, then the user should not be considered at all for the purpose of that                     //
//           experiment. Each experiment may define a point in time where users not already in the                    //
//           experiment may be added to it and assigned to a group (for example, at user creation                     //
//           time). Current experiments:                                                                              //
//       firstTimeBillingPrompt: Value is "control" or "test". Users are assigned to groups at                        //
//               account creation on servers where billing is enabled (i.e. Oasis). Users in the                      //
//               test group will see a plan selection dialog and asked to make an explitic choice                     //
//               (possibly "free") before they can create grains (but not when opening someone                        //
//               else's shared grain). The goal of the experiment is to determine whether this                        //
//               prompt scares users away -- and also whether it increases paid signups.                              //
//   stashedOldUser: A complete copy of this user from before the accounts/identities migration.                      //
//                   TODO(cleanup): Delete this field once we're sure it's safe to do so.                             //
                                                                                                                      //
Meteor.users.ensureIndexOnServer("services.google.email", { sparse: 1 });                                             // 126
Meteor.users.ensureIndexOnServer("services.github.emails.email", { sparse: 1 });                                      // 127
Meteor.users.ensureIndexOnServer("services.email.email", { unique: 1, sparse: 1 });                                   // 128
Meteor.users.ensureIndexOnServer("loginIdentities.id", { unique: 1, sparse: 1 });                                     // 129
Meteor.users.ensureIndexOnServer("nonloginIdentities.id", { sparse: 1 });                                             // 130
Meteor.users.ensureIndexOnServer("services.google.id", { unique: 1, sparse: 1 });                                     // 131
Meteor.users.ensureIndexOnServer("services.github.id", { unique: 1, sparse: 1 });                                     // 132
                                                                                                                      //
// TODO(cleanup): This index is obsolete; delete it.                                                                  //
Meteor.users.ensureIndexOnServer("identities.id", { unique: 1, sparse: 1 });                                          // 135
                                                                                                                      //
Packages = new Mongo.Collection("packages");                                                                          // 137
// Packages which are installed or downloading.                                                                       //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                                   //
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"                  //
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                               //
//       >1 = download byte count.                                                                                    //
//   error:  If status is "failed", error message string.                                                             //
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                         //
//   appId:  If status is "ready", the application ID string.  Packages representing different                        //
//       versions of the same app have the same appId.  The spk tool defines the app ID format                        //
//       and can cryptographically verify that a package belongs to a particular app ID.                              //
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package                       //
//       collector should at some point check whether there are any other references and, if not,                     //
//       delete the package.                                                                                          //
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.                        //
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up                   //
//     even if it has no users.                                                                                       //
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                          //
//     packager.                                                                                                      //
                                                                                                                      //
DevPackages = new Mongo.Collection("devpackages");                                                                    // 159
// List of packages currently made available via the dev tools running on the local machine.                          //
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool                    //
// on the local machine to publish an under-development app to this server. That should only ever                     //
// happen on developers' desktop machines.                                                                            //
//                                                                                                                    //
// While a dev package is published, it automatically appears as installed by every user of the                       //
// server, and it overrides all packages with the same application ID. If any instances of those                      //
// packages are currently open, they are killed and reset on publish.                                                 //
//                                                                                                                    //
// When the dev tool disconnects, the package is automatically unpublished, and any open instances                    //
// are again killed and refreshed.                                                                                    //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  The package ID string (as with Packages._id).                                                              //
//   appId: The app ID this package is intended to override (as with Packages.appId).                                 //
//   timestamp:  Time when the package was last updated. If this changes while the package is                         //
//     published, all running instances are reset. This is used e.g. to reset the app each time                       //
//     changes are made to the source code.                                                                           //
//   manifest:  The app's manifest, as with Packages.manifest.                                                        //
                                                                                                                      //
UserActions = new Mongo.Collection("userActions");                                                                    // 180
// List of actions that each user has installed which create new grains.  Each app may install                        //
// some number of actions (usually, one).                                                                             //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   userId:  User who has installed this action.                                                                     //
//   packageId:  Package used to run this action.                                                                     //
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                               //
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                                //
//       that clients can access it without subscribing to the Packages collection.                                   //
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                           //
//       searchability.                                                                                               //
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                              //
//   title: JSON-encoded LocalizedText title for this action, e.g.                                                    //
//       `{defaultText: "New Spreadsheet"}`.                                                                          //
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.                       //
//   command:  Manifest.Command to run this action (see package.capnp).                                               //
                                                                                                                      //
Grains = new Mongo.Collection("grains");                                                                              // 199
// Grains belonging to users.                                                                                         //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   packageId:  _id of the package of which this grain is an instance.                                               //
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                               //
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                           //
//       searchability.                                                                                               //
//   userId: The _id of the account that owns this grain.                                                             //
//   identityId: The identity with which the owning account prefers to open this grain.                               //
//   title:  Human-readable string title, as chosen by the user.                                                      //
//   lastUsed:  Date when the grain was last used by a user.                                                          //
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                                   //
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent                   //
//                   time a session to this grain was opened.                                                         //
//                                                                                                                    //
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and                    //
// web publishing functionality without powerbox support; they will be replaced once the powerbox                     //
// is implemented.                                                                                                    //
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and                    //
//       web publishing. This field is initialized when first requested by the app.                                   //
                                                                                                                      //
RoleAssignments = new Mongo.Collection("roleAssignments");                                                            // 222
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges                   //
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                            //
// collection.                                                                                                        //
                                                                                                                      //
Contacts = new Mongo.Collection("contacts");                                                                          // 227
// Edges in the social graph.                                                                                         //
//                                                                                                                    //
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob                        //
// will show up in her user-picker UI for actions like share-by-identity.                                             //
//                                                                                                                    //
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of                       //
// Bob's.                                                                                                             //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: random                                                                                                      //
//   ownerId: The accountId of the user account who owns this contact.                                                //
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies                       //
//            the contact to the owner.                                                                               //
//   created: Date when this contact was created.                                                                     //
//   identityId: The `_id` of the user whose contact info this contains.                                              //
                                                                                                                      //
Sessions = new Mongo.Collection("sessions");                                                                          // 244
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.                   //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   grainId:  _id of the grain to which this session is connected.                                                   //
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the                    //
//       '*' in WILDCARD_HOST.                                                                                        //
//   tabId: Random value unique to the grain tab in which this session is displayed. Typically                        //
//       every session has a different `tabId`, but embedded sessions (including in the powerbox)                     //
//       have the same `tabId` as the outer session.                                                                  //
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some                       //
//       period.                                                                                                      //
//   userId:  User ID of the user who owns this session.                                                              //
//   identityId:  Identity ID of the user who owns this session.                                                      //
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens                     //
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),                        //
//       anonymous users can get access without an API token and so neither userId nor hashedToken                    //
//       are present.                                                                                                 //
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                          //
//       interaction.                                                                                                 //
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                                  //
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated                    //
//       until newSession is called on the UiView.                                                                    //
//   permissions: The permissions for the current identity on this UiView. This isn't populated                       //
//       until newSession is called on the UiView.                                                                    //
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first                  //
//       request                                                                                                      //
                                                                                                                      //
SignupKeys = new Mongo.Collection("signupKeys");                                                                      // 273
// Invite keys which may be used by users to get access to Sandstorm.                                                 //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   used:  Boolean indicating whether this key has already been consumed.                                            //
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                         //
//   email: If this key was sent as an email invite, the email address to which it was sent.                          //
                                                                                                                      //
ActivityStats = new Mongo.Collection("activityStats");                                                                // 282
// Contains usage statistics taken on a regular interval. Each entry is a data point.                                 //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   timestamp: Date when measurements were taken.                                                                    //
//   daily: Contains stats counts pertaining to the last day before the sample time.                                  //
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                          //
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                          //
//                                                                                                                    //
// Each of daily, weekly, and monthly contains:                                                                       //
//   activeUsers: The number of unique users who have used a grain on the server in the time                          //
//       interval. Only counts logged-in users.                                                                       //
//   demoUsers: Demo users.                                                                                           //
//   appDemoUsers: Users that came in through "app demo".                                                             //
//   activeGrains: The number of unique grains that have been used in the time interval.                              //
//   apps: An object indexed by app ID recording, for each app:                                                       //
//       owners: Number of unique owners of this app (counting only grains that still exist).                         //
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting                   //
//         only grains that still exist).                                                                             //
//       grains: Number of active grains of this app (that still exist).                                              //
//       deleted: Number of non-demo grains of this app that were deleted.                                            //
//       demoed: Number of demo grains created and expired.                                                           //
//       appDemoUsers: Number of app demos initiated with this app.                                                   //
                                                                                                                      //
DeleteStats = new Mongo.Collection("deleteStats");                                                                    // 306
// Contains records of objects that were deleted, for stat-keeping purposes.                                          //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                            //
//   lastActive: Date of the user's or grain's last activity.                                                         //
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they                    //
//     arrived to demo. For others, undefined.                                                                        //
                                                                                                                      //
FileTokens = new Mongo.Collection("fileTokens");                                                                      // 315
// Tokens corresponding to backup files that are currently stored on the server. A user receives                      //
// a token when they create a backup file (either by uploading it, or by backing up one of their                      //
// grains) and may use the token to read the file (either to download it, or to restore a new                         //
// grain from it).                                                                                                    //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The unguessable token string.                                                                         //
//   name:      Suggested filename.                                                                                   //
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.                       //
                                                                                                                      //
ApiTokens = new Mongo.Collection("apiTokens");                                                                        // 326
// Access tokens for APIs exported by apps.                                                                           //
//                                                                                                                    //
// Originally API tokens were only used by external users through the HTTP API endpoint. However,                     //
// now they are also used to implement SturdyRefs, not just held by external users, but also when                     //
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                          //
// `restore()`, and `drop()` methods in `grain.capnp` (on `SandstormApi`, `AppPersistent`, and                        //
// `MainView`) -- the fields of type `Data` are API tokens.                                                           //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                          //
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)                     //
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.                      //
//              That is, the UiView's newSession() method will intersect the requested permissions                    //
//              with this identity's permissions before forwarding on to the underlying app. If                       //
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is                    //
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may                       //
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that                    //
//              this is NOT the identity against which the `requiredPermissions` parameter of                         //
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.                  //
//   accountId: For tokens where `identityId` is set, the `_id` (in the Users table) of the account                   //
//              that created the token.                                                                               //
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded                        //
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These                   //
//              permissions will be intersected with those held by `identityId` when the view is                      //
//              opened.                                                                                               //
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as                   //
//              anonymous rather than as directly associated with `identityId`. This has no effect                    //
//              on the permissions granted.                                                                           //
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by                     //
//              the app or its supervisor (whereas without this it strictly represents UiView).                       //
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).                     //
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                            //
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                              //
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,                   //
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object                   //
//              containing exactly one of the following fields:                                                       //
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                                   //
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                          //
//                           `Notifications` collection.                                                              //
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this                     //
//                  will be moved out of the frontend and into the backend, but we'll migrate the                     //
//                  database when that happens. This field contains the boolean true to signify that                  //
//                  it has been set.                                                                                  //
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                                    //
//       emailVerifier: An EmailVerifier capability that is implemented by the frontend. The                          //
//                      value is an object containing the field `services`, which itself is a                         //
//                      list of names of identity providers that are trusted to verify addresses.                     //
//                      If `services` is omitted or falsy, all configured identity providers are                      //
//                      trusted. Note that a malicious user could specify invalid names in the                        //
//                      list; they should be ignored.                                                                 //
//       verifiedEmail: An VerifiedEmail capability that is implemented by the frontend.                              //
//                      An object containing `verifierId`, `tabId`, and `address`.                                    //
//   parentToken: If present, then this token represents exactly the capability represented by                        //
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated                   //
//              by `roleAssignment` (if present). To facilitate permissions computations, if the                      //
//              capability is a UiView, then `grainId` is set to the backing grain, `identityId`                      //
//              is set to the identity that shared the view, and `accountId` is set to the account                    //
//              that shared the view. Neither `objectId` nor `frontendRef` is present when                            //
//              `parentToken` is present.                                                                             //
//   petname:   Human-readable label for this access token, useful for identifying tokens for                         //
//              revocation. This should be displayed when visualizing incoming capabilities to                        //
//              the grain identified by `grainId`.                                                                    //
//   created:   Date when this token was created.                                                                     //
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may                   //
//              become un-revoked in the future.                                                                      //
//   expires:   Optional expiration Date. If undefined, the token does not expire.                                    //
//   owner:     A `ApiTokenOwner` (defined in `supervisor.capnp`, stored as a JSON object)                            //
//              as passed to the `save()` call that created this token. If not present, treat                         //
//              as `webkey` (the default for `ApiTokenOwner`).                                                        //
//   expiresIfUnused:                                                                                                 //
//              Optional Date after which the token, if it has not been used yet, expires.                            //
//              This field should be cleared on a token's first use.                                                  //
//   requirements: List of conditions which must hold for this token to be considered valid.                          //
//              Semantically, this list specifies the powers which were *used* to originally                          //
//              create the token. If any condition in the list becomes untrue, then the token must                    //
//              be considered revoked, and all live refs and sturdy refs obtained transitively                        //
//              through it must also become revoked. Each item is the JSON serialization of the                       //
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                                        //
//   hasApiHost: If true, there is an entry in ApiHosts for this token, which will need to be                         //
//              cleaned up when the token is.                                                                         //
//                                                                                                                    //
// It is important to note that a token's owner and provider are independent from each other. To                      //
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:                   //
//                                                                                                                    //
// struct ApiToken {                                                                                                  //
//   owner :ApiTokenOwner;                                                                                            //
//   provider :union {                                                                                                //
//     grain :group {                                                                                                 //
//       grainId :Text;                                                                                               //
//       union {                                                                                                      //
//         uiView :group {                                                                                            //
//           identityId :Text;                                                                                        //
//           roleAssignment :RoleAssignment;                                                                          //
//           forSharing :Bool;                                                                                        //
//         }                                                                                                          //
//         objectId :SupervisorObjectId;                                                                              //
//       }                                                                                                            //
//     }                                                                                                              //
//     frontendRef :union {                                                                                           //
//       notificationHandle :Text;                                                                                    //
//       ipNetwork :Bool;                                                                                             //
//       ipInterface :Bool;                                                                                           //
//       emailVerifier :group {                                                                                       //
//         services :List(String);                                                                                    //
//       }                                                                                                            //
//       verifiedEmail :group {                                                                                       //
//         verifierId :Text;                                                                                          //
//         tabId :Text;                                                                                               //
//         address :Text;                                                                                             //
//       }                                                                                                            //
//     }                                                                                                              //
//     child :group {                                                                                                 //
//       parentToken :Text;                                                                                           //
//       union {                                                                                                      //
//         uiView :group {                                                                                            //
//           grainId :Text;                                                                                           //
//           identityId :Text;                                                                                        //
//           roleAssignment :RoleAssignment = (allAccess = ());                                                       //
//         }                                                                                                          //
//         other :Void;                                                                                               //
//       }                                                                                                            //
//     }                                                                                                              //
//   }                                                                                                                //
//   requirements: List(Supervisor.MembraneRequirement);                                                              //
//   ...                                                                                                              //
// }                                                                                                                  //
                                                                                                                      //
ApiHosts = new Mongo.Collection("apiHosts");                                                                          // 454
// Allows defining some limited static behavior for an API host when accessed unauthenticated. This                   //
// mainly exists to allow backwards-compatibility with client applications that expect to be able                     //
// to probe an API host without authentication to determine capabilities such as DAV protocols                        //
// supported, before authenticating to perform real requests. An app can specify these properties                     //
// when creating an offerTemplate.                                                                                    //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:          apiHostIdHashForToken() of the corresponding API token.                                            //
//   hash2:        hash(hash(token)), aka hash(ApiToken._id). Used to allow ApiHosts to be cleaned                    //
//                 up when ApiTokens are deleted.                                                                     //
//   options:      Specifies how to respond to unauthenticated OPTIONS requests on this host.                         //
//                 This is an object containing fields:                                                               //
//     dav:        List of strings specifying DAV header `compliance-class`es, e.g. "1" or                            //
//                 "calendar-access". https://tools.ietf.org/html/rfc4918#section-10.1                                //
//   resources:    Object mapping URL paths (including initial '/') to static HTTP responses to                       //
//                 give when those paths are accessed unauthenticated. Due to Mongo disliking '.'                     //
//                 and '$' in keys, these characters must be escaped as '\uFF0E' and '\uFF04'                         //
//                 (see SandstormDb.escapeMongoKey). Each value in this map is an object with                         //
//                 fields:                                                                                            //
//     type:       Content-Type.                                                                                      //
//     language:   Content-Language.                                                                                  //
//     encoding:   Content-Encoding.                                                                                  //
//     body:       Entity-body as a string or buffer.                                                                 //
                                                                                                                      //
Notifications = new Mongo.Collection("notifications");                                                                // 479
// Notifications for a user.                                                                                          //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:          random                                                                                             //
//   grainId:      The grain originating this notification, if any.                                                   //
//   userId:       The user receiving the notification.                                                               //
//   text:         The JSON-ified LocalizedText to display in the notification.                                       //
//   isUnread:     Boolean indicating if this notification is unread.                                                 //
//   timestamp:    Date when this notification was last updated                                                       //
//   ongoing:      If present, this is an ongoing notification, and this field contains an                            //
//                 ApiToken referencing the `OngoingNotification` capability.                                         //
//   admin:        If present, this is a notification intended for an admin.                                          //
//     action:     If present, this is a (string) link that the notification should direct the                        //
//                 admin to.                                                                                          //
//     type:       The type of notification (currently only "reportStats").                                           //
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds                    //
//                 as keys.                                                                                           //
//     $appId:     The appId that has an outstanding update.                                                          //
//       packageId: The packageId that it will update to.                                                             //
//       name: The name of the app. (appTitle from package.manifest)                                                  //
//       version: The app's version number. (appVersion from package.manifest)                                        //
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)          //
//   referral:     If this boolean field is true, then treat this notification as a referral                          //
//                 notification. This causes text to be ignored, since we need custom logic.                          //
//   mailingListBonus: Like `referral`, but notify the user about the mailing list bonus. This is                     //
//                 a one-time notification only to Oasis users who existed when the bonus program                     //
//                 was implemented.                                                                                   //
                                                                                                                      //
StatsTokens = new Mongo.Collection("statsTokens");                                                                    // 508
// Access tokens for the Stats collection                                                                             //
//                                                                                                                    //
// These tokens are used for accessing the ActivityStats collection remotely                                          //
// (ie. from a dashboard webapp)                                                                                      //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                                 //
                                                                                                                      //
Misc = new Mongo.Collection("misc");                                                                                  // 517
// Miscellaneous configuration and other settings                                                                     //
//                                                                                                                    //
// This table is currently only used for persisting BASE_URL from one session to the next,                            //
// but in general any miscellaneous settings should go in here                                                        //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The name of the setting. eg. "BASE_URL"                                                               //
//   value:     The value of the setting.                                                                             //
                                                                                                                      //
Settings = new Mongo.Collection("settings");                                                                          // 527
// Settings for this Sandstorm instance go here. They are configured through the adminSettings                        //
// route. This collection differs from misc in that any admin user can update it through the admin                    //
// interface.                                                                                                         //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The name of the setting. eg. "smtpConfig"                                                             //
//   value:     The value of the setting.                                                                             //
//   automaticallyReset: Sometimes the server needs to automatically reset a setting. When it does                    //
//                       so, it will also write an object to this field indicating why the reset was                  //
//                       needed. That object can have the following variants:                                         //
//       baseUrlChangedFrom: The reset was due to BASE_URL changing. This field contains a string                     //
//                           with the old BASE_URL.                                                                   //
//                                                                                                                    //
//   potentially other fields that are unique to the setting                                                          //
                                                                                                                      //
Migrations = new Mongo.Collection("migrations");                                                                      // 543
// This table tracks which migrations we have applied to this instance.                                               //
// It contains a single entry:                                                                                        //
//   _id:       "migrations_applied"                                                                                  //
//   value:     The number of migrations this instance has successfully completed.                                    //
                                                                                                                      //
StaticAssets = new Mongo.Collection("staticAssets");                                                                  // 549
// Collection of static assets served up from the Sandstorm server's "static" host. We only                           //
// support relatively small assets: under 1MB each.                                                                   //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       Random ID; will be used in the URL.                                                                   //
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                           //
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                             //
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                                   //
//   content:   The asset content (byte buffer).                                                                      //
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't                   //
//       have transactions, this needs to bias towards over-counting; a backup GC could be used                       //
//       to catch leaked assets, although it's probably not a big deal in practice.                                   //
                                                                                                                      //
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                                        // 563
// Collection of tokens representing a single-use permission to upload an asset, such as a new                        //
// profile picture.                                                                                                   //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       Random ID.                                                                                            //
//   purpose:   Contains one of the following, indicating how the asset is to be used:                                //
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                         //
//           userId: User whose picture shall be replaced.                                                            //
//           identityId: Which of the user's identities shall be updated.                                             //
//   expires:   Time when this token will go away if unused.                                                          //
                                                                                                                      //
Plans = new Mongo.Collection("plans");                                                                                // 575
// Subscription plans, which determine quota.                                                                         //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                               //
//   storage: Number of bytes this user is allowed to store.                                                          //
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                         //
//   computeLabel: Label to display to the user describing this plan's compute units.                                 //
//   grains: Total number of grains this user can create (often `Infinity`).                                          //
//   price: Price per month in US cents.                                                                              //
//   hidden: If true, a user cannot switch to this plan, but some users may be on it and are                          //
//       allowed to switch away.                                                                                      //
//   title: Title from display purposes. If missing, default to capitalizing _id.                                     //
                                                                                                                      //
AppIndex = new Mongo.Collection("appIndex");                                                                          // 589
// A mirror of the data from the App Market index                                                                     //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: the appId of the app                                                                                        //
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket                   //
                                                                                                                      //
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                            // 596
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package                    //
// by that user is installed, as well as if the keybase profile is requested and not already                          //
// present for some reason.                                                                                           //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                                  //
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                                   //
//   handle: Keybase handle.                                                                                          //
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase                     //
//     docs for details. We also add a boolean "status" field to each proof indicating whether                        //
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                         //
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it                    //
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the                    //
//     server is currently actively checking this proof). Note that if a check fails due to network                   //
//     errors, the status goes back to "unverified".                                                                  //
//                                                                                                                    //
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                         //
//       for now and we just trust Keybase.                                                                           //
                                                                                                                      //
FeatureKey = new Mongo.Collection("featureKey");                                                                      // 616
// Responsible for storing the current feature key that is active on the server.  Contains a single                   //
// document with two keys:                                                                                            //
//                                                                                                                    //
//   _id: "currentFeatureKey"                                                                                         //
//   value: the still-signed, binary-encoded feature key                                                              //
//          (a feature key with comments removed and base64 decoded)                                                  //
//                                                                                                                    //
// This is only intended to be visible on the server.                                                                 //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 626
  Meteor.publish("credentials", function () {                                                                         // 627
    // Data needed for isSignedUp() and isAdmin() to work.                                                            //
                                                                                                                      //
    if (this.userId) {                                                                                                // 630
      return [Meteor.users.find({ _id: this.userId }, { fields: { signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,
          plan: 1, planBonus: 1, hasCompletedSignup: 1, experiments: 1,                                               // 634
          referredIdentityIds: 1 } }), Plans.find()];                                                                 // 635
    } else {                                                                                                          //
      return [];                                                                                                      // 639
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
var countReferrals = function (user) {                                                                                // 644
  var referredIdentityIds = user.referredIdentityIds;                                                                 // 645
  return referredIdentityIds && referredIdentityIds.length || 0;                                                      // 646
};                                                                                                                    //
                                                                                                                      //
var calculateReferralBonus = function (user) {                                                                        // 649
  // This function returns an object of the form:                                                                     //
  //                                                                                                                  //
  // - {grains: 0, storage: 0}                                                                                        //
  //                                                                                                                  //
  // which are extra resources this account gets as part of participating in the referral                             //
  // program. (Storage is measured in bytes, as usual for plans.)                                                     //
                                                                                                                      //
  // TODO(cleanup): Consider moving referral bonus logic into Oasis payments module (since it's                       //
  //   payments-specific) and aggregating into `planBonus`.                                                           //
                                                                                                                      //
  // Authorization note: Only call this if accountId is the current user!                                             //
  var isPaid = user.plan && user.plan !== "free";                                                                     // 661
                                                                                                                      //
  successfulReferralsCount = countReferrals(user);                                                                    // 663
  if (isPaid) {                                                                                                       // 664
    var maxPaidStorageBonus = 30 * 1e9;                                                                               // 665
    return { grains: 0,                                                                                               // 666
      storage: Math.min(successfulReferralsCount * 2 * 1e9, maxPaidStorageBonus) };                                   // 667
  } else {                                                                                                            //
    var maxFreeStorageBonus = 2 * 1e9;                                                                                // 671
    var bonus = {                                                                                                     // 672
      storage: Math.min(successfulReferralsCount * 50 * 1e6, maxFreeStorageBonus)                                     // 673
    };                                                                                                                //
    if (successfulReferralsCount > 0) {                                                                               // 677
      bonus.grains = Infinity;                                                                                        // 678
    } else {                                                                                                          //
      bonus.grains = 0;                                                                                               // 680
    }                                                                                                                 //
                                                                                                                      //
    return bonus;                                                                                                     // 683
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
getUserQuota = function (user) {                                                                                      // 687
  var plan = Plans.findOne(user.plan || "free");                                                                      // 688
  var referralBonus = calculateReferralBonus(user);                                                                   // 689
  var bonus = user.planBonus || {};                                                                                   // 690
  var userQuota = {                                                                                                   // 691
    storage: plan.storage + referralBonus.storage + (bonus.storage || 0),                                             // 692
    grains: plan.grains + referralBonus.grains + (bonus.grains || 0),                                                 // 693
    compute: plan.compute + (bonus.compute || 0)                                                                      // 694
  };                                                                                                                  //
  return userQuota;                                                                                                   // 696
};                                                                                                                    //
                                                                                                                      //
isUserOverQuota = function (user) {                                                                                   // 699
  // Return false if user has quota space remaining, true if it is full. When this returns true,                      //
  // we will not allow the user to create new grains, though they may be able to open existing ones                   //
  // which may still increase their storage usage.                                                                    //
  //                                                                                                                  //
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                                 //
                                                                                                                      //
  if (!Meteor.settings["public"].quotaEnabled || user.isAdmin) return false;                                          // 706
                                                                                                                      //
  var plan = getUserQuota(user);                                                                                      // 708
  if (plan.grains < Infinity) {                                                                                       // 709
    var count = Grains.find({ userId: user._id }, { fields: {}, limit: plan.grains }).count();                        // 710
    if (count >= plan.grains) return "outOfGrains";                                                                   // 711
  }                                                                                                                   //
                                                                                                                      //
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                            // 714
};                                                                                                                    //
                                                                                                                      //
isUserExcessivelyOverQuota = function (user) {                                                                        // 717
  // Return true if user is so far over quota that we should prevent their existing grains from                       //
  // running at all.                                                                                                  //
  //                                                                                                                  //
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                                 //
                                                                                                                      //
  if (!Meteor.settings["public"].quotaEnabled || user.isAdmin) return false;                                          // 723
                                                                                                                      //
  var quota = getUserQuota(user);                                                                                     // 725
                                                                                                                      //
  // quota.grains = Infinity means unlimited grains. IEEE754 defines Infinity == Infinity.                            //
  if (quota.grains < Infinity) {                                                                                      // 728
    var count = Grains.find({ userId: user._id }, { fields: {}, limit: quota.grains * 2 }).count();                   // 729
    if (count >= quota.grains * 2) return "outOfGrains";                                                              // 730
  }                                                                                                                   //
                                                                                                                      //
  return quota && user.storageUsage && user.storageUsage >= quota.storage * 1.2 && "outOfStorage";                    // 733
};                                                                                                                    //
                                                                                                                      //
isAdmin = function () {                                                                                               // 736
  // Returns true if the user is the administrator.                                                                   //
                                                                                                                      //
  var user = Meteor.user();                                                                                           // 739
  if (user && user.isAdmin) {                                                                                         // 740
    return true;                                                                                                      // 741
  } else {                                                                                                            //
    return false;                                                                                                     // 743
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
isAdminById = function (id) {                                                                                         // 747
  // Returns true if the user's id is the administrator.                                                              //
                                                                                                                      //
  var user = Meteor.users.findOne({ _id: id }, { fields: { isAdmin: 1 } });                                           // 750
  if (user && user.isAdmin) {                                                                                         // 751
    return true;                                                                                                      // 752
  } else {                                                                                                            //
    return false;                                                                                                     // 754
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
findAdminUserForToken = function (token) {                                                                            // 758
  if (!token.requirements) {                                                                                          // 759
    return;                                                                                                           // 760
  }                                                                                                                   //
                                                                                                                      //
  var requirements = token.requirements.filter(function (requirement) {                                               // 763
    return "userIsAdmin" in requirement;                                                                              // 764
  });                                                                                                                 //
                                                                                                                      //
  if (requirements.length > 1) {                                                                                      // 767
    return;                                                                                                           // 768
  }                                                                                                                   //
                                                                                                                      //
  if (requirements.length === 0) {                                                                                    // 771
    return;                                                                                                           // 772
  }                                                                                                                   //
                                                                                                                      //
  return requirements[0].userIsAdmin;                                                                                 // 775
};                                                                                                                    //
                                                                                                                      //
var wildcardHost = Meteor.settings["public"].wildcardHost.toLowerCase().split("*");                                   // 778
                                                                                                                      //
if (wildcardHost.length != 2) {                                                                                       // 780
  throw new Error("Wildcard host must contain exactly one asterisk.");                                                // 781
}                                                                                                                     //
                                                                                                                      //
matchWildcardHost = function (host) {                                                                                 // 784
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                          //
                                                                                                                      //
  // We remove everything after the first ":" character so that our                                                   //
  // comparison logic ignores port numbers.                                                                           //
  var prefix = wildcardHost[0];                                                                                       // 789
  var suffix = wildcardHost[1].split(":")[0];                                                                         // 790
  var hostSansPort = host.split(":")[0];                                                                              // 791
                                                                                                                      //
  if (hostSansPort.lastIndexOf(prefix, 0) >= 0 && hostSansPort.indexOf(suffix, -suffix.length) >= 0 && hostSansPort.length >= prefix.length + suffix.length) {
    var id = hostSansPort.slice(prefix.length, -suffix.length);                                                       // 796
    if (id.match(/^[-a-z0-9]*$/)) {                                                                                   // 797
      return id;                                                                                                      // 798
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  return null;                                                                                                        // 802
};                                                                                                                    //
                                                                                                                      //
makeWildcardHost = function (id) {                                                                                    // 805
  return wildcardHost[0] + id + wildcardHost[1];                                                                      // 806
};                                                                                                                    //
                                                                                                                      //
var isApiHostId = function (hostId) {                                                                                 // 809
  if (hostId) {                                                                                                       // 810
    var split = hostId.split("-");                                                                                    // 811
    if (split[0] === "api") return split[1] || "*";                                                                   // 812
  }                                                                                                                   //
                                                                                                                      //
  return false;                                                                                                       // 815
};                                                                                                                    //
                                                                                                                      //
var isTokenSpecificHostId = function (hostId) {                                                                       // 818
  return hostId.lastIndexOf("api-", 0) === 0;                                                                         // 819
};                                                                                                                    //
                                                                                                                      //
var apiHostIdHashForToken = undefined;                                                                                // 822
if (Meteor.isServer) {                                                                                                // 823
  (function () {                                                                                                      //
    var Crypto = Npm.require("crypto");                                                                               // 824
    apiHostIdHashForToken = function (token) {                                                                        // 825
      // Given an API token, compute the host ID that must be used when requesting this token.                        //
                                                                                                                      //
      // We add a leading 'x' to the hash so that knowing the hostname alone is not sufficient to                     //
      // find the corresponding API token in the ApiTokens table (whose _id values are also hashes                    //
      // of tokens). This doesn't technically add any security, but helps prove that we don't have                    //
      // any bugs which would allow someone who knows only the hostname to access the app API.                        //
      return Crypto.createHash("sha256").update("x" + token).digest("hex").slice(0, 32);                              // 832
    };                                                                                                                //
  })();                                                                                                               //
} else {                                                                                                              //
  apiHostIdHashForToken = function (token) {                                                                          // 835
    // Given an API token, compute the host ID that must be used when requesting this token.                          //
                                                                                                                      //
    // We add a leading 'x' to the hash so that knowing the hostname alone is not sufficient to                       //
    // find the corresponding API token in the ApiTokens table (whose _id values are also hashes                      //
    // of tokens). This doesn't technically add any security, but helps prove that we don't have                      //
    // any bugs which would allow someone who knows only the hostname to access the app API.                          //
    return SHA256("x" + token).slice(0, 32);                                                                          // 842
  };                                                                                                                  //
}                                                                                                                     //
                                                                                                                      //
var apiHostIdForToken = function (token) {                                                                            // 846
  return "api-" + apiHostIdHashForToken(token);                                                                       // 847
};                                                                                                                    //
                                                                                                                      //
var makeApiHost = function (token) {                                                                                  // 850
  return makeWildcardHost(apiHostIdForToken(token));                                                                  // 851
};                                                                                                                    //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 854
  (function () {                                                                                                      //
    var Url = Npm.require("url");                                                                                     // 855
    getWildcardOrigin = function () {                                                                                 // 856
      // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're                     //
      // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),                   //
      // an astrisk is only allowed as the first character and must be followed by a period. So we need               //
      // "*.example.com" instead -- which matches more than we actually want, but is the best we can                  //
      // really do. We also add the protocol to the front (again, that's what CSP wants).                             //
                                                                                                                      //
      // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.                //
      var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                        // 864
                                                                                                                      //
      var dotPos = wildcardHost[1].indexOf(".");                                                                      // 866
      if (dotPos < 0) {                                                                                               // 867
        return protocol + "//*";                                                                                      // 868
      } else {                                                                                                        //
        return protocol + "//*" + wildcardHost[1].slice(dotPos);                                                      // 870
      }                                                                                                               //
    };                                                                                                                //
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
allowDevAccounts = function () {                                                                                      // 875
  var setting = Settings.findOne({ _id: "devAccounts" });                                                             // 876
  if (setting) {                                                                                                      // 877
    return setting.value;                                                                                             // 878
  } else {                                                                                                            //
    return Meteor.settings && Meteor.settings["public"] && Meteor.settings["public"].allowDevAccounts;                // 880
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
sendReferralProgramNotification = function (userId) {                                                                 // 885
  Notifications.upsert({                                                                                              // 886
    userId: userId,                                                                                                   // 887
    referral: true                                                                                                    // 888
  }, {                                                                                                                //
    userId: userId,                                                                                                   // 890
    referral: true,                                                                                                   // 891
    timestamp: new Date(),                                                                                            // 892
    isUnread: true                                                                                                    // 893
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
roleAssignmentPattern = {                                                                                             // 897
  none: Match.Optional(null),                                                                                         // 898
  allAccess: Match.Optional(null),                                                                                    // 899
  roleId: Match.Optional(Match.Integer),                                                                              // 900
  addPermissions: Match.Optional([Boolean]),                                                                          // 901
  removePermissions: Match.Optional([Boolean])                                                                        // 902
};                                                                                                                    //
                                                                                                                      //
SandstormDb = function () {                                                                                           // 905
  this.collections = {                                                                                                // 906
    // Direct access to underlying collections. DEPRECATED.                                                           //
    //                                                                                                                //
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove                     //
    //   direct access to the collections.                                                                            //
                                                                                                                      //
    packages: Packages,                                                                                               // 912
    devPackages: DevPackages,                                                                                         // 913
    userActions: UserActions,                                                                                         // 914
    grains: Grains,                                                                                                   // 915
    contacts: Contacts,                                                                                               // 916
    sessions: Sessions,                                                                                               // 917
    signupKeys: SignupKeys,                                                                                           // 918
    activityStats: ActivityStats,                                                                                     // 919
    deleteStats: DeleteStats,                                                                                         // 920
    fileTokens: FileTokens,                                                                                           // 921
    apiTokens: ApiTokens,                                                                                             // 922
    apiHosts: ApiHosts,                                                                                               // 923
    notifications: Notifications,                                                                                     // 924
    statsTokens: StatsTokens,                                                                                         // 925
    misc: Misc,                                                                                                       // 926
    settings: Settings,                                                                                               // 927
    appIndex: AppIndex,                                                                                               // 928
    keybaseProfiles: KeybaseProfiles,                                                                                 // 929
    featureKey: FeatureKey                                                                                            // 930
                                                                                                                      //
  };                                                                                                                  //
};                                                                                                                    //
                                                                                                                      //
// TODO(cleanup): These methods should not be defined freestanding and should use collection                          //
//   objects created in SandstormDb's constructor rather than globals.                                                //
                                                                                                                      //
// Intentionally omitted:                                                                                             //
// - Migrations, since it's used only within this package.                                                            //
// - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.                       //
_.extend(SandstormDb.prototype, {                                                                                     // 941
  getUserQuota: getUserQuota,                                                                                         // 942
  isUserOverQuota: isUserOverQuota,                                                                                   // 943
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                             // 944
  isAdmin: isAdmin,                                                                                                   // 945
  isAdminById: isAdminById,                                                                                           // 946
  findAdminUserForToken: findAdminUserForToken,                                                                       // 947
  matchWildcardHost: matchWildcardHost,                                                                               // 948
  makeWildcardHost: makeWildcardHost,                                                                                 // 949
  isApiHostId: isApiHostId,                                                                                           // 950
  isTokenSpecificHostId: isTokenSpecificHostId,                                                                       // 951
  apiHostIdHashForToken: apiHostIdHashForToken,                                                                       // 952
  apiHostIdForToken: apiHostIdForToken,                                                                               // 953
  makeApiHost: makeApiHost,                                                                                           // 954
  allowDevAccounts: allowDevAccounts,                                                                                 // 955
  roleAssignmentPattern: roleAssignmentPattern,                                                                       // 956
                                                                                                                      //
  isDemoUser: function () {                                                                                           // 958
    // Returns true if this is a demo user.                                                                           //
                                                                                                                      //
    var user = Meteor.user();                                                                                         // 961
    if (user && user.expires) {                                                                                       // 962
      return true;                                                                                                    // 963
    } else {                                                                                                          //
      return false;                                                                                                   // 965
    }                                                                                                                 //
  },                                                                                                                  //
                                                                                                                      //
  isSignedUp: function () {                                                                                           // 969
    // Returns true if the user has presented an invite key.                                                          //
                                                                                                                      //
    var user = Meteor.user();                                                                                         // 972
                                                                                                                      //
    if (!user) return false; // not signed in                                                                         // 974
                                                                                                                      //
    if (!user.loginIdentities) return false; // not an account                                                        // 976
                                                                                                                      //
    if (user.expires) return false; // demo user.                                                                     // 978
                                                                                                                      //
    if (Meteor.settings["public"].allowUninvited) return true; // all accounts qualify                                // 980
                                                                                                                      //
    if (user.signupKey) return true; // user is invited                                                               // 982
                                                                                                                      //
    if (this.isUserInOrganization(user)) return true;                                                                 // 984
                                                                                                                      //
    return false;                                                                                                     // 986
  },                                                                                                                  //
                                                                                                                      //
  isSignedUpOrDemo: function () {                                                                                     // 989
    var user = Meteor.user();                                                                                         // 990
                                                                                                                      //
    if (!user) return false; // not signed in                                                                         // 992
                                                                                                                      //
    if (!user.loginIdentities) return false; // not an account                                                        // 994
                                                                                                                      //
    if (user.expires) return true; // demo user.                                                                      // 996
                                                                                                                      //
    if (Meteor.settings["public"].allowUninvited) return true; // all accounts qualify                                // 998
                                                                                                                      //
    if (user.signupKey) return true; // user is invited                                                               // 1000
                                                                                                                      //
    if (this.isUserInOrganization(user)) return true;                                                                 // 1002
                                                                                                                      //
    return false;                                                                                                     // 1004
  },                                                                                                                  //
                                                                                                                      //
  identityInOrganization: function (identityId) {                                                                     // 1007
    var identity = Meteor.users.findOne({ _id: identityId });                                                         // 1008
    if (!identity || !identity.services) {                                                                            // 1009
      return false;                                                                                                   // 1010
    }                                                                                                                 //
                                                                                                                      //
    var orgMembership = this.getOrganizationMembership();                                                             // 1013
    var googleEnabled = orgMembership && orgMembership.google && orgMembership.google.enabled;                        // 1014
    var googleDomain = orgMembership && orgMembership.google && orgMembership.google.domain;                          // 1015
    var emailEnabled = orgMembership && orgMembership.emailToken && orgMembership.emailToken.enabled;                 // 1016
    var emailDomain = orgMembership && orgMembership.emailToken && orgMembership.emailToken.domain;                   // 1017
    var ldapEnabled = orgMembership && orgMembership.ldap && orgMembership.ldap.enabled;                              // 1018
    var samlEnabled = orgMembership && orgMembership.saml && orgMembership.saml.enabled;                              // 1019
    if (emailEnabled && emailDomain && identity.services.email) {                                                     // 1020
      if (identity.services.email.email.toLowerCase().split("@").pop() === emailDomain) {                             // 1021
        return true;                                                                                                  // 1022
      }                                                                                                               //
    } else if (ldapEnabled && identity.services.ldap) {                                                               //
      return true;                                                                                                    // 1025
    } else if (samlEnabled && identity.services.saml) {                                                               //
      return true;                                                                                                    // 1027
    } else if (googleEnabled && googleDomain && identity.services.google && identity.services.google.hd) {            //
      if (identity.services.google.hd.toLowerCase() === googleDomain) {                                               // 1029
        return true;                                                                                                  // 1030
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return false;                                                                                                     // 1034
  },                                                                                                                  //
                                                                                                                      //
  isUserInOrganization: function (user) {                                                                             // 1037
    if (!this.isFeatureKeyValid()) {                                                                                  // 1038
      return false;                                                                                                   // 1039
    }                                                                                                                 //
                                                                                                                      //
    for (var i = 0; i < user.loginIdentities.length; i++) {                                                           // 1042
      if (this.identityInOrganization(user.loginIdentities[i].id)) {                                                  // 1043
        return true;                                                                                                  // 1044
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return false;                                                                                                     // 1048
  }                                                                                                                   //
});                                                                                                                   //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 1052
  (function () {                                                                                                      //
    SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                                      // 1053
                                                                                                                      //
    var Crypto = Npm.require("crypto");                                                                               // 1055
    SandstormDb.prototype.removeApiTokens = function (query) {                                                        // 1056
      // Remove all API tokens matching the query, making sure to clean up ApiHosts as well.                          //
                                                                                                                      //
      this.collections.apiTokens.find(query).forEach(function (token) {                                               // 1059
        // Clean up ApiHosts for webkey tokens.                                                                       //
        if (token.hasApiHost) {                                                                                       // 1061
          var hash2 = Crypto.createHash("sha256").update(token._id).digest("base64");                                 // 1062
          ApiHosts.remove({ hash2: hash2 });                                                                          // 1063
        }                                                                                                             //
      });                                                                                                             //
                                                                                                                      //
      this.collections.apiTokens.remove(query);                                                                       // 1067
    };                                                                                                                //
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
// =======================================================================================                            //
// Below this point are newly-written or refactored functions.                                                        //
                                                                                                                      //
_.extend(SandstormDb.prototype, {                                                                                     // 1074
  getUser: (function () {                                                                                             // 1075
    function getUser(userId) {                                                                                        // 1075
      check(userId, Match.OneOf(String, undefined, null));                                                            // 1076
      if (userId) {                                                                                                   // 1077
        return Meteor.users.findOne(userId);                                                                          // 1078
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return getUser;                                                                                                   //
  })(),                                                                                                               //
                                                                                                                      //
  getIdentity: (function () {                                                                                         // 1082
    function getIdentity(identityId) {                                                                                // 1082
      check(identityId, String);                                                                                      // 1083
      var identity = Meteor.users.findOne({ _id: identityId });                                                       // 1084
      if (identity) {                                                                                                 // 1085
        SandstormDb.fillInProfileDefaults(identity);                                                                  // 1086
        SandstormDb.fillInIntrinsicName(identity);                                                                    // 1087
        SandstormDb.fillInPictureUrl(identity);                                                                       // 1088
        return identity;                                                                                              // 1089
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return getIdentity;                                                                                               //
  })(),                                                                                                               //
                                                                                                                      //
  userHasIdentity: function (userId, identityId) {                                                                    // 1093
    check(userId, String);                                                                                            // 1094
    check(identityId, String);                                                                                        // 1095
                                                                                                                      //
    if (userId === identityId) return true;                                                                           // 1097
                                                                                                                      //
    var user = Meteor.users.findOne(userId);                                                                          // 1099
    return SandstormDb.getUserIdentityIds(user).indexOf(identityId) != -1;                                            // 1100
  },                                                                                                                  //
                                                                                                                      //
  userGrains: (function () {                                                                                          // 1103
    function userGrains(userId) {                                                                                     // 1103
      check(userId, Match.OneOf(String, undefined, null));                                                            // 1104
      return this.collections.grains.find({ userId: userId });                                                        // 1105
    }                                                                                                                 //
                                                                                                                      //
    return userGrains;                                                                                                //
  })(),                                                                                                               //
                                                                                                                      //
  currentUserGrains: (function () {                                                                                   // 1108
    function currentUserGrains() {                                                                                    // 1108
      return this.userGrains(Meteor.userId());                                                                        // 1109
    }                                                                                                                 //
                                                                                                                      //
    return currentUserGrains;                                                                                         //
  })(),                                                                                                               //
                                                                                                                      //
  getGrain: (function () {                                                                                            // 1112
    function getGrain(grainId) {                                                                                      // 1112
      check(grainId, String);                                                                                         // 1113
      return this.collections.grains.findOne(grainId);                                                                // 1114
    }                                                                                                                 //
                                                                                                                      //
    return getGrain;                                                                                                  //
  })(),                                                                                                               //
                                                                                                                      //
  userApiTokens: (function () {                                                                                       // 1117
    function userApiTokens(userId) {                                                                                  // 1117
      check(userId, Match.OneOf(String, undefined, null));                                                            // 1118
      var identityIds = SandstormDb.getUserIdentityIds(this.getUser(userId));                                         // 1119
      return this.collections.apiTokens.find({ "owner.user.identityId": { $in: identityIds } });                      // 1120
    }                                                                                                                 //
                                                                                                                      //
    return userApiTokens;                                                                                             //
  })(),                                                                                                               //
                                                                                                                      //
  currentUserApiTokens: (function () {                                                                                // 1123
    function currentUserApiTokens() {                                                                                 // 1123
      return this.userApiTokens(Meteor.userId());                                                                     // 1124
    }                                                                                                                 //
                                                                                                                      //
    return currentUserApiTokens;                                                                                      //
  })(),                                                                                                               //
                                                                                                                      //
  userActions: (function () {                                                                                         // 1127
    function userActions(user) {                                                                                      // 1127
      return this.collections.userActions.find({ userId: user });                                                     // 1128
    }                                                                                                                 //
                                                                                                                      //
    return userActions;                                                                                               //
  })(),                                                                                                               //
                                                                                                                      //
  currentUserActions: (function () {                                                                                  // 1131
    function currentUserActions() {                                                                                   // 1131
      return this.userActions(Meteor.userId());                                                                       // 1132
    }                                                                                                                 //
                                                                                                                      //
    return currentUserActions;                                                                                        //
  })(),                                                                                                               //
                                                                                                                      //
  iconSrcForPackage: (function () {                                                                                   // 1135
    function iconSrcForPackage(pkg, usage) {                                                                          // 1135
      return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                                // 1136
    }                                                                                                                 //
                                                                                                                      //
    return iconSrcForPackage;                                                                                         //
  })(),                                                                                                               //
                                                                                                                      //
  getDenormalizedGrainInfo: (function () {                                                                            // 1139
    function getDenormalizedGrainInfo(grainId) {                                                                      // 1139
      var grain = this.getGrain(grainId);                                                                             // 1140
      var pkg = this.collections.packages.findOne(grain.packageId);                                                   // 1141
                                                                                                                      //
      if (!pkg) {                                                                                                     // 1143
        pkg = this.collections.devPackages.findOne(grain.packageId);                                                  // 1144
      }                                                                                                               //
                                                                                                                      //
      var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };                             // 1147
      var grainInfo = { appTitle: appTitle };                                                                         // 1148
                                                                                                                      //
      if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {                              // 1150
        var icons = pkg.manifest.metadata.icons;                                                                      // 1151
        grainInfo.icon = icons.grain || icons.appGrid;                                                                // 1152
      }                                                                                                               //
                                                                                                                      //
      // Only provide an app ID if we have no icon asset to provide and need to offer an identicon.                   //
      if (!grainInfo.icon && pkg) {                                                                                   // 1156
        grainInfo.appId = pkg.appId;                                                                                  // 1157
      }                                                                                                               //
                                                                                                                      //
      return grainInfo;                                                                                               // 1160
    }                                                                                                                 //
                                                                                                                      //
    return getDenormalizedGrainInfo;                                                                                  //
  })(),                                                                                                               //
                                                                                                                      //
  getPlan: function (id) {                                                                                            // 1163
    check(id, String);                                                                                                // 1164
    var plan = Plans.findOne(id);                                                                                     // 1165
    if (!plan) {                                                                                                      // 1166
      throw new Error("no such plan: " + id);                                                                         // 1167
    }                                                                                                                 //
                                                                                                                      //
    return plan;                                                                                                      // 1170
  },                                                                                                                  //
                                                                                                                      //
  listPlans: function () {                                                                                            // 1173
    return Plans.find({}, { sort: { price: 1 } });                                                                    // 1174
  },                                                                                                                  //
                                                                                                                      //
  getMyPlan: function () {                                                                                            // 1177
    var user = Meteor.user();                                                                                         // 1178
    return user && Plans.findOne(user.plan || "free");                                                                // 1179
  },                                                                                                                  //
                                                                                                                      //
  getMyReferralBonus: function (user) {                                                                               // 1182
    // This function is called from the server and from the client, similar to getMyPlan().                           //
    //                                                                                                                //
    // The parameter may be omitted in which case the current user is assumed.                                        //
                                                                                                                      //
    return calculateReferralBonus(user || Meteor.user());                                                             // 1187
  },                                                                                                                  //
                                                                                                                      //
  getMyUsage: function (user) {                                                                                       // 1190
    user = user || Meteor.user();                                                                                     // 1191
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                              // 1192
      if (Meteor.isClient) {                                                                                          // 1193
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly                       //
        // not reactive.                                                                                              //
        return user.pseudoUsage;                                                                                      // 1196
      } else {                                                                                                        //
        return {                                                                                                      // 1198
          grains: Grains.find({ userId: user._id }).count(),                                                          // 1199
          storage: user.storageUsage || 0,                                                                            // 1200
          compute: 0 };                                                                                               // 1201
      }                                                                                                               //
    } else // not tracked yet                                                                                         //
      {                                                                                                               // 1204
        return { grains: 0, storage: 0, compute: 0 };                                                                 // 1205
      }                                                                                                               //
  },                                                                                                                  //
                                                                                                                      //
  isUninvitedFreeUser: function () {                                                                                  // 1209
    if (!Meteor.settings["public"].allowUninvited) return false;                                                      // 1210
                                                                                                                      //
    var user = Meteor.user();                                                                                         // 1212
    return user && !user.expires && (!user.plan || user.plan === "free");                                             // 1213
  },                                                                                                                  //
                                                                                                                      //
  getSetting: function (name) {                                                                                       // 1216
    var setting = Settings.findOne(name);                                                                             // 1217
    return setting && setting.value;                                                                                  // 1218
  },                                                                                                                  //
                                                                                                                      //
  addUserActions: function (packageId) {                                                                              // 1221
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.                      //
    var pack = Packages.findOne(packageId);                                                                           // 1223
    if (pack) {                                                                                                       // 1224
      // Remove old versions.                                                                                         //
      UserActions.find({ userId: Meteor.userId(), appId: pack.appId }).forEach(function (action) {                    // 1226
        UserActions.remove(action._id);                                                                               // 1228
      });                                                                                                             //
                                                                                                                      //
      // Install new.                                                                                                 //
      var actions = pack.manifest.actions;                                                                            // 1232
      for (var i in babelHelpers.sanitizeForInObject(actions)) {                                                      // 1233
        var action = actions[i];                                                                                      // 1234
        if ("none" in action.input) {                                                                                 // 1235
          var userAction = {                                                                                          // 1236
            userId: Meteor.userId(),                                                                                  // 1237
            packageId: pack._id,                                                                                      // 1238
            appId: pack.appId,                                                                                        // 1239
            appTitle: pack.manifest.appTitle,                                                                         // 1240
            appMarketingVersion: pack.manifest.appMarketingVersion,                                                   // 1241
            appVersion: pack.manifest.appVersion,                                                                     // 1242
            title: action.title,                                                                                      // 1243
            nounPhrase: action.nounPhrase,                                                                            // 1244
            command: action.command                                                                                   // 1245
          };                                                                                                          //
          UserActions.insert(userAction);                                                                             // 1247
        } else {}                                                                                                     //
        // TODO(someday):  Implement actions with capability inputs.                                                  //
        //jscs:ignore disallowEmptyBlocks                                                                             //
      }                                                                                                               // 1250
                                                                                                                      //
      Meteor.call("deleteUnusedPackages", pack.appId);                                                                // 1253
    }                                                                                                                 //
  },                                                                                                                  //
                                                                                                                      //
  sendAdminNotification: function (message, link) {                                                                   // 1257
    Meteor.users.find({ isAdmin: true }, { fields: { _id: 1 } }).forEach(function (user) {                            // 1258
      Notifications.insert({                                                                                          // 1259
        admin: {                                                                                                      // 1260
          action: link,                                                                                               // 1261
          type: "reportStats"                                                                                         // 1262
        },                                                                                                            //
        userId: user._id,                                                                                             // 1264
        text: { defaultText: message },                                                                               // 1265
        timestamp: new Date(),                                                                                        // 1266
        isUnread: true                                                                                                // 1267
      });                                                                                                             //
    });                                                                                                               //
  },                                                                                                                  //
                                                                                                                      //
  getKeybaseProfile: function (keyFingerprint) {                                                                      // 1272
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                            // 1273
  },                                                                                                                  //
                                                                                                                      //
  getServerTitle: function () {                                                                                       // 1276
    var setting = Settings.findOne({ _id: "serverTitle" });                                                           // 1277
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1278
  },                                                                                                                  //
                                                                                                                      //
  getSmtpConfig: function () {                                                                                        // 1281
    var setting = Settings.findOne({ _id: "smtpConfig" });                                                            // 1282
    return setting ? setting.value : undefined; // undefined if subscription is not ready.                            // 1283
  },                                                                                                                  //
                                                                                                                      //
  getReturnAddress: function () {                                                                                     // 1286
    var config = this.getSmtpConfig();                                                                                // 1287
    return config && config.returnAddress || ""; // empty if subscription is not ready.                               // 1288
  },                                                                                                                  //
                                                                                                                      //
  getReturnAddressWithDisplayName: function (identityId) {                                                            // 1291
    check(identityId, String);                                                                                        // 1292
    var identity = this.getIdentity(identityId);                                                                      // 1293
    var displayName = identity.profile.name + " (via " + this.getServerTitle() + ")";                                 // 1294
                                                                                                                      //
    // First remove any instances of characters that cause trouble for SimpleSmtp. Ideally,                           //
    // we could escape such characters with a backslash, but that does not seem to help here.                         //
    var sanitized = displayName.replace(/"|<|>|\\|\r/g, "");                                                          // 1298
                                                                                                                      //
    return "\"" + sanitized + "\" <" + this.getReturnAddress() + ">";                                                 // 1300
  },                                                                                                                  //
                                                                                                                      //
  getPrimaryEmail: function (accountId, identityId) {                                                                 // 1303
    check(accountId, String);                                                                                         // 1304
    check(identityId, String);                                                                                        // 1305
                                                                                                                      //
    var identity = this.getIdentity(identityId);                                                                      // 1307
    var senderEmails = SandstormDb.getVerifiedEmails(identity);                                                       // 1308
    var senderPrimaryEmail = _.findWhere(senderEmails, { primary: true });                                            // 1309
    var accountPrimaryEmailAddress = this.getUser(accountId).primaryEmail;                                            // 1310
    if (_.findWhere(senderEmails, { email: accountPrimaryEmailAddress })) {                                           // 1311
      return accountPrimaryEmailAddress;                                                                              // 1312
    } else if (senderPrimaryEmail) {                                                                                  //
      return senderPrimaryEmail.email;                                                                                // 1314
    } else {                                                                                                          //
      return null;                                                                                                    // 1316
    }                                                                                                                 //
  },                                                                                                                  //
                                                                                                                      //
  isFeatureKeyValid: function () {                                                                                    // 1320
    var featureKey = this.currentFeatureKey();                                                                        // 1321
    return !!featureKey;                                                                                              // 1322
  },                                                                                                                  //
                                                                                                                      //
  isFeatureKeyValidAndNotExpired: function () {                                                                       // 1325
    var featureKey = this.currentFeatureKey();                                                                        // 1326
    return featureKey && parseInt(featureKey.expires) > Date.now() / 1000;                                            // 1327
  },                                                                                                                  //
                                                                                                                      //
  getLdapUrl: function () {                                                                                           // 1330
    var setting = Settings.findOne({ _id: "ldapUrl" });                                                               // 1331
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1332
  },                                                                                                                  //
                                                                                                                      //
  getLdapBase: function () {                                                                                          // 1335
    var setting = Settings.findOne({ _id: "ldapBase" });                                                              // 1336
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1337
  },                                                                                                                  //
                                                                                                                      //
  getLdapDnPattern: function () {                                                                                     // 1340
    var setting = Settings.findOne({ _id: "ldapDnPattern" });                                                         // 1341
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1342
  },                                                                                                                  //
                                                                                                                      //
  getLdapSearchUsername: function () {                                                                                // 1345
    var setting = Settings.findOne({ _id: "ldapSearchUsername" });                                                    // 1346
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1347
  },                                                                                                                  //
                                                                                                                      //
  getLdapNameField: function () {                                                                                     // 1350
    var setting = Settings.findOne({ _id: "ldapNameField" });                                                         // 1351
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1352
  },                                                                                                                  //
                                                                                                                      //
  getLdapEmailField: function () {                                                                                    // 1355
    var setting = Settings.findOne({ _id: "ldapEmailField" });                                                        // 1356
    return setting ? setting.value : "mail";                                                                          // 1357
    // default to "mail". This setting was added later, and so could potentially be unset.                            //
  },                                                                                                                  //
                                                                                                                      //
  getLdapExplicitDnSelected: function () {                                                                            // 1361
    var setting = Settings.findOne({ _id: "ldapExplicitDnSelected" });                                                // 1362
    return setting && setting.value;                                                                                  // 1363
  },                                                                                                                  //
                                                                                                                      //
  getLdapFilter: function () {                                                                                        // 1366
    var setting = Settings.findOne({ _id: "ldapFilter" });                                                            // 1367
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1368
  },                                                                                                                  //
                                                                                                                      //
  getLdapSearchBindDn: function () {                                                                                  // 1371
    var setting = Settings.findOne({ _id: "ldapSearchBindDn" });                                                      // 1372
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1373
  },                                                                                                                  //
                                                                                                                      //
  getLdapSearchBindPassword: function () {                                                                            // 1376
    var setting = Settings.findOne({ _id: "ldapSearchBindPassword" });                                                // 1377
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1378
  },                                                                                                                  //
                                                                                                                      //
  getOrganizationMembership: function () {                                                                            // 1381
    var setting = Settings.findOne({ _id: "organizationMembership" });                                                // 1382
    return setting && setting.value;                                                                                  // 1383
  },                                                                                                                  //
                                                                                                                      //
  getOrganizationEmailEnabled: function () {                                                                          // 1386
    var membership = this.getOrganizationMembership();                                                                // 1387
    return membership && membership.emailToken && membership.emailToken.enabled;                                      // 1388
  },                                                                                                                  //
                                                                                                                      //
  getOrganizationEmailDomain: function () {                                                                           // 1391
    var membership = this.getOrganizationMembership();                                                                // 1392
    return membership && membership.emailToken && membership.emailToken.domain;                                       // 1393
  },                                                                                                                  //
                                                                                                                      //
  getOrganizationGoogleEnabled: function () {                                                                         // 1396
    var membership = this.getOrganizationMembership();                                                                // 1397
    return membership && membership.google && membership.google.enabled;                                              // 1398
  },                                                                                                                  //
                                                                                                                      //
  getOrganizationGoogleDomain: function () {                                                                          // 1401
    var membership = this.getOrganizationMembership();                                                                // 1402
    return membership && membership.google && membership.google.domain;                                               // 1403
  },                                                                                                                  //
                                                                                                                      //
  getOrganizationLdapEnabled: function () {                                                                           // 1406
    var membership = this.getOrganizationMembership();                                                                // 1407
    return membership && membership.ldap && membership.ldap.enabled;                                                  // 1408
  },                                                                                                                  //
                                                                                                                      //
  getOrganizationSamlEnabled: function () {                                                                           // 1411
    var membership = this.getOrganizationMembership();                                                                // 1412
    return membership && membership.ldap && membership.saml.enabled;                                                  // 1413
  },                                                                                                                  //
                                                                                                                      //
  getSamlEntryPoint: function () {                                                                                    // 1416
    var setting = Settings.findOne({ _id: "samlEntryPoint" });                                                        // 1417
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1418
  },                                                                                                                  //
                                                                                                                      //
  getSamlPublicCert: function () {                                                                                    // 1421
    var setting = Settings.findOne({ _id: "samlPublicCert" });                                                        // 1422
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1423
  }                                                                                                                   //
});                                                                                                                   //
                                                                                                                      //
SandstormDb.escapeMongoKey = function (key) {                                                                         // 1427
  // This incredibly poor mechanism for escaping Mongo keys is recommended by the Mongo docs here:                    //
  //   https://docs.mongodb.org/manual/faq/developers/#dollar-sign-operator-escaping                                  //
  // and seems to be a de facto standard, for example:                                                                //
  //   https://www.npmjs.com/package/mongo-key-escape                                                                 //
  return key.replace(".", "．").replace("$", "＄");                                                                     // 1432
};                                                                                                                    //
                                                                                                                      //
var appNameFromPackage = function (packageObj) {                                                                      // 1435
  // This function takes a Package object from Mongo and returns an                                                   //
  // app title.                                                                                                       //
  var manifest = packageObj.manifest;                                                                                 // 1438
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                              // 1439
  var action = manifest.actions[0];                                                                                   // 1440
  appName = manifest.appTitle && manifest.appTitle.defaultText || appNameFromActionName(action.title.defaultText);    // 1441
  return appName;                                                                                                     // 1443
};                                                                                                                    //
                                                                                                                      //
var appNameFromActionName = function (name) {                                                                         // 1446
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                              //
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                                 //
  //   derive the app title from the action title.                                                                    //
  // TODO(cleanup): Get rid of this once apps have real titles.                                                       //
  if (!name) {                                                                                                        // 1451
    return "(unnamed)";                                                                                               // 1452
  }                                                                                                                   //
                                                                                                                      //
  if (name.lastIndexOf("New ", 0) === 0) {                                                                            // 1455
    name = name.slice(4);                                                                                             // 1456
  }                                                                                                                   //
                                                                                                                      //
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                                      // 1459
    name = "Hacker CMS";                                                                                              // 1460
  } else {                                                                                                            //
    var space = name.indexOf(" ");                                                                                    // 1462
    if (space > 0) {                                                                                                  // 1463
      name = name.slice(0, space);                                                                                    // 1464
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  return name;                                                                                                        // 1468
};                                                                                                                    //
                                                                                                                      //
var appShortDescriptionFromPackage = function (pkg) {                                                                 // 1471
  return pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.shortDescription && pkg.manifest.metadata.shortDescription.defaultText;
};                                                                                                                    //
                                                                                                                      //
var nounPhraseForActionAndAppTitle = function (action, appTitle) {                                                    // 1477
  // A hack to deal with legacy apps not including fields in their manifests.                                         //
  // I look forward to the day I can remove most of this code.                                                        //
  // Attempt to figure out the appropriate noun that this action will create.                                         //
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                           //
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                                        // 1482
  // Otherwise, try to guess one from the structure of the action title field                                         //
  if (action.title && action.title.defaultText) {                                                                     // 1484
    var text = action.title.defaultText;                                                                              // 1485
    // Strip a leading "New "                                                                                         //
    if (text.lastIndexOf("New ", 0) === 0) {                                                                          // 1487
      var candidate = text.slice(4);                                                                                  // 1488
      // Strip a leading appname too, if provided                                                                     //
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                                 // 1490
        var newCandidate = candidate.slice(appTitle.length);                                                          // 1491
        // Unless that leaves you with no noun, in which case, use "grain"                                            //
        if (newCandidate.length > 0) {                                                                                // 1493
          return newCandidate.toLowerCase();                                                                          // 1494
        } else {                                                                                                      //
          return "grain";                                                                                             // 1496
        }                                                                                                             //
      }                                                                                                               //
                                                                                                                      //
      return candidate.toLowerCase();                                                                                 // 1500
    }                                                                                                                 //
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates                       //
    // the package soon.                                                                                              //
    return text;                                                                                                      // 1504
  } else {                                                                                                            //
    return "grain";                                                                                                   // 1506
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
// Static methods on SandstormDb that don't need an instance.                                                         //
// Largely things that deal with backwards-compatibility.                                                             //
_.extend(SandstormDb, {                                                                                               // 1512
  appNameFromActionName: appNameFromActionName,                                                                       // 1513
  appNameFromPackage: appNameFromPackage,                                                                             // 1514
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                                     // 1515
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle                                                      // 1516
});                                                                                                                   //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 1519
  (function () {                                                                                                      //
    var Crypto = Npm.require("crypto");                                                                               // 1520
    var ContentType = Npm.require("content-type");                                                                    // 1521
    var Zlib = Npm.require("zlib");                                                                                   // 1522
                                                                                                                      //
    var replicaNumber = Meteor.settings.replicaNumber || 0;                                                           // 1524
                                                                                                                      //
    var computeStagger = function (n) {                                                                               // 1526
      // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                       //
      // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                        //
      // The sequence looks like:                                                                                     //
      //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                            //
      //                                                                                                              //
      // We use this to determine how we'll stagger periodic events performed by this replica.                        //
      // Notice that this allows us to compute a stagger which is independent of the number of                        //
      // front-end replicas present; we can add more replicas to the end without affecting how the                    //
      // earlier ones schedule their events.                                                                          //
      var denom = 1;                                                                                                  // 1536
      while (denom <= n) denom <<= 1;                                                                                 // 1537
      var num = n * 2 - denom + 1;                                                                                    // 1538
      return num / denom;                                                                                             // 1539
    };                                                                                                                //
                                                                                                                      //
    var stagger = computeStagger(replicaNumber);                                                                      // 1542
                                                                                                                      //
    SandstormDb.periodicCleanup = function (intervalMs, callback) {                                                   // 1544
      // Register a database cleanup function than should run periodically, roughly once every                        //
      // interval of the given length.                                                                                //
      //                                                                                                              //
      // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be                     //
      // scaled appropriately on the assumption that more data is being generated demanding more                      //
      // frequent cleanups.                                                                                           //
                                                                                                                      //
      check(intervalMs, Number);                                                                                      // 1552
      check(callback, Function);                                                                                      // 1553
                                                                                                                      //
      if (intervalMs < 120000) {                                                                                      // 1555
        throw new Error("less than 2-minute cleanup interval seems too fast; " + "are you using the right units?");   // 1556
      }                                                                                                               //
                                                                                                                      //
      // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that                     //
      // the schedule is independent of the exact startup time.                                                       //
      var first = intervalMs - Date.now() % intervalMs;                                                               // 1562
                                                                                                                      //
      // Stagger cleanups across replicas so that we don't have all replicas trying to clean the                      //
      // same data at the same time.                                                                                  //
      first += Math.floor(intervalMs * computeStagger(replicaNumber));                                                // 1566
                                                                                                                      //
      // If the stagger put us more than an interval away from now, back up.                                          //
      if (first > intervalMs) first -= intervalMs;                                                                    // 1569
                                                                                                                      //
      Meteor.setTimeout(function () {                                                                                 // 1571
        callback();                                                                                                   // 1572
        Meteor.setInterval(callback, intervalMs);                                                                     // 1573
      }, first);                                                                                                      //
    };                                                                                                                //
                                                                                                                      //
    // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                         //
    var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                                 // 1578
                                                                                                                      //
    var BufferSmallerThan = function (limit) {                                                                        // 1580
      return Match.Where(function (buf) {                                                                             // 1581
        check(buf, Buffer);                                                                                           // 1582
        return buf.length < limit;                                                                                    // 1583
      });                                                                                                             //
    };                                                                                                                //
                                                                                                                      //
    var DatabaseId = Match.Where(function (s) {                                                                       // 1587
      check(s, String);                                                                                               // 1588
      return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                            // 1589
    });                                                                                                               //
                                                                                                                      //
    addStaticAsset = function (metadata, content) {                                                                   // 1592
      // Add a new static asset to the database. If `content` is a string rather than a buffer, it                    //
      // will be automatically gzipped before storage; do not specify metadata.encoding in this case.                 //
                                                                                                                      //
      if (typeof content === "string" && !metadata.encoding) {                                                        // 1596
        content = gzipSync(new Buffer(content, "utf8"));                                                              // 1597
        metadata.encoding = "gzip";                                                                                   // 1598
      }                                                                                                               //
                                                                                                                      //
      check(metadata, {                                                                                               // 1601
        mimeType: String,                                                                                             // 1602
        encoding: Match.Optional("gzip")                                                                              // 1603
      });                                                                                                             //
      check(content, BufferSmallerThan(1 << 20));                                                                     // 1605
                                                                                                                      //
      // Validate content type.                                                                                       //
      metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                                   // 1608
                                                                                                                      //
      var hasher = Crypto.createHash("sha256");                                                                       // 1610
      hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                                     // 1611
      hasher.update(content);                                                                                         // 1612
      var hash = hasher.digest("base64");                                                                             // 1613
                                                                                                                      //
      var existing = StaticAssets.findAndModify({                                                                     // 1615
        query: { hash: hash, refcount: { $gte: 1 } },                                                                 // 1616
        update: { $inc: { refcount: 1 } },                                                                            // 1617
        fields: { _id: 1, refcount: 1 }                                                                               // 1618
      });                                                                                                             //
      if (existing) {                                                                                                 // 1620
        return existing._id;                                                                                          // 1621
      }                                                                                                               //
                                                                                                                      //
      return StaticAssets.insert(_.extend({                                                                           // 1624
        hash: hash,                                                                                                   // 1625
        content: content,                                                                                             // 1626
        refcount: 1                                                                                                   // 1627
      }, metadata));                                                                                                  //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                            // 1631
                                                                                                                      //
    SandstormDb.prototype.refStaticAsset = function (id) {                                                            // 1633
      // Increment the refcount on an existing static asset.                                                          //
      //                                                                                                              //
      // You must call this BEFORE adding the new reference to the DB, in case of failure between                     //
      // the two calls. (This way, the failure case is a storage leak, which is probably not a big                    //
      // deal and can be fixed by GC, rather than a mysteriously missing asset.)                                      //
                                                                                                                      //
      check(id, String);                                                                                              // 1640
                                                                                                                      //
      var existing = StaticAssets.findAndModify({                                                                     // 1642
        query: { hash: hash },                                                                                        // 1643
        update: { $inc: { refcount: 1 } },                                                                            // 1644
        fields: { _id: 1, refcount: 1 }                                                                               // 1645
      });                                                                                                             //
      if (!existing) {                                                                                                // 1647
        throw new Error("refStaticAsset() called on asset that doesn't exist");                                       // 1648
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.unrefStaticAsset = function (id) {                                                          // 1652
      // Decrement refcount on a static asset and delete if it has reached zero.                                      //
      //                                                                                                              //
      // You must call this AFTER removing the reference from the DB, in case of failure between                      //
      // the two calls. (This way, the failure case is a storage leak, which is probably not a big                    //
      // deal and can be fixed by GC, rather than a mysteriously missing asset.)                                      //
                                                                                                                      //
      check(id, String);                                                                                              // 1659
                                                                                                                      //
      var existing = StaticAssets.findAndModify({                                                                     // 1661
        query: { _id: id },                                                                                           // 1662
        update: { $inc: { refcount: -1 } },                                                                           // 1663
        fields: { _id: 1, refcount: 1 },                                                                              // 1664
        "new": true                                                                                                   // 1665
      });                                                                                                             //
      if (!existing) {                                                                                                // 1667
        console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);                      // 1668
      } else if (existing.refcount <= 0) {                                                                            //
        StaticAssets.remove({ _id: existing._id });                                                                   // 1670
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.getStaticAsset = function (id) {                                                            // 1674
      // Get a static asset's mimeType, encoding, and raw content.                                                    //
                                                                                                                      //
      check(id, String);                                                                                              // 1677
                                                                                                                      //
      var asset = StaticAssets.findOne(id, { fields: { _id: 0, mimeType: 1, encoding: 1, content: 1 } });             // 1679
      if (asset) {                                                                                                    // 1680
        // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy                     //
        //   here.                                                                                                    //
        asset.content = new Buffer(asset.content);                                                                    // 1683
      }                                                                                                               //
                                                                                                                      //
      return asset;                                                                                                   // 1686
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.newAssetUpload = function (purpose) {                                                       // 1689
      check(purpose, { profilePicture: { userId: DatabaseId, identityId: DatabaseId } });                             // 1690
                                                                                                                      //
      return AssetUploadTokens.insert({                                                                               // 1692
        purpose: purpose,                                                                                             // 1693
        expires: new Date(Date.now() + 300000) });                                                                    // 1694
    };                                                                                                                //
                                                                                                                      //
    // in 5 minutes                                                                                                   //
    SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                        // 1698
      // Indicates that the given asset upload has completed. It will be removed and its purpose                      //
      // returned. If no matching upload exists, returns undefined.                                                   //
                                                                                                                      //
      check(id, String);                                                                                              // 1702
                                                                                                                      //
      var upload = AssetUploadTokens.findAndModify({                                                                  // 1704
        query: { _id: id },                                                                                           // 1705
        remove: true                                                                                                  // 1706
      });                                                                                                             //
                                                                                                                      //
      if (upload.expires.valueOf() < Date.now()) {                                                                    // 1709
        return undefined; // already expired                                                                          // 1710
      } else {                                                                                                        //
          return upload.purpose;                                                                                      // 1712
        }                                                                                                             //
    };                                                                                                                //
                                                                                                                      //
    function cleanupExpiredAssetUploads() {                                                                           // 1716
      AssetUploadTokens.remove({ expires: { $lt: Date.now() } });                                                     // 1717
    }                                                                                                                 //
                                                                                                                      //
    // Cleanup tokens every hour.                                                                                     //
    SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                                 // 1721
                                                                                                                      //
    var packageCache = {};                                                                                            // 1723
    // Package info is immutable. Let's cache to save on mongo queries.                                               //
                                                                                                                      //
    SandstormDb.prototype.getPackage = function (packageId) {                                                         // 1726
      // Get the given package record. Since package info is immutable, cache the data in the server                  //
      // to reduce mongo query overhead, since it turns out we have to fetch specific packages a                      //
      // lot.                                                                                                         //
                                                                                                                      //
      if (packageId in packageCache) {                                                                                // 1731
        return packageCache[packageId];                                                                               // 1732
      }                                                                                                               //
                                                                                                                      //
      var pkg = Packages.findOne(packageId);                                                                          // 1735
      if (pkg && pkg.status === "ready") {                                                                            // 1736
        packageCache[packageId] = pkg;                                                                                // 1737
      }                                                                                                               //
                                                                                                                      //
      return pkg;                                                                                                     // 1740
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name, versionNumber, marketingVersion) {
      var _this = this;                                                                                               // 1745
      var actions = _this.collections.userActions.find({ appId: appId, appVersion: { $lt: versionNumber } }, { fields: { userId: 1 } });
      actions.forEach(function (action) {                                                                             // 1748
        var userId = action.userId;                                                                                   // 1749
        var updater = {                                                                                               // 1750
          userId: userId,                                                                                             // 1751
          timestamp: new Date(),                                                                                      // 1752
          isUnread: true                                                                                              // 1753
        };                                                                                                            //
                                                                                                                      //
        // Set only the appId that we care about. Use mongo's dot notation to specify only a single                   //
        // field inside of an object to update                                                                        //
        updater["appUpdates." + appId] = {                                                                            // 1758
          marketingVersion: marketingVersion,                                                                         // 1759
          packageId: packageId,                                                                                       // 1760
          name: name,                                                                                                 // 1761
          version: versionNumber                                                                                      // 1762
        };                                                                                                            //
        _this.collections.notifications.upsert({ userId: userId }, { $set: updater });                                // 1764
      });                                                                                                             //
                                                                                                                      //
      _this.collections.appIndex.update({ _id: appId }, { $set: { hasSentNotifications: true } });                    // 1767
                                                                                                                      //
      // In the case where we replaced a previous notification and that was the only reference to the                 //
      // package, we need to clean it up                                                                              //
      Meteor.call("deleteUnusedPackages", appId);                                                                     // 1771
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.sendReferralProgramNotification = sendReferralProgramNotification;                          // 1774
                                                                                                                      //
    SandstormDb.prototype.upgradeGrains = function (appId, version, packageId, backend) {                             // 1776
      check(appId, String);                                                                                           // 1777
      check(version, Match.Integer);                                                                                  // 1778
      check(packageId, String);                                                                                       // 1779
                                                                                                                      //
      var selector = {                                                                                                // 1781
        userId: Meteor.userId(),                                                                                      // 1782
        appId: appId,                                                                                                 // 1783
        appVersion: { $lte: version },                                                                                // 1784
        packageId: { $ne: packageId }                                                                                 // 1785
      };                                                                                                              //
                                                                                                                      //
      if (!this.isSimulation) {                                                                                       // 1788
        Grains.find(selector).forEach(function (grain) {                                                              // 1789
          backend.shutdownGrain(grain._id, grain.userId);                                                             // 1790
        });                                                                                                           //
      }                                                                                                               //
                                                                                                                      //
      Grains.update(selector, { $set: { appVersion: version, packageId: packageId } }, { multi: true });              // 1794
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {                      // 1797
      // Mark package for possible installation.                                                                      //
                                                                                                                      //
      var fields = {                                                                                                  // 1800
        status: "download",                                                                                           // 1801
        progress: 0,                                                                                                  // 1802
        url: url,                                                                                                     // 1803
        isAutoUpdated: !!isAutoUpdated                                                                                // 1804
      };                                                                                                              //
                                                                                                                      //
      if (retryFailed) {                                                                                              // 1807
        Packages.update({ _id: packageId, status: "failed" }, { $set: fields });                                      // 1808
      } else {                                                                                                        //
        try {                                                                                                         // 1810
          fields._id = packageId;                                                                                     // 1811
          Packages.insert(fields);                                                                                    // 1812
        } catch (err) {                                                                                               //
          console.error("Simultaneous startInstall()s?", err.stack);                                                  // 1814
        }                                                                                                             //
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                                 // 1819
      check(keyFingerprint, String);                                                                                  // 1820
      return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                                // 1821
    });                                                                                                               //
                                                                                                                      //
    SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                                     // 1824
      // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.                  //
                                                                                                                      //
      check(keyFingerprint, ValidKeyFingerprint);                                                                     // 1827
                                                                                                                      //
      console.log("fetching keybase", keyFingerprint);                                                                // 1829
                                                                                                                      //
      HTTP.get("https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint + "&fields=basics,profile,proofs_summary", {
        timeout: 5000                                                                                                 // 1834
      }, function (err, keybaseResponse) {                                                                            //
        if (err) {                                                                                                    // 1836
          console.log("keybase lookup error:", err.stack);                                                            // 1837
          return;                                                                                                     // 1838
        }                                                                                                             //
                                                                                                                      //
        if (!keybaseResponse.data) {                                                                                  // 1841
          console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                               // 1842
          return;                                                                                                     // 1843
        }                                                                                                             //
                                                                                                                      //
        var profile = (keybaseResponse.data.them || [])[0];                                                           // 1846
                                                                                                                      //
        if (profile) {                                                                                                // 1848
          // jscs:disable requireCamelCaseOrUpperCaseIdentifiers                                                      //
          var record = {                                                                                              // 1850
            displayName: (profile.profile || {}).full_name,                                                           // 1851
            handle: (profile.basics || {}).username,                                                                  // 1852
            proofs: (profile.proofs_summary || {}).all || []                                                          // 1853
          };                                                                                                          //
          // jscs:enable requireCamelCaseOrUpperCaseIdentifiers                                                       //
                                                                                                                      //
          record.proofs.forEach(function (proof) {                                                                    // 1857
            // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that                   //
            // this would filter.)                                                                                    //
            for (var field in babelHelpers.sanitizeForInObject(proof)) {                                              // 1860
              // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly                   //
              // because I'm too lazy to check the field names recursively (and Keybase doesn't                       //
              // return any objects anyway).                                                                          //
              if (field.match(/[.$]/) || typeof proof[field] === "object") {                                          // 1864
                delete proof[field];                                                                                  // 1865
              }                                                                                                       //
            }                                                                                                         //
                                                                                                                      //
            // Indicate not verified.                                                                                 //
            // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the                   //
            //   Keybase node-based CLI.                                                                              //
            proof.status = "unverified";                                                                              // 1872
          });                                                                                                         //
                                                                                                                      //
          KeybaseProfiles.update(keyFingerprint, { $set: record }, { upsert: true });                                 // 1875
        } else {                                                                                                      //
          // Keybase reports no match, so remove what we know of this user. We don't want to remove                   //
          // the item entirely from the cache as this will cause us to repeatedly re-fetch the data                   //
          // from Keybase.                                                                                            //
          //                                                                                                          //
          // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,                    //
          //   but at present we don't have the ability to verify proofs.                                             //
          KeybaseProfiles.update(keyFingerprint, { $unset: { displayName: "", handle: "", proofs: "" } }, { upsert: true });
        }                                                                                                             //
      });                                                                                                             //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.deleteUnusedAccount = function (backend, identityId) {                                      // 1889
      // If there is an *unused* account that has `identityId` as a login identity, deletes it.                       //
                                                                                                                      //
      check(identityId, String);                                                                                      // 1892
      var account = Meteor.users.findOne({ "loginIdentities.id": identityId });                                       // 1893
      if (account && account.loginIdentities.length == 1 && account.nonloginIdentities.length == 0 && !Grains.findOne({ userId: account._id }) && !ApiTokens.findOne({ accountId: account._id }) && (!account.plan || account.plan === "free") && !(account.payments && account.payments.id) && !Contacts.findOne({ ownerId: account._id })) {
        Meteor.users.remove({ _id: account._id });                                                                    // 1902
        backend.deleteUser(account._id);                                                                              // 1903
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    Meteor.publish("keybaseProfile", function (keyFingerprint) {                                                      // 1907
      check(keyFingerprint, ValidKeyFingerprint);                                                                     // 1908
      var db = this.connection.sandstormDb;                                                                           // 1909
                                                                                                                      //
      var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                               // 1911
      if (cursor.count() === 0) {                                                                                     // 1912
        // Fire off async update.                                                                                     //
        db.updateKeybaseProfileAsync(keyFingerprint);                                                                 // 1914
      }                                                                                                               //
                                                                                                                      //
      return cursor;                                                                                                  // 1917
    });                                                                                                               //
                                                                                                                      //
    Meteor.publish("appIndex", function (appId) {                                                                     // 1920
      check(appId, String);                                                                                           // 1921
      var db = this.connection.sandstormDb;                                                                           // 1922
      var cursor = db.collections.appIndex.find({ _id: appId });                                                      // 1923
      return cursor;                                                                                                  // 1924
    });                                                                                                               //
                                                                                                                      //
    Meteor.publish("userPackages", function () {                                                                      // 1927
      var _this2 = this;                                                                                              //
                                                                                                                      //
      // Users should be able to see packages that are either:                                                        //
      // 1. referenced by one of their userActions                                                                    //
      // 2. referenced by one of their grains                                                                         //
      var db = this.connection.sandstormDb;                                                                           // 1931
                                                                                                                      //
      // Note that package information, once it is in the database, is static. There's no need to                     //
      // reactively subscribe to changes to a package since they don't change. It's also unecessary                   //
      // to reactively remove a package from the client side when it is removed on the server, or                     //
      // when the client stops using it, because the worst case is the client has a small amount                      //
      // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                       //
      // So, we implement this as a cache: the first time each package ID shows up among the user's                   //
      // stuff, we push the package info to the client, and then we never update it.                                  //
      //                                                                                                              //
      // Alternatively, we could subscribe to each individual package query, but this would waste                     //
      // lots of server-side resources watching for events that will never happen or don't matter.                    //
      var hasPackage = {};                                                                                            // 1943
      var refPackage = function (packageId) {                                                                         // 1944
        // Ignore dev apps.                                                                                           //
        if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                           // 1946
                                                                                                                      //
        if (!hasPackage[packageId]) {                                                                                 // 1948
          hasPackage[packageId] = true;                                                                               // 1949
          var pkg = db.getPackage(packageId);                                                                         // 1950
          if (pkg) {                                                                                                  // 1951
            _this2.added("packages", packageId, pkg);                                                                 // 1952
          } else {                                                                                                    //
            console.error("shouldn't happen: missing package referenced by user's stuff:", packageId);                // 1954
          }                                                                                                           //
        }                                                                                                             //
      };                                                                                                              //
                                                                                                                      //
      // package source 1: packages referred to by actions                                                            //
      var actions = db.userActions(this.userId);                                                                      // 1961
      var actionsHandle = actions.observe({                                                                           // 1962
        added: function (newAction) {                                                                                 // 1963
          refPackage(newAction.packageId);                                                                            // 1964
        },                                                                                                            //
                                                                                                                      //
        changed: function (oldAction, newAction) {                                                                    // 1967
          refPackage(newAction.packageId);                                                                            // 1968
        }                                                                                                             //
      });                                                                                                             //
                                                                                                                      //
      // package source 2: packages referred to by grains directly                                                    //
      var grains = db.userGrains(this.userId);                                                                        // 1973
      var grainsHandle = grains.observe({                                                                             // 1974
        added: function (newGrain) {                                                                                  // 1975
          refPackage(newGrain.packageId);                                                                             // 1976
        },                                                                                                            //
                                                                                                                      //
        changed: function (oldGrain, newGrain) {                                                                      // 1979
          refPackage(newGrain.packageId);                                                                             // 1980
        }                                                                                                             //
      });                                                                                                             //
                                                                                                                      //
      this.onStop(function () {                                                                                       // 1984
        actionsHandle.stop();                                                                                         // 1985
        grainsHandle.stop();                                                                                          // 1986
      });                                                                                                             //
                                                                                                                      //
      this.ready();                                                                                                   // 1989
    });                                                                                                               //
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 1993
  SandstormDb.prototype.currentFeatureKey = function () {                                                             // 1994
    // Returns an object with all of the current signed feature key properties,                                       //
    // or undefined, if the feature key is missing or not correctly signed.                                           //
    var doc = this.collections.featureKey.findOne({ _id: "currentFeatureKey" });                                      // 1997
    if (!doc) return undefined;                                                                                       // 1998
    var buf = new Buffer(doc.value);                                                                                  // 1999
    // We use loadSignedFeatureKey from server/feature-key.js.  This should probably get refactored                   //
    // once we can use ES6 modules.                                                                                   //
    return loadSignedFeatureKey(buf);                                                                                 // 2002
  };                                                                                                                  //
} else {                                                                                                              //
  SandstormDb.prototype.currentFeatureKey = function () {                                                             // 2005
    return this.collections.featureKey.findOne({ _id: "currentFeatureKey" });                                         // 2006
  };                                                                                                                  //
}                                                                                                                     //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/profile.js                                                                                   //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                                 //
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                              //
// All rights reserved.                                                                                               //
//                                                                                                                    //
// Licensed under the Apache License, Version 2.0 (the "License");                                                    //
// you may not use this file except in compliance with the License.                                                   //
// You may obtain a copy of the License at                                                                            //
//                                                                                                                    //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                       //
//                                                                                                                    //
// Unless required by applicable law or agreed to in writing, software                                                //
// distributed under the License is distributed on an "AS IS" BASIS,                                                  //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                           //
// See the License for the specific language governing permissions and                                                //
// limitations under the License.                                                                                     //
                                                                                                                      //
var makeIdenticon = undefined;                                                                                        // 17
var httpProtocol = undefined;                                                                                         // 18
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 20
  SandstormDb.ensureSubscriberHasIdentity = function (publishHandler, identityId) {                                   // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber                      //
    // has a given identity linked. Automatically stops the subscription if the user loses the                        //
    // identity. Returns a boolean indicating whether the user initially has the identity.                            //
                                                                                                                      //
    var userId = publishHandler.userId;                                                                               // 26
    if (userId === identityId) {                                                                                      // 27
      return true;                                                                                                    // 28
    } else {                                                                                                          //
      var _ret = (function () {                                                                                       //
        var hasIdentityCursor = Meteor.users.find({                                                                   // 30
          $or: [{                                                                                                     // 31
            _id: userId,                                                                                              // 33
            "loginIdentities.id": identityId                                                                          // 34
          }, {                                                                                                        //
            _id: userId,                                                                                              // 37
            "nonloginIdentities.id": identityId                                                                       // 38
          }]                                                                                                          //
        });                                                                                                           //
        if (hasIdentityCursor.count() == 0) {                                                                         // 42
          publishHandler.stop();                                                                                      // 43
          return {                                                                                                    // 44
            v: false                                                                                                  //
          };                                                                                                          //
        }                                                                                                             //
                                                                                                                      //
        var handle = hasIdentityCursor.observe({ removed: function () {                                               // 47
            publishHandler.stop();                                                                                    // 47
          } });                                                                                                       //
                                                                                                                      //
        publishHandler.onStop(function () {                                                                           // 49
          handle.stop();                                                                                              // 49
        });                                                                                                           //
                                                                                                                      //
        return {                                                                                                      // 51
          v: true                                                                                                     //
        };                                                                                                            //
      })();                                                                                                           //
                                                                                                                      //
      if (typeof _ret === "object") return _ret.v;                                                                    //
    }                                                                                                                 //
  };                                                                                                                  //
                                                                                                                      //
  Meteor.publish("identityProfile", function (identityId) {                                                           // 55
    check(identityId, String);                                                                                        // 56
    if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                           // 57
                                                                                                                      //
    return Meteor.users.find({ _id: identityId }, { fields: {                                                         // 59
        profile: 1,                                                                                                   // 61
        unverifiedEmail: 1,                                                                                           // 62
        expires: 1,                                                                                                   // 63
        createdAt: 1,                                                                                                 // 64
                                                                                                                      //
        "services.dev.name": 1,                                                                                       // 66
                                                                                                                      //
        "services.google.id": 1,                                                                                      // 68
        "services.google.email": 1,                                                                                   // 69
        "services.google.verified_email": 1,                                                                          // 70
        "services.google.name": 1,                                                                                    // 71
        "services.google.picture": 1,                                                                                 // 72
        "services.google.gender": 1,                                                                                  // 73
        "services.google.hd": 1,                                                                                      // 74
                                                                                                                      //
        "services.github.id": 1,                                                                                      // 76
        "services.github.email": 1,                                                                                   // 77
        "services.github.emails": 1,                                                                                  // 78
        "services.github.username": 1,                                                                                // 79
                                                                                                                      //
        "services.email.email": 1,                                                                                    // 81
                                                                                                                      //
        "services.ldap.id": 1,                                                                                        // 83
        "services.ldap.username": 1,                                                                                  // 84
        "services.ldap.rawAttrs": 1,                                                                                  // 85
                                                                                                                      //
        "services.saml.id": 1,                                                                                        // 87
        "services.saml.email": 1,                                                                                     // 88
        "services.saml.displayName": 1                                                                                // 89
      } });                                                                                                           //
  }), Meteor.publish("accountIdentities", function () {                                                               //
    // Maybe this should be folded into the "credentials" subscription?                                               //
                                                                                                                      //
    if (!this.userId) return [];                                                                                      // 96
                                                                                                                      //
    return Meteor.users.find({ _id: this.userId }, { fields: {                                                        // 98
        profile: 1,                                                                                                   // 101
        verifiedEmail: 1,                                                                                             // 102
        loginIdentities: 1,                                                                                           // 103
        nonloginIdentities: 1,                                                                                        // 104
        expires: 1,                                                                                                   // 105
        primaryEmail: 1                                                                                               // 106
      } });                                                                                                           //
  });                                                                                                                 //
                                                                                                                      //
  makeIdenticon = function (id) {                                                                                     // 110
    // We only make identicons client-side.                                                                           //
    return undefined;                                                                                                 // 112
  };                                                                                                                  //
                                                                                                                      //
  var Url = Npm.require("url");                                                                                       // 115
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                            // 116
} else {                                                                                                              //
  (function () {                                                                                                      //
    var identiconCache = {};                                                                                          // 118
                                                                                                                      //
    makeIdenticon = function (id) {                                                                                   // 120
      // Given a cryptographic hash as input, generate an identicon. We always use the user's                         //
      // identity ID as the input hash (the same thing passed to apps as the user ID).                                //
                                                                                                                      //
      // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so                      //
      // that apps can themselves use identicon.js to produce consistent identicons. As it turns out                  //
      // identicon.js doesn't use the second half of the hash even if we provide it, but slice it                     //
      // anyway to be safe.                                                                                           //
      id = id.slice(0, 32);                                                                                           // 128
                                                                                                                      //
      if (id in identiconCache) {                                                                                     // 130
        return identiconCache[id];                                                                                    // 131
      }                                                                                                               //
                                                                                                                      //
      // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.                    //
      var data = new Identicon(id, 512).toString();                                                                   // 135
      var result = "data:image/png;base64," + data;                                                                   // 136
      identiconCache[id] = result;                                                                                    // 137
      return result;                                                                                                  // 138
    };                                                                                                                //
                                                                                                                      //
    httpProtocol = window.location.protocol;                                                                          // 141
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
var GENDERS = { male: "male", female: "female", neutral: "neutral", robot: "robot" };                                 // 144
                                                                                                                      //
function staticAssetUrl(id, staticHost) {                                                                             // 146
  if (id) {                                                                                                           // 147
    return staticHost + "/" + id;                                                                                     // 148
  } else {                                                                                                            //
    return undefined;                                                                                                 // 150
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function filterHandle(handle) {                                                                                       // 154
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                           //
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                              // 156
}                                                                                                                     //
                                                                                                                      //
function emailToHandle(email) {                                                                                       // 159
  // Turn an email address into a handle.                                                                             //
                                                                                                                      //
  if (!email) return undefined;                                                                                       // 162
                                                                                                                      //
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters                        //
  // on the same account.                                                                                             //
  var parts = email.split("@");                                                                                       // 166
  var base = filterHandle(parts[0].split("+")[0]);                                                                    // 167
                                                                                                                      //
  var domain = (parts[1] || "").split(".");                                                                           // 169
  if (domain[domain.length - 1] === "name") {                                                                         // 170
    // Oh, a .name domain. Let's use                                                                                  //
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                                // 172
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info", "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"], base)) {
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                           //
    base = filterHandle(domain[0]);                                                                                   // 180
  }                                                                                                                   //
                                                                                                                      //
  return filterHandle(base);                                                                                          // 183
}                                                                                                                     //
                                                                                                                      //
SandstormDb.fillInProfileDefaults = function (user) {                                                                 // 186
  var profile = user.profile;                                                                                         // 187
  if (profile.service === "github") {                                                                                 // 188
    profile.name = profile.name || user.services.github.username || "Name Unknown";                                   // 189
    profile.handle = profile.handle || filterHandle(user.services.github.username) || filterHandle(profile.name);     // 190
  } else if (profile.service === "google") {                                                                          //
    profile.name = profile.name || user.services.google.name || "Name Unknown";                                       // 193
    profile.handle = profile.handle || emailToHandle(user.services.google.email) || filterHandle(profile.name);       // 194
    profile.pronoun = profile.pronoun || GENDERS[user.services.google.gender] || "neutral";                           // 196
  } else if (profile.service === "email") {                                                                           //
    var email = user.services.email.email;                                                                            // 198
    profile.name = profile.name || emailToHandle(email);                                                              // 199
    profile.handle = profile.handle || emailToHandle(email);                                                          // 200
  } else if (profile.service === "dev") {                                                                             //
    var lowerCaseName = user.services.dev.name.split(" ")[0].toLowerCase();                                           // 202
    profile.name = profile.name || user.services.dev.name;                                                            // 203
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                                   // 204
    profile.pronoun = profile.pronoun || (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" : _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");
  } else if (profile.service === "demo") {                                                                            //
    profile.name = profile.name || "Demo User";                                                                       // 209
    profile.handle = profile.handle || "demo";                                                                        // 210
  } else if (profile.service === "ldap") {                                                                            //
    var setting = Settings.findOne({ _id: "ldapNameField" });                                                         // 212
    var key = setting ? setting.value : "";                                                                           // 213
    profile.handle = profile.handle || user.services.ldap.username;                                                   // 214
    profile.name = profile.name || user.services.ldap.rawAttrs[key] || profile.handle;                                // 215
  } else if (profile.service === "saml") {                                                                            //
    profile.handle = profile.handle || emailToHandle(user.services.saml.email);                                       // 217
    profile.name = profile.name || user.services.saml.displayName || profile.handle;                                  // 218
  } else {                                                                                                            //
    throw new Error("unrecognized identity service: ", profile.service);                                              // 220
  }                                                                                                                   //
                                                                                                                      //
  profile.pronoun = profile.pronoun || "neutral";                                                                     // 223
};                                                                                                                    //
                                                                                                                      //
SandstormDb.fillInIntrinsicName = function (user) {                                                                   // 226
  var profile = user.profile;                                                                                         // 227
  if (profile.service === "github") {                                                                                 // 228
    profile.intrinsicName = user.services.github.username;                                                            // 229
  } else if (profile.service === "google") {                                                                          //
    profile.intrinsicName = user.services.google.name;                                                                // 231
    user.privateIntrinsicName = user.services.google.email;                                                           // 232
  } else if (profile.service === "email") {                                                                           //
    profile.intrinsicName = user.services.email.email;                                                                // 234
  } else if (profile.service === "dev") {                                                                             //
    profile.intrinsicName = user.services.dev.name;                                                                   // 236
  } else if (profile.service === "demo") {                                                                            //
    profile.intrinsicName = "demo on " + user.createdAt.toISOString().substring(0, 10);                               // 238
  } else if (profile.service === "ldap") {                                                                            //
    profile.intrinsicName = user.services.ldap.username;                                                              // 240
  } else if (profile.service === "saml") {                                                                            //
    profile.intrinsicName = user.services.saml.id;                                                                    // 242
  } else {                                                                                                            //
    throw new Error("unrecognized identity service: ", profile.service);                                              // 244
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
SandstormDb.getVerifiedEmails = function (identity) {                                                                 // 248
  if (identity.services.google && identity.services.google.email && identity.services.google.verified_email) {        // 249
    // jscs:ignore requireCamelCaseOrUpperCaseIdentifiers                                                             //
    return [{ email: identity.services.google.email, primary: true }];                                                // 251
  } else if (identity.services.email) {                                                                               //
    return [{ email: identity.services.email.email, primary: true }];                                                 // 253
  } else if (identity.services.github && identity.services.github.emails) {                                           //
    return _.chain(identity.services.github.emails).filter(function (email) {                                         // 255
      return email.verified;                                                                                          // 256
    }).map(function (email) {                                                                                         //
      return _.pick(email, "email", "primary");                                                                       //
    }).value();                                                                                                       //
  } else if (identity.services.ldap) {                                                                                //
    var email = identity.services.ldap.rawAttrs[SandstormDb.prototype.getLdapEmailField()];                           // 260
    if (email) {                                                                                                      // 261
      return [{ email: email, primary: true }];                                                                       // 262
    }                                                                                                                 //
  } else if (identity.services.saml && identity.services.saml.email) {                                                //
    return [{ email: identity.services.saml.email, primary: true }];                                                  // 265
  }                                                                                                                   //
                                                                                                                      //
  return [];                                                                                                          // 268
};                                                                                                                    //
                                                                                                                      //
SandstormDb.prototype.findIdentitiesByEmail = function (email) {                                                      // 271
  // Returns an array of identities which have the given email address as one of their verified                       //
  // addresses.                                                                                                       //
                                                                                                                      //
  check(email, String);                                                                                               // 275
                                                                                                                      //
  return Meteor.users.find({ $or: [{ "services.google.email": email }, { "services.email.email": email }, { "services.github.emails.email": email }, { "services.saml.email": email }] }).fetch().filter(function (identity) {
    // Verify that the email is verified, since our query doesn't technically do that.                                //
    return !!_findWhere(SandstormDb.getVerifiedEmails(identity), { email: email });                                   // 284
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
SandstormDb.prototype.findAccountsByEmail = function (email) {                                                        // 288
  var identityIds = _.pluck(this.findIdentitiesByEmail(email), "_id");                                                // 289
  return Meteor.users.find({ $or: [{ "loginIdentities.id": { $in: identityIds } }, { "nonloginIdentities.id": { $in: identityIds } }] }).fetch();
};                                                                                                                    //
                                                                                                                      //
SandstormDb.fillInPictureUrl = function (user) {                                                                      // 296
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                                  // 297
  user.profile.pictureUrl = staticAssetUrl(user.profile.picture, staticHost) || makeIdenticon(user._id);              // 298
};                                                                                                                    //
                                                                                                                      //
SandstormDb.getUserIdentityIds = function (user) {                                                                    // 302
  // Given an account user object, returns an array containing the ID of each identity linked to the                  //
  // account. Always returns the most recently added login identity first.                                            //
  if (user && user.loginIdentities) {                                                                                 // 305
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                             // 306
  } else {                                                                                                            //
    return [];                                                                                                        // 308
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
SandstormDb.getUserEmails = function (user) {                                                                         // 312
  // Given a user object, returns an array containing all email addresses associated with that user.                  //
  // Each entry in the array is an object of the form:                                                                //
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                                   //
  //                                                                                                                  //
  // At most one entry in the result has `primary = true`.                                                            //
  //                                                                                                                  //
  // TODO(cleanup): This actually does need to query the database to fetch profile information                        //
  //   for linked identities, so it probably makes more sense for it to be a non-static method                        //
  //   on SandstormDb.                                                                                                //
                                                                                                                      //
  var identityIds = SandstormDb.getUserIdentityIds(user);                                                             // 323
  var verifiedEmails = {};                                                                                            // 324
  var unverifiedEmails = {};                                                                                          // 325
                                                                                                                      //
  identityIds.forEach(function (id) {                                                                                 // 327
    var identity = Meteor.users.findOne({ _id: id });                                                                 // 328
    if (identity && identity.services) {                                                                              // 329
      SandstormDb.getVerifiedEmails(identity).forEach(function (verifiedEmail) {                                      // 330
        if (verifiedEmail) {                                                                                          // 331
          verifiedEmails[verifiedEmail.email] = true;                                                                 // 332
        }                                                                                                             //
      });                                                                                                             //
    }                                                                                                                 //
                                                                                                                      //
    if (identity && identity.unverifiedEmail) {                                                                       // 337
      unverifiedEmails[identity.unverifiedEmail] = true;                                                              // 338
    }                                                                                                                 //
  });                                                                                                                 //
                                                                                                                      //
  var result = [];                                                                                                    // 342
  _.keys(verifiedEmails).map(function (email) {                                                                       // 343
    result.push({ email: email,                                                                                       // 344
      verified: true,                                                                                                 // 345
      primary: email === user.primaryEmail });                                                                        // 346
  });                                                                                                                 //
                                                                                                                      //
  _.keys(unverifiedEmails).map(function (email) {                                                                     // 349
    if (!(email in verifiedEmails)) {                                                                                 // 350
      result.push({ email: email, verified: false });                                                                 // 350
    }                                                                                                                 //
  });                                                                                                                 //
                                                                                                                      //
  // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                              //
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                                  // 354
    result[0].primary = true;                                                                                         // 355
  }                                                                                                                   //
                                                                                                                      //
  return result;                                                                                                      // 358
};                                                                                                                    //
                                                                                                                      //
SandstormDb.prototype.addContact = (function () {                                                                     // 361
  function addContact(ownerId, identityId) {                                                                          // 361
    check(ownerId, String);                                                                                           // 362
    check(identityId, String);                                                                                        // 363
    var profile = this.getIdentity(identityId).profile;                                                               // 364
    this.collections.contacts.upsert({ ownerId: ownerId, identityId: identityId }, {                                  // 365
      ownerId: ownerId,                                                                                               // 366
      petname: profile && profile.name,                                                                               // 367
      created: new Date(),                                                                                            // 368
      identityId: identityId                                                                                          // 369
    });                                                                                                               //
  }                                                                                                                   //
                                                                                                                      //
  return addContact;                                                                                                  //
})();                                                                                                                 //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/user.js                                                                                      //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                                 //
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                              //
// All rights reserved.                                                                                               //
//                                                                                                                    //
// Licensed under the Apache License, Version 2.0 (the "License");                                                    //
// you may not use this file except in compliance with the License.                                                   //
// You may obtain a copy of the License at                                                                            //
//                                                                                                                    //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                       //
//                                                                                                                    //
// Unless required by applicable law or agreed to in writing, software                                                //
// distributed under the License is distributed on an "AS IS" BASIS,                                                  //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                           //
// See the License for the specific language governing permissions and                                                //
// limitations under the License.                                                                                     //
                                                                                                                      //
var Crypto = Npm.require("crypto");                                                                                   // 17
var Future = Npm.require("fibers/future");                                                                            // 18
                                                                                                                      //
userPictureUrl = function (user) {                                                                                    // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                                     // 21
    // Try to determine user's avatar URL from login service.                                                         //
                                                                                                                      //
    var google = user.services.google;                                                                                // 24
    if (google && google.picture) {                                                                                   // 25
      return google.picture;                                                                                          // 26
    }                                                                                                                 //
                                                                                                                      //
    var github = user.services.github;                                                                                // 29
    if (github && github.id) {                                                                                        // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                                  // 31
    }                                                                                                                 //
                                                                                                                      //
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be                     //
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                         //
    // Github are different because they are actually the identity providers, so they already know                    //
    // the user logged in.                                                                                            //
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
fetchPicture = function (url) {                                                                                       // 41
  try {                                                                                                               // 42
    var result = HTTP.get(url, {                                                                                      // 43
      npmRequestOptions: { encoding: null },                                                                          // 44
      timeout: 5000                                                                                                   // 45
    });                                                                                                               //
                                                                                                                      //
    var metadata = {};                                                                                                // 48
                                                                                                                      //
    metadata.mimeType = result.headers["content-type"];                                                               // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 && metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                                 // 53
    }                                                                                                                 //
                                                                                                                      //
    var enc = result.headers["content-encoding"];                                                                     // 56
    if (enc && enc !== "identity") {                                                                                  // 57
      metadata.encoding = enc;                                                                                        // 58
    }                                                                                                                 //
                                                                                                                      //
    return addStaticAsset(metadata, result.content);                                                                  // 61
  } catch (err) {                                                                                                     //
    console.error("failed to fetch user profile picture:", url, err.stack);                                           // 63
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
var ValidHandle = Match.Where(function (handle) {                                                                     // 67
  check(handle, String);                                                                                              // 68
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                                        // 69
});                                                                                                                   //
                                                                                                                      //
Accounts.onCreateUser(function (options, user) {                                                                      // 72
  if (user.loginIdentities) {                                                                                         // 73
    // it's an account                                                                                                //
    check(user, { _id: String,                                                                                        // 75
      createdAt: Date,                                                                                                // 76
      isAdmin: Match.Optional(Boolean),                                                                               // 77
      hasCompletedSignup: Match.Optional(Boolean),                                                                    // 78
      signupKey: Match.Optional(String),                                                                              // 79
      signupNote: Match.Optional(String),                                                                             // 80
      signupEmail: Match.Optional(String),                                                                            // 81
      expires: Match.Optional(Date),                                                                                  // 82
      appDemoId: Match.Optional(String),                                                                              // 83
      loginIdentities: [{ id: String }],                                                                              // 84
      nonloginIdentities: [{ id: String }] });                                                                        // 85
                                                                                                                      //
    if (Meteor.settings["public"].quotaEnabled) {                                                                     // 87
      user.experiments = user.experiments || {};                                                                      // 88
      user.experiments = {                                                                                            // 89
        firstTimeBillingPrompt: Math.random() < 0.5 ? "control" : "test"                                              // 90
      };                                                                                                              //
      if (!("expires" in user)) {                                                                                     // 92
        sendReferralProgramNotification(user._id);                                                                    // 93
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return user;                                                                                                      // 97
  }                                                                                                                   //
                                                                                                                      //
  // Check profile.                                                                                                   //
  if (options.profile) {                                                                                              // 101
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                              //
    check(options.profile, Match.ObjectIncluding({                                                                    // 103
      name: Match.OneOf(null, Match.Optional(String)),                                                                // 104
      handle: Match.Optional(ValidHandle),                                                                            // 105
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot"))                                      // 106
    }));                                                                                                              //
  }                                                                                                                   //
                                                                                                                      //
  check(options.unverifiedEmail, Match.Optional(String));                                                             // 110
                                                                                                                      //
  if (options.unverifiedEmail) {                                                                                      // 112
    user.unverifiedEmail = options.unverifiedEmail;                                                                   // 113
  }                                                                                                                   //
                                                                                                                      //
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronouns");                                         // 116
                                                                                                                      //
  // Try downloading avatar.                                                                                          //
  var url = userPictureUrl(user);                                                                                     // 119
  if (url) {                                                                                                          // 120
    var assetId = fetchPicture(url);                                                                                  // 121
    if (assetId) {                                                                                                    // 122
      user.profile.picture = assetId;                                                                                 // 123
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  var serviceUserId = undefined;                                                                                      // 127
  if (user.services && user.services.dev) {                                                                           // 128
    check(user.services.dev, { name: String, isAdmin: Boolean, hasCompletedSignup: Boolean });                        // 129
    serviceUserId = user.services.dev.name;                                                                           // 130
    user.profile.service = "dev";                                                                                     // 131
  } else if ("expires" in user) {                                                                                     //
    serviceUserId = user._id;                                                                                         // 133
    user.profile.service = "demo";                                                                                    // 134
  } else if (user.services && user.services.email) {                                                                  //
    check(user.services.email, { email: String,                                                                       // 136
      tokens: [{ digest: String, algorithm: String, createdAt: Date }] });                                            // 138
    serviceUserId = user.services.email.email;                                                                        // 139
    user.profile.service = "email";                                                                                   // 140
  } else if (user.services && "google" in user.services) {                                                            //
    serviceUserId = user.services.google.id;                                                                          // 142
    user.profile.service = "google";                                                                                  // 143
  } else if (user.services && "github" in user.services) {                                                            //
    serviceUserId = user.services.github.id;                                                                          // 145
    user.profile.service = "github";                                                                                  // 146
  } else if (user.services && "ldap" in user.services) {                                                              //
    serviceUserId = user.services.ldap.id;                                                                            // 148
    user.profile.service = "ldap";                                                                                    // 149
  } else if (user.services && "saml" in user.services) {                                                              //
    serviceUserId = user.services.saml.id;                                                                            // 151
    user.profile.service = "saml";                                                                                    // 152
  } else {                                                                                                            //
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " + JSON.stringify(user));        // 154
  }                                                                                                                   //
                                                                                                                      //
  user._id = Crypto.createHash("sha256").update(user.profile.service + ":" + serviceUserId).digest("hex");            // 158
                                                                                                                      //
  return user;                                                                                                        // 161
});                                                                                                                   //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/migrations.js                                                                                //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                                   //
//                                                                                                                    //
// Because migrations can experience partial failure and likely have                                                  //
// side-effects, we should be careful to make sure all migrations are                                                 //
// idempotent and safe to accidentally run multiple times.                                                            //
                                                                                                                      //
var Future = Npm.require("fibers/future");                                                                            // 7
var Url = Npm.require("url");                                                                                         // 8
var Crypto = Npm.require("crypto");                                                                                   // 9
                                                                                                                      //
var updateLoginStyleToRedirect = function () {                                                                        // 11
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                          // 12
  ["google", "github"].forEach(function (serviceName) {                                                               // 13
    var config = configurations.findOne({ service: serviceName });                                                    // 14
    if (config && config.loginStyle !== "redirect") {                                                                 // 15
      configurations.update({ service: serviceName }, { $set: { loginStyle: "redirect" } });                          // 16
    }                                                                                                                 //
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
var enableLegacyOAuthProvidersIfNotInSettings = function () {                                                         // 21
  // In the before time, Google and Github login were enabled by default.                                             //
  //                                                                                                                  //
  // This actually didn't make much sense, required the first user to configure                                       //
  // OAuth, and had some trust-the-first-user properties that weren't totally                                         //
  // secure.                                                                                                          //
  //                                                                                                                  //
  // Now, we have admin-token, so we wish to disable all logins by default                                            //
  // (since they need to be configured anyway) but since users may have never                                         //
  // explicitly told Sandstorm that Google or Github login should be enabled,                                         //
  // we can't just use the value in the Settings collection, since it might                                           //
  // never have been set.                                                                                             //
  //                                                                                                                  //
  // Thus, if the service is configured but absent from Settings, we should                                           //
  // explicitly enable it in Settings, and then the rest of the logic can just                                        //
  // depend on what value is in Settings and default to false without breaking                                        //
  // user installations.                                                                                              //
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                          // 38
  ["google", "github"].forEach(function (serviceName) {                                                               // 39
    var config = configurations.findOne({ service: serviceName });                                                    // 40
    var serviceConfig = Settings.findOne({ _id: serviceName });                                                       // 41
    if (config && !serviceConfig) {                                                                                   // 42
      // Only explicitly enable the login service if:                                                                 //
      // 1) the service is already configured                                                                         //
      // 2) there is no sandstorm configuration already present (the user was                                         //
      //    using the previous default behavior).                                                                     //
      Settings.insert({ _id: serviceName, value: true });                                                             // 47
    }                                                                                                                 //
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
var denormalizeInviteInfo = function () {                                                                             // 52
  // When a user is invited via a signup token, the `signupKey` field of their user table entry                       //
  // has always been populated to indicate the key they used. This points into the SignupKeys table                   //
  // which has more information about the key, namely a freeform note entered by the admin when                       //
  // they created the key. In the case that the email invite form was used, the note has the form                     //
  // "E-mail invite to <address>".                                                                                    //
  //                                                                                                                  //
  // Later, we decided it was useful to indicate in the users table visible to the admin                              //
  // information about the invite terms. Namely, for email invites we want to show the address                        //
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                         //
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing                       //
  // from the SignupKeys table.                                                                                       //
                                                                                                                      //
  Meteor.users.find().forEach(function (user) {                                                                       // 65
    if (user.signupKey && typeof user.signupKey === "string" && user.signupKey !== "admin") {                         // 66
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                            // 67
      if (signupInfo && signupInfo.note) {                                                                            // 68
        var newFields = { signupNote: signupInfo.note };                                                              // 69
                                                                                                                      //
        var prefix = "E-mail invite to ";                                                                             // 71
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                              // 72
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                               // 73
        }                                                                                                             //
                                                                                                                      //
        Meteor.users.update(user._id, { $set: newFields });                                                           // 76
      }                                                                                                               //
    }                                                                                                                 //
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
function mergeRoleAssignmentsIntoApiTokens() {                                                                        // 82
  RoleAssignments.find().forEach(function (roleAssignment) {                                                          // 83
    ApiTokens.insert({                                                                                                // 84
      grainId: roleAssignment.grainId,                                                                                // 85
      userId: roleAssignment.sharer,                                                                                  // 86
      roleAssignment: roleAssignment.roleAssignment,                                                                  // 87
      petname: roleAssignment.petname,                                                                                // 88
      created: roleAssignment.created,                                                                                // 89
      owner: { user: { userId: roleAssignment.recipient,                                                              // 90
          title: roleAssignment.title } }                                                                             // 91
    });                                                                                                               //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function fixOasisStorageUsageStats() {}                                                                               // 96
// This migration only pertained to Oasis and it was successfully applied there. Since it referred                    //
// to some global variables that we later wanted to remove and/or rename, we've since replaced it                     //
// with a no-op.                                                                                                      //
                                                                                                                      //
function fetchProfilePictures() {                                                                                     // 101
  Meteor.users.find({}).forEach(function (user) {                                                                     // 102
    var url = userPictureUrl(user);                                                                                   // 103
    if (url) {                                                                                                        // 104
      console.log("Fetching user picture:", url);                                                                     // 105
      var assetId = fetchPicture(url);                                                                                // 106
      if (assetId) {                                                                                                  // 107
        Meteor.users.update(user._id, { $set: { "profile.picture": assetId } });                                      // 108
      }                                                                                                               //
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function assignPlans() {                                                                                              // 114
  if (Meteor.settings["public"].quotaEnabled && SandstormDb.paymentsMigrationHook) {                                  // 115
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                              // 116
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function removeKeyrings() {                                                                                           // 120
  // These blobs full of public keys were not intended to find their way into mongo and while                         //
  // harmless they slow things down because they're huge. Remove them.                                                //
  Packages.update({ "manifest.metadata.pgpKeyring": { $exists: true } }, { $unset: { "manifest.metadata.pgpKeyring": "" } }, { multi: true });
}                                                                                                                     //
                                                                                                                      //
function useLocalizedTextInUserActions() {                                                                            // 128
  function toLocalizedText(newObj, oldObj, field) {                                                                   // 129
    if (field in oldObj) {                                                                                            // 130
      if (typeof oldObj[field] === "string") {                                                                        // 131
        newObj[field] = { defaultText: oldObj[field] };                                                               // 132
      } else {                                                                                                        //
        newObj[field] = oldObj[field];                                                                                // 134
      }                                                                                                               //
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  UserActions.find({}).forEach(function (userAction) {                                                                // 139
    var fields = {};                                                                                                  // 140
    toLocalizedText(fields, userAction, "appTitle");                                                                  // 141
    toLocalizedText(fields, userAction, "title");                                                                     // 142
    toLocalizedText(fields, userAction, "nounPhrase");                                                                // 143
    UserActions.update(userAction._id, { $set: fields });                                                             // 144
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function verifyAllPgpSignatures() {                                                                                   // 148
  Packages.find({}).forEach(function (pkg) {                                                                          // 149
    try {                                                                                                             // 150
      console.log("checking PGP signature for package:", pkg._id);                                                    // 151
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                             // 152
      if (info.authorPgpKeyFingerprint) {                                                                             // 153
        console.log("  " + info.authorPgpKeyFingerprint);                                                             // 154
        Packages.update(pkg._id, { $set: { authorPgpKeyFingerprint: info.authorPgpKeyFingerprint } });                // 155
      } else {                                                                                                        //
        console.log("  no signature");                                                                                // 158
      }                                                                                                               //
    } catch (err) {                                                                                                   //
      console.error(err.stack);                                                                                       // 161
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                                 // 166
  Meteor.users.find().forEach(function (user) {                                                                       // 167
    var identity = {};                                                                                                // 168
    var serviceUserId = undefined;                                                                                    // 169
    if ("devName" in user) {                                                                                          // 170
      identity.service = "dev";                                                                                       // 171
      serviceUserId = user.devName;                                                                                   // 172
    } else if ("expires" in user) {                                                                                   //
      identity.service = "demo";                                                                                      // 174
      serviceUserId = user._id;                                                                                       // 175
    } else if (user.services && "google" in user.services) {                                                          //
      identity.service = "google";                                                                                    // 177
      if (user.services.google.email && user.services.google.verified_email) {                                        // 178
        // jscs:ignore requireCamelCaseOrUpperCaseIdentifiers                                                         //
        identity.verifiedEmail = user.services.google.email;                                                          // 179
      }                                                                                                               //
                                                                                                                      //
      serviceUserId = user.services.google.id;                                                                        // 182
    } else if (user.services && "github" in user.services) {                                                          //
      identity.service = "github";                                                                                    // 184
      identity.unverifiedEmail = user.services.github.email;                                                          // 185
      serviceUserId = user.services.github.id;                                                                        // 186
    } else if (user.services && "emailToken" in user.services) {                                                      //
      identity.service = "emailToken";                                                                                // 188
      identity.verifiedEmail = user.services.emailToken.email;                                                        // 189
      serviceUserId = user.services.emailToken.email;                                                                 // 190
    }                                                                                                                 //
                                                                                                                      //
    identity.id = Crypto.createHash("sha256").update(identity.service + ":" + serviceUserId).digest("hex");           // 193
                                                                                                                      //
    if (user.profile) {                                                                                               // 196
      if (user.profile.name) {                                                                                        // 197
        identity.name = user.profile.name;                                                                            // 198
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.handle) {                                                                                      // 201
        identity.handle = user.profile.handle;                                                                        // 202
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.picture) {                                                                                     // 205
        identity.picture = user.profile.picture;                                                                      // 206
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.pronoun) {                                                                                     // 209
        identity.pronoun = user.profile.pronoun;                                                                      // 210
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.email) {                                                                                       // 213
        identity.unverifiedEmail = user.profile.email;                                                                // 214
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    identity.main = true;                                                                                             // 218
                                                                                                                      //
    Meteor.users.update(user._id, { $set: { identities: [identity] } });                                              // 220
                                                                                                                      //
    Grains.update({ userId: user._id }, { $set: { identityId: identity.id } }, { multi: true });                      // 222
    Sessions.update({ userId: user._id }, { $set: { identityId: identity.id } }, { multi: true });                    // 223
    ApiTokens.update({ userId: user._id }, { $set: { identityId: identity.id } }, { multi: true });                   // 224
    ApiTokens.update({ "owner.user.userId": user._id }, { $set: { "owner.user.identityId": identity.id } }, { multi: true });
    ApiTokens.update({ "owner.grain.introducerUser": user._id }, { $set: { "owner.grain.introducerIdentity": identity.id } }, { multi: true });
                                                                                                                      //
    while (ApiTokens.update({ "requirements.permissionsHeld.userId": user._id }, { $set: { "requirements.$.permissionsHeld.identityId": identity.id },
      $unset: { "requirements.$.permissionsHeld.userId": 1 } }, { multi: true }) > 0);                                // 236
    // The `$` operatorer modifies the first element in the array that matches the query. Since                       //
    // there may be many matches, we need to repeat until no documents are modified.                                  //
  });                                                                                                                 //
                                                                                                                      //
  ApiTokens.remove({ userInfo: { $exists: true } });                                                                  // 243
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                             //
  // that this field could show up in the database was in this deprecated, no-longer-functional                       //
  // form of API token.                                                                                               //
}                                                                                                                     //
                                                                                                                      //
function appUpdateSettings() {                                                                                        // 249
  Settings.insert({ _id: "appMarketUrl", value: "https://apps.sandstorm.io" });                                       // 250
  Settings.insert({ _id: "appIndexUrl", value: "https://app-index.sandstorm.io" });                                   // 251
  Settings.insert({ _id: "appUpdatesEnabled", value: true });                                                         // 252
}                                                                                                                     //
                                                                                                                      //
function moveDevAndEmailLoginDataIntoIdentities() {                                                                   // 255
  Meteor.users.find().forEach(function (user) {                                                                       // 256
    if (user.identities.length != 1) {                                                                                // 257
      throw new Error("User does not have exactly one identity: ", user);                                             // 258
    }                                                                                                                 //
                                                                                                                      //
    var identity = user.identities[0];                                                                                // 261
    if (Match.test(identity.service, Object)) {                                                                       // 262
      return;                                                                                                         // 262
    } // Already migrated.                                                                                            //
                                                                                                                      //
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");                  // 264
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                                   // 265
                                                                                                                      //
    var serviceObject = {};                                                                                           // 267
    var fieldsToUnset = {};                                                                                           // 268
                                                                                                                      //
    if (identity.service === "dev") {                                                                                 // 270
      serviceObject.name = user.devName;                                                                              // 271
      fieldsToUnset.devName = 1;                                                                                      // 272
    } else if (identity.service === "emailToken") {                                                                   //
      serviceObject.tokens = user.services.emailToken.tokens;                                                         // 274
      serviceObject.email = user.services.emailToken.email;                                                           // 275
      fieldsToUnset["services.emailToken"] = 1;                                                                       // 276
    }                                                                                                                 //
                                                                                                                      //
    newIdentity.service = {};                                                                                         // 279
    newIdentity.service[identity.service] = serviceObject;                                                            // 280
                                                                                                                      //
    var modifier = { $set: { identities: [newIdentity] } };                                                           // 282
    if (Object.keys(fieldsToUnset).length > 0) {                                                                      // 283
      modifier.$unset = fieldsToUnset;                                                                                // 284
    }                                                                                                                 //
                                                                                                                      //
    Meteor.users.update({ _id: user._id }, modifier);                                                                 // 287
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function repairEmailIdentityIds() {                                                                                   // 291
  Meteor.users.find({ "identities.service.emailToken": { $exists: 1 } }).forEach(function (user) {                    // 292
    if (user.identities.length != 1) {                                                                                // 293
      throw new Error("User does not have exactly one identity: ", user);                                             // 294
    }                                                                                                                 //
                                                                                                                      //
    var identity = user.identities[0];                                                                                // 297
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail", "profile");              // 298
    newIdentity.service = { email: identity.service.emailToken };                                                     // 300
    newIdentity.id = Crypto.createHash("sha256").update("email:" + identity.service.emailToken.email).digest("hex");  // 301
                                                                                                                      //
    Grains.update({ identityId: identity.id }, { $set: { identityId: newIdentity.id } }, { multi: true });            // 304
    Sessions.update({ identityId: identity.id }, { $set: { identityId: newIdentity.id } }, { multi: true });          // 305
    ApiTokens.update({ identityId: identity.id }, { $set: { identityId: newIdentity.id } }, { multi: true });         // 306
    ApiTokens.update({ "owner.user.identityId": identity.id }, { $set: { "owner.user.identityId": newIdentity.id } }, { multi: true });
    ApiTokens.update({ "owner.grain.introducerIdentity": identity.id }, { $set: { "owner.grain.introducerIdentity": newIdentity.id } }, { multi: true });
                                                                                                                      //
    while (ApiTokens.update({ "requirements.permissionsHeld.identityId": identity.id }, { $set: { "requirements.$.permissionsHeld.identityId": newIdentity.id } }, { multi: true }) > 0);
                                                                                                                      //
    Meteor.users.update({ _id: user._id }, { $set: { identities: [newIdentity] } });                                  // 320
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function splitAccountUsersAndIdentityUsers() {                                                                        // 324
  Meteor.users.find({ identities: { $exists: true } }).forEach(function (user) {                                      // 325
    if (user.identities.length != 1) {                                                                                // 326
      throw new Error("User does not have exactly one identity: ", user);                                             // 327
    }                                                                                                                 //
                                                                                                                      //
    var identity = user.identities[0];                                                                                // 330
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires");                                            // 331
    identityUser._id = identity.id;                                                                                   // 332
    identityUser.profile = identity.profile;                                                                          // 333
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                                      // 334
    identityUser.profile.service = Object.keys(identity.service)[0];                                                  // 335
                                                                                                                      //
    // Updating this user needs to be a two step process because the `services` field typically                       //
    // contains subfields that are constrained to be unique by Mongo indices.                                         //
    identityUser.stagedServices = _.omit(user.services, "resume");                                                    // 339
    if (identity.service.dev) {                                                                                       // 340
      identityUser.stagedServices.dev = identity.service.dev;                                                         // 341
    } else if (identity.service.email) {                                                                              //
      identityUser.stagedServices.email = identity.service.email;                                                     // 343
    }                                                                                                                 //
                                                                                                                      //
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires", "isAdmin", "signupKey", "signupNote", "signupEmail", "plan", "storageUsage", "isAppDemoUser", "appDemoId", "payments", "dailySentMailCount", "hasCompletedSignup");
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                           // 350
    accountUser.nonloginIdentities = [];                                                                              // 351
    if (user.services && user.services.resume) {                                                                      // 352
      accountUser.services = { resume: user.services.resume };                                                        // 353
    }                                                                                                                 //
                                                                                                                      //
    accountUser.stashedOldUser = user;                                                                                // 356
                                                                                                                      //
    ApiTokens.update({ identityId: identityUser._id }, { $set: { accountId: user._id } }, { multi: true });           // 358
                                                                                                                      //
    Meteor.users.upsert({ _id: identityUser._id }, identityUser);                                                     // 361
    Meteor.users.update({ _id: user._id }, accountUser);                                                              // 362
  });                                                                                                                 //
                                                                                                                      //
  Meteor.users.find({ stagedServices: { $exists: true } }).forEach(function (identity) {                              // 365
    Meteor.users.update({ _id: identity._id }, { $unset: { stagedServices: 1 },                                       // 366
      $set: { services: identity.stagedServices } });                                                                 // 367
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function populateContactsFromApiTokens() {                                                                            // 371
  ApiTokens.find({ "owner.user.identityId": { $exists: 1 },                                                           // 372
    accountId: { $exists: 1 } }).forEach(function (token) {                                                           // 373
    var identityId = token.owner.user.identityId;                                                                     // 374
    var identity = SandstormDb.prototype.getIdentity(identityId);                                                     // 375
    if (identity) {                                                                                                   // 376
      var profile = identity.profile;                                                                                 // 377
      Contacts.upsert({ ownerId: token.accountId, identityId: identityId }, {                                         // 378
        ownerId: token.accountId,                                                                                     // 379
        petname: profile && profile.name,                                                                             // 380
        created: new Date(),                                                                                          // 381
        identityId: identityId,                                                                                       // 382
        profile: profile                                                                                              // 383
      });                                                                                                             //
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function cleanUpApiTokens() {                                                                                         // 389
  // The `splitUserIdsIntoAccountIdsAndIdentityIds()` migration only added `identityId` in cases                      //
  // where the user still existed in the database.                                                                    //
  ApiTokens.remove({ userId: { $exists: true }, identityId: { $exists: false } });                                    // 392
  ApiTokens.remove({ "owner.user.userId": { $exists: true },                                                          // 393
    "owner.user.identityId": { $exists: false } });                                                                   // 394
                                                                                                                      //
  // For a while we were accidentally setting `appIcon` instead of `icon`.                                            //
  ApiTokens.find({ "owner.user.denormalizedGrainMetadata.appIcon": { $exists: true } }).forEach(function (apiToken) {
    var icon = apiToken.owner.user.denormalizedGrainMetadata.appIcon;                                                 // 399
    ApiTokens.update({ _id: apiToken._id }, { $set: { "owner.user.denormalizedGrainMetadata.icon": icon },            // 400
      $unset: { "owner.user.denormalizedGrainMetadata.appIcon": true } });                                            // 402
  });                                                                                                                 //
                                                                                                                      //
  // For a while the `identityId` field of child UiView tokens was not getting set.                                   //
  function repairChain(parentToken) {                                                                                 // 406
    ApiTokens.find({ parentToken: parentToken._id, grainId: { $exists: true },                                        // 407
      identityId: { $exists: false } }).forEach(function (childToken) {                                               // 408
      ApiTokens.update({ _id: childToken._id }, { $set: { identityId: parentToken.identityId } });                    // 409
      repairChain(childToken);                                                                                        // 410
    });                                                                                                               //
  }                                                                                                                   //
                                                                                                                      //
  ApiTokens.find({ grainId: { $exists: true }, identityId: { $exists: true },                                         // 414
    parentToken: { $exists: false } }).forEach(repairChain);                                                          // 415
}                                                                                                                     //
                                                                                                                      //
function initServerTitleAndReturnAddress() {                                                                          // 418
  var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                            // 419
  Settings.insert({ _id: "serverTitle", value: hostname });                                                           // 420
  Settings.insert({ _id: "returnAddress", value: "no-reply@" + hostname });                                           // 421
}                                                                                                                     //
                                                                                                                      //
function sendReferralNotifications() {                                                                                // 424
  if (Meteor.settings["public"].quotaEnabled) {                                                                       // 425
    Meteor.users.find({                                                                                               // 426
      loginIdentities: { $exists: true },                                                                             // 427
      expires: { $exists: false }                                                                                     // 428
    }, { fields: { _id: 1 } }).forEach(function (user) {                                                              //
      sendReferralProgramNotification(user._id);                                                                      // 430
    });                                                                                                               //
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function assignBonuses() {                                                                                            // 435
  if (Meteor.settings["public"].quotaEnabled && SandstormDb.bonusesMigrationHook) {                                   // 436
    SandstormDb.bonusesMigrationHook();                                                                               // 437
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function splitSmtpUrl() {                                                                                             // 441
  var smtpUrlSetting = Settings.findOne({ _id: "smtpUrl" });                                                          // 442
  var smtpUrl = smtpUrlSetting ? smtpUrlSetting.value : process.env.MAIL_URL;                                         // 443
  var returnAddress = Settings.findOne({ _id: "returnAddress" });                                                     // 444
                                                                                                                      //
  // Default values.                                                                                                  //
  var smtpConfig = {                                                                                                  // 447
    hostname: "localhost",                                                                                            // 448
    port: "25",                                                                                                       // 449
    auth: undefined,                                                                                                  // 450
    returnAddress: returnAddress.value                                                                                // 451
  };                                                                                                                  //
                                                                                                                      //
  var parsed = undefined;                                                                                             // 454
  try {                                                                                                               // 455
    parsed = smtpUrl && Url.parse(smtpUrl);                                                                           // 456
  } catch (e) {}                                                                                                      //
                                                                                                                      //
  if (parsed) {                                                                                                       // 459
    // If there was a SMTP URL previously defined, import its data.                                                   //
    var auth = undefined;                                                                                             // 461
    if (parsed.auth) {                                                                                                // 462
      var colonIndex = parsed.auth.indexOf(":");                                                                      // 463
      var user = undefined;                                                                                           // 464
      var pass = undefined;                                                                                           // 465
      if (colonIndex !== -1) {                                                                                        // 466
        user = parsed.auth.slice(0, colonIndex);                                                                      // 467
        pass = parsed.auth.slice(colonIndex + 1);                                                                     // 468
      }                                                                                                               //
                                                                                                                      //
      auth = {                                                                                                        // 471
        user: user,                                                                                                   // 472
        pass: pass                                                                                                    // 473
      };                                                                                                              //
    }                                                                                                                 //
                                                                                                                      //
    // Override defaults with previous config's values.                                                               //
    smtpConfig.hostname = parsed.hostname || "localhost";                                                             // 478
    smtpConfig.port = parsed.port || "25";                                                                            // 479
    smtpConfig.auth = auth;                                                                                           // 480
  }                                                                                                                   //
                                                                                                                      //
  Settings.upsert({ _id: "smtpConfig" }, { value: smtpConfig });                                                      // 483
  Settings.remove({ _id: "returnAddress" });                                                                          // 484
  Settings.remove({ _id: "smtpUrl" });                                                                                // 485
}                                                                                                                     //
                                                                                                                      //
function smtpPortShouldBeNumber() {                                                                                   // 488
  var entry = Settings.findOne({ _id: "smtpConfig" });                                                                // 489
  if (entry) {                                                                                                        // 490
    var setting = entry.value;                                                                                        // 491
    if (setting.port) {                                                                                               // 492
      setting.port = _.isNumber(setting.port) ? setting.port : parseInt(setting.port);                                // 493
      Settings.upsert({ _id: "smtpConfig" }, { value: setting });                                                     // 494
    }                                                                                                                 //
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function consolidateOrgSettings() {                                                                                   // 499
  var orgGoogleDomain = Settings.findOne({ _id: "organizationGoogle" });                                              // 500
  var orgEmailDomain = Settings.findOne({ _id: "organizationEmail" });                                                // 501
  var orgLdap = Settings.findOne({ _id: "organizationLdap" });                                                        // 502
  var orgSaml = Settings.findOne({ _id: "organizationSaml" });                                                        // 503
                                                                                                                      //
  var orgMembership = {                                                                                               // 505
    google: {                                                                                                         // 506
      enabled: orgGoogleDomain ? !!orgGoogleDomain.value : false,                                                     // 507
      domain: orgGoogleDomain ? orgGoogleDomain.value : ""                                                            // 508
    },                                                                                                                //
    email: {                                                                                                          // 510
      enabled: orgEmailDomain ? !!orgEmailDomain.value : false,                                                       // 511
      domain: orgEmailDomain ? orgEmailDomain.value : ""                                                              // 512
    },                                                                                                                //
    ldap: {                                                                                                           // 514
      enabled: orgLdap ? orgLdap.value : false                                                                        // 515
    },                                                                                                                //
    saml: {                                                                                                           // 517
      enabled: orgSaml ? orgSaml.value : false                                                                        // 518
    }                                                                                                                 //
  };                                                                                                                  //
                                                                                                                      //
  Settings.upsert({ _id: "organizationMembership" }, { value: orgMembership });                                       // 522
  Settings.remove({ _id: "organizationGoogle" });                                                                     // 523
  Settings.remove({ _id: "organizationEmail" });                                                                      // 524
  Settings.remove({ _id: "organizationLdap" });                                                                       // 525
  Settings.remove({ _id: "organizationSaml" });                                                                       // 526
}                                                                                                                     //
                                                                                                                      //
// This must come after all the functions named within are defined.                                                   //
// Only append to this list!  Do not modify or remove list entries;                                                   //
// doing so is likely change the meaning and semantics of user databases.                                             //
var MIGRATIONS = [updateLoginStyleToRedirect, enableLegacyOAuthProvidersIfNotInSettings, denormalizeInviteInfo, mergeRoleAssignmentsIntoApiTokens, fixOasisStorageUsageStats, fetchProfilePictures, assignPlans, removeKeyrings, useLocalizedTextInUserActions, verifyAllPgpSignatures, splitUserIdsIntoAccountIdsAndIdentityIds, appUpdateSettings, moveDevAndEmailLoginDataIntoIdentities, repairEmailIdentityIds, splitAccountUsersAndIdentityUsers, populateContactsFromApiTokens, cleanUpApiTokens, initServerTitleAndReturnAddress, sendReferralNotifications, assignBonuses, splitSmtpUrl, smtpPortShouldBeNumber, consolidateOrgSettings];
                                                                                                                      //
function migrateToLatest() {                                                                                          // 558
  if (Meteor.settings.replicaNumber) {                                                                                // 559
    (function () {                                                                                                    //
      // This is a replica. Wait for the first replica to perform migrations.                                         //
                                                                                                                      //
      console.log("Waiting for migrations on replica zero...");                                                       // 562
                                                                                                                      //
      var done = new Future();                                                                                        // 564
      var change = function (doc) {                                                                                   // 565
        console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                          // 566
        if (doc.value >= MIGRATIONS.length) done["return"]();                                                         // 567
      };                                                                                                              //
                                                                                                                      //
      var observer = Migrations.find({ _id: "migrations_applied" }).observe({                                         // 570
        added: change,                                                                                                // 571
        changed: change                                                                                               // 572
      });                                                                                                             //
                                                                                                                      //
      done.wait();                                                                                                    // 575
      observer.stop();                                                                                                // 576
      console.log("Migrations have completed on replica zero.");                                                      // 577
    })();                                                                                                             //
  } else {                                                                                                            //
    var applied = Migrations.findOne({ _id: "migrations_applied" });                                                  // 580
    var start = undefined;                                                                                            // 581
    if (!applied) {                                                                                                   // 582
      // Migrations table is not yet seeded with a value.  This means it has                                          //
      // applied 0 migrations.  Persist this.                                                                         //
      Migrations.insert({ _id: "migrations_applied", value: 0 });                                                     // 585
      start = 0;                                                                                                      // 586
    } else {                                                                                                          //
      start = applied.value;                                                                                          // 588
    }                                                                                                                 //
                                                                                                                      //
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                                    // 591
                                                                                                                      //
    for (var i = start; i < MIGRATIONS.length; i++) {                                                                 // 593
      // Apply migration i, then record that migration i was successfully run.                                        //
      console.log("Applying migration " + (i + 1));                                                                   // 595
      MIGRATIONS[i]();                                                                                                // 596
      Migrations.update({ _id: "migrations_applied" }, { $set: { value: i + 1 } });                                   // 597
      console.log("Applied migration " + (i + 1));                                                                    // 598
    }                                                                                                                 //
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                              // 603
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
