(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var ECMAScript = Package.ecmascript.ECMAScript;
var check = Package.check.check;
var Match = Package.check.Match;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var SandstormPowerboxRequest;

(function(){

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                             //
// packages/sandstorm-ui-powerbox/powerbox-server.js                                                           //
//                                                                                                             //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                               //
// Sandstorm - Personal Cloud Sandbox                                                                          //
// Copyright (c) 2016 Sandstorm Development Group, Inc. and contributors                                       //
// All rights reserved.                                                                                        //
//                                                                                                             //
// Licensed under the Apache License, Version 2.0 (the "License");                                             //
// you may not use this file except in compliance with the License.                                            //
// You may obtain a copy of the License at                                                                     //
//                                                                                                             //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                //
//                                                                                                             //
// Unless required by applicable law or agreed to in writing, software                                         //
// distributed under the License is distributed on an "AS IS" BASIS,                                           //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                    //
// See the License for the specific language governing permissions and                                         //
// limitations under the License.                                                                              //
                                                                                                               //
var Capnp = Npm.require("capnp");                                                                              // 17
var Grain = Capnp.importSystem("sandstorm/grain.capnp");                                                       // 18
var Ip = Capnp.importSystem("sandstorm/ip.capnp");                                                             // 19
var Email = Capnp.importSystem("sandstorm/email.capnp");                                                       // 20
                                                                                                               //
function encodePowerboxDescriptor(desc) {                                                                      // 22
  return Capnp.serializePacked(Grain.PowerboxDescriptor, desc).toString("base64").replace(/\+/g, "-").replace(/\//g, "_");
}                                                                                                              //
                                                                                                               //
Meteor.methods({                                                                                               // 29
  newFrontendRef: function (sessionId, frontendRefVariety, saveLabel) {                                        // 30
    // Checks if the requester is an admin, and if so, provides a new frontendref of the desired               //
    // variety, provided by the requesting user, owned by the grain for this session.                          //
    check(sessionId, String);                                                                                  // 33
    check(frontendRefVariety, Match.OneOf({ ipNetwork: true }, { ipInterface: true }, { emailVerifier: { services: Match.Optional([String]) } }, { verifiedEmail: { verifierId: Match.Optional(String), address: String } }));
    check(saveLabel, {                                                                                         // 40
      defaultText: String,                                                                                     // 41
      localizations: Match.Optional([{ locale: String, text: String }])                                        // 42
    });                                                                                                        //
                                                                                                               //
    var db = this.connection.sandstormDb;                                                                      // 45
                                                                                                               //
    var session = db.collections.sessions.findOne(sessionId);                                                  // 47
    if (!session) {                                                                                            // 48
      throw new Meteor.Error(403, "Invalid session ID");                                                       // 49
    }                                                                                                          //
                                                                                                               //
    var descriptor = undefined;                                                                                // 52
    var requirements = [];                                                                                     // 53
    if (frontendRefVariety.ipNetwork) {                                                                        // 54
      if (!db.isAdmin(this.userId)) {                                                                          // 55
        throw new Meteor.Error(403, "User must be an admin to powerbox offer IpNetwork");                      // 56
      }                                                                                                        //
                                                                                                               //
      descriptor = encodePowerboxDescriptor({ tags: [{ id: Ip.IpNetwork.typeId }] });                          // 59
      requirements.push({ userIsAdmin: Meteor.userId() });                                                     // 60
    } else if (frontendRefVariety.ipInterface) {                                                               //
      if (!db.isAdmin(this.userId)) {                                                                          // 62
        throw new Meteor.Error(403, "User must be an admin to powerbox offer IpInterface");                    // 63
      }                                                                                                        //
                                                                                                               //
      descriptor = encodePowerboxDescriptor({ tags: [{ id: Ip.IpInterface.typeId }] });                        // 66
      requirements.push({ userIsAdmin: Meteor.userId() });                                                     // 67
    } else if (frontendRefVariety.emailVerifier) {                                                             //
      var services = frontendRefVariety.emailVerifier.services;                                                // 69
      if (services) {                                                                                          // 70
        services.forEach(function (service) {                                                                  // 71
          if (!Accounts.identityServices[service]) {                                                           // 72
            throw new Error("No such identity service: " + service);                                           // 73
          }                                                                                                    //
        });                                                                                                    //
      }                                                                                                        //
                                                                                                               //
      descriptor = encodePowerboxDescriptor({ tags: [{ id: Email.EmailVerifier.typeId }] });                   // 78
    } else if (frontendRefVariety.verifiedEmail) {                                                             //
      // Verify that the address actually belongs to the user.                                                 //
                                                                                                               //
      if (!_.contains(getVerifiedEmails(db, this.userId, frontendRefVariety.verifiedEmail.verifierId), frontendRefVariety.verifiedEmail.address)) {
        throw new Meteor.Error(403, "User has no such verified address");                                      // 85
      }                                                                                                        //
                                                                                                               //
      // Add the session's tabId.                                                                              //
      frontendRefVariety.verifiedEmail.tabId = session.tabId;                                                  // 89
                                                                                                               //
      // Build the descriptor, which contains the verifier ID.                                                 //
      var tagValue = frontendRefVariety.verifiedEmail.verifierId && Capnp.serialize(Email.VerifiedEmail.PowerboxTag, { verifierId: new Buffer(frontendRefVariety.verifiedEmail.verifierId, "hex") });
      descriptor = encodePowerboxDescriptor({                                                                  // 95
        tags: [{ id: Email.VerifiedEmail.typeId, value: tagValue }]                                            // 96
      });                                                                                                      //
    } else {                                                                                                   //
      throw new Meteor.Error(500, "Unimplemented frontendRef type");                                           // 99
    }                                                                                                          //
                                                                                                               //
    var grainId = session.grainId;                                                                             // 102
    var apiTokenOwner = {                                                                                      // 103
      grain: {                                                                                                 // 104
        grainId: grainId,                                                                                      // 105
        saveLabel: saveLabel,                                                                                  // 106
        introducerIdentity: session.identityId                                                                 // 107
      }                                                                                                        //
    };                                                                                                         //
                                                                                                               //
    // TODO(cleanup): refactor: reaches out into core.js                                                       //
    var sturdyRef = waitPromise(saveFrontendRef(frontendRefVariety, apiTokenOwner, requirements)).sturdyRef.toString();
    return { sturdyRef: sturdyRef, descriptor: descriptor };                                                   // 113
  }                                                                                                            //
});                                                                                                            //
                                                                                                               //
var PowerboxOption = (function () {                                                                            //
  function PowerboxOption(fields) {                                                                            // 118
    babelHelpers.classCallCheck(this, PowerboxOption);                                                         //
                                                                                                               //
    _.extend(this, fields);                                                                                    // 119
  }                                                                                                            //
                                                                                                               //
  PowerboxOption.prototype.intersect = (function () {                                                          // 117
    function intersect(other) {                                                                                // 122
      // Intersect two options with the same ID. Used when combining matches from multiple tags                //
      // in the same descriptor. The tags are a conjunction.                                                   //
      //                                                                                                       //
      // Returns true if there was any overlap, or false if there was no overlap and therefore the             //
      // option should be dropped.                                                                             //
                                                                                                               //
      if (other._id != this._id) {                                                                             // 129
        throw new Error("can only merge options with the same ID");                                            // 130
      }                                                                                                        //
                                                                                                               //
      if (other.grainId) {                                                                                     // 133
        if (this.uiView && !other.uiView) delete this.uiView;                                                  // 134
        if (this.hostedObject && !other.hostedObject) delete this.hostedObject;                                // 135
        return !!(this.uiView || this.hostedObject);                                                           // 136
      } else if ((this.frontendRef || {}).verifiedEmail) {                                                     //
        // Try to hint which verifierId matched.                                                               //
        if (this.frontendRef.verifiedEmail.verifierId) {                                                       // 139
          if (other.frontendRef.verifiedEmail.verifierId) {                                                    // 140
            // Can't match multiple verifiers at once.                                                         //
            return false;                                                                                      // 142
          }                                                                                                    //
        } else if (other.frontendRef.verifiedEmail.verifierId) {                                               //
          // One request did not specify any verifierId, the other did, and both matched, so use               //
          // the verifierId.                                                                                   //
          this.frontendRef.verifiedEmail.verifierId = other.frontendRef.verifiedEmail.verifierId;              // 147
        }                                                                                                      //
                                                                                                               //
        return true;                                                                                           // 150
      } else {                                                                                                 //
        // No intersection logic needed for other types.                                                       //
        return true;                                                                                           // 153
      }                                                                                                        //
    }                                                                                                          //
                                                                                                               //
    return intersect;                                                                                          //
  })();                                                                                                        //
                                                                                                               //
  PowerboxOption.prototype.union = (function () {                                                              // 117
    function union(other) {                                                                                    // 157
      // Union two options with the same ID. Used when combining matches from multiple descriptors.            //
      // The descriptors are a disjunction.                                                                    //
                                                                                                               //
      if (other._id != this._id) {                                                                             // 161
        throw new Error("can only merge options with the same ID");                                            // 162
      }                                                                                                        //
                                                                                                               //
      if (other.grainId) {                                                                                     // 165
        if (!this.uiView && other.uiView) this.uiView = other.uiView;                                          // 166
        if (!this.hostedObject && other.hostedObject) this.hostedObject = other.hostedObject;                  // 167
      } else if ((this.frontendRef || {}).verifiedEmail) {                                                     //
        // If this doesn't have a verifierId but other does, copy it over.                                     //
        if (!this.frontendRef.verifiedEmail.verifierId && other.frontendRef.verifiedEmail.verifierId) {        // 170
          this.frontendRef.verifiedEmail.verifierId = other.frontendRef.verifiedEmail.verifierId;              // 172
        }                                                                                                      //
      }                                                                                                        //
    }                                                                                                          //
                                                                                                               //
    return union;                                                                                              //
  })();                                                                                                        //
                                                                                                               //
  PowerboxOption.prototype.subtract = (function () {                                                           // 117
    function subtract(other) {                                                                                 // 177
      // Remove `other` from this. Used when `other` is a match deemed unacceptable. Returns true              //
      // if there's anything left, false otherwise.                                                            //
                                                                                                               //
      if (other._id != this._id) {                                                                             // 181
        throw new Error("can only merge options with the same ID");                                            // 182
      }                                                                                                        //
                                                                                                               //
      if (other.grainId) {                                                                                     // 185
        if (this.uiView && other.uiView) delete this.uiView;                                                   // 186
        if (this.hostedObject && other.hostedObject) delete this.hostedObject;                                 // 187
        return !!(this.uiView || this.hostedObject);                                                           // 188
      } else {                                                                                                 //
        return false;                                                                                          // 190
      }                                                                                                        //
    }                                                                                                          //
                                                                                                               //
    return subtract;                                                                                           //
  })();                                                                                                        //
                                                                                                               //
  return PowerboxOption;                                                                                       //
})();                                                                                                          //
                                                                                                               //
var specialCaseTypes = {};                                                                                     // 195
// This object maps tag IDs to functions which return lists of matches for that tag.                           //
                                                                                                               //
specialCaseTypes[Grain.UiView.typeId] = function (db, userId, value) {                                         // 198
  if (!userId) return [];                                                                                      // 199
                                                                                                               //
  // TODO(someday): Allow `value` to specify app IDs to filter for.                                            //
                                                                                                               //
  var sharedGrainIds = db.userApiTokens(userId).map(function (token) {                                         // 203
    return token.grainId;                                                                                      //
  });                                                                                                          //
  var ownedGrainIds = Grains.find({ userId: userId }, { fields: { _id: 1 } }).map(function (grain) {           // 204
    return grain._id;                                                                                          //
  });                                                                                                          //
                                                                                                               //
  return _.uniq(sharedGrainIds.concat(ownedGrainIds)).map(function (grainId) {                                 // 206
    return new PowerboxOption({                                                                                // 207
      _id: "grain-" + grainId,                                                                                 // 208
      grainId: grainId,                                                                                        // 209
      uiView: {}                                                                                               // 210
    });                                                                                                        //
  });                                                                                                          //
};                                                                                                             //
                                                                                                               //
specialCaseTypes[Ip.IpNetwork.typeId] = function (db, userId, value) {                                         // 215
  if (Meteor.users.findOne(userId).isAdmin) {                                                                  // 216
    return [new PowerboxOption({                                                                               // 217
      _id: "frontendref-ipnetwork",                                                                            // 219
      frontendRef: { ipNetwork: true }                                                                         // 220
    })];                                                                                                       //
  } else {                                                                                                     //
    return [];                                                                                                 // 224
  }                                                                                                            //
};                                                                                                             //
                                                                                                               //
specialCaseTypes[Ip.IpInterface.typeId] = function (db, userId, value) {                                       // 228
  if (Meteor.users.findOne(userId).isAdmin) {                                                                  // 229
    return [new PowerboxOption({                                                                               // 230
      _id: "frontendref-ipinterface",                                                                          // 232
      frontendRef: { ipInterface: true }                                                                       // 233
    })];                                                                                                       //
  } else {                                                                                                     //
    return [];                                                                                                 // 237
  }                                                                                                            //
};                                                                                                             //
                                                                                                               //
specialCaseTypes[Email.EmailVerifier.typeId] = function (db, userId, value) {                                  // 241
  var results = [];                                                                                            // 242
                                                                                                               //
  results.push({                                                                                               // 244
    _id: "emailverifier-all",                                                                                  // 245
    frontendRef: { emailVerifier: {} }                                                                         // 246
  });                                                                                                          //
                                                                                                               //
  for (var _name in babelHelpers.sanitizeForInObject(Accounts.identityServices)) {                             // 249
    if (Accounts.identityServices[_name].isEnabled()) {                                                        // 250
      results.push({                                                                                           // 251
        _id: "emailverifier-" + _name,                                                                         // 252
        frontendRef: { emailVerifier: { services: [_name] } }                                                  // 253
      });                                                                                                      //
    }                                                                                                          //
  };                                                                                                           //
                                                                                                               //
  return results;                                                                                              // 258
};                                                                                                             //
                                                                                                               //
specialCaseTypes[Email.VerifiedEmail.typeId] = function (db, userId, value) {                                  // 261
  var verifierId = value && Capnp.parse(Email.VerifiedEmail.PowerboxTag, value).verifierId.toString("base64");
  return getVerifiedEmails(db, userId, verifierId).map(function (address) {                                    // 264
    return {                                                                                                   //
      _id: "email-" + address,                                                                                 // 265
      frontendRef: { verifiedEmail: { verifierId: verifierId, address: address } }                             // 266
    };                                                                                                         //
  });                                                                                                          //
};                                                                                                             //
                                                                                                               //
function getVerifiedEmails(db, userId, verifierId) {                                                           // 270
  // Get all of the email addresses verified as belonging to the given user using the given                    //
  // verifier.                                                                                                 //
                                                                                                               //
  var services = null;                                                                                         // 274
                                                                                                               //
  if (verifierId) {                                                                                            // 276
    var verifier = db.collections.apiTokens.findOne(verifierId);                                               // 277
    var verifierInfo = ((verifier || {}).frontendRef || {}).emailVerifier;                                     // 278
    if (!verifierInfo) return []; // invalid verifier                                                          // 279
                                                                                                               //
    if (verifierInfo.services) {                                                                               // 281
      // Limit to the listed services.                                                                         //
      services = {};                                                                                           // 283
      verifierInfo.services.forEach(function (service) {                                                       // 284
        return services[service] = true;                                                                       //
      });                                                                                                      //
    }                                                                                                          //
  }                                                                                                            //
                                                                                                               //
  var user = Meteor.users.findOne(userId);                                                                     // 288
  var emails = {}; // map address -> true, for uniquification                                                  // 289
  Meteor.users.find({ _id: { $in: SandstormDb.getUserIdentityIds(user) } }).forEach(function (identity) {      // 290
    if (!services || services[identity.profile.service]) {                                                     // 291
      SandstormDb.getVerifiedEmails(identity).forEach(function (email) {                                       // 292
        emails[email.email] = true;                                                                            // 292
      });                                                                                                      //
    }                                                                                                          //
  });                                                                                                          //
                                                                                                               //
  return Object.keys(emails);                                                                                  // 296
}                                                                                                              //
                                                                                                               //
Meteor.publish("powerboxOptions", function (requestId, descriptorList) {                                       // 299
  var _this = this;                                                                                            //
                                                                                                               //
  // Performs a powerbox query, returning options to present to the user in the powerbox.                      //
  // `descriptorList` is an array of `PowerboxDescriptor`s, each individually serialized in                    //
  // packed format and base64-encoded. The publish populates a pseudo-collection called                        //
  // `powerboxOptions`. Each item has the following fields:                                                    //
  //                                                                                                           //
  //   _id: Object with two fields: `requestId` (as passed when subscribing) and `optionId` (a                 //
  //       unique identifier for the option).                                                                  //
  //   matchQuality: "preferred" or "acceptable" ("unacceptable" options aren't returned).                     //
  //   frontendRef: If present, selecting this option means creating a simple frontendRef. The                 //
  //       field value should be passed back to the method `newFrontendRef` verbatim. The format               //
  //       of the field is the same as ApiTokens.frontendRef.                                                  //
  //   grainId: If present, this option selects a grain to satisfy the request. One or both of                 //
  //       `uiView` and `hostedObject` will be present. Mutually exclusive with `frontendRef`.                 //
  //   uiView: If present, this option creates a UiView for a grain. The UI should allow the user              //
  //       to choose a role to grant. Only present when `grainId` is also present. `uiView` is an              //
  //       object with the fields:                                                                             //
  //     TODO(someday): Some sort of indication of what options the powerbox should give the user              //
  //         as far as which permissions to grant? Unclear if any app would ever want to specify               //
  //         permissions when requesting a plain UiView.                                                       //
  //   hostedObject: If present, this grain advertises that it publishes capabilities that might               //
  //       match the query. If selected, a request session to the grain embedded should be shown               //
  //       embedded in the powerbox and the same request should be passed to it. Only present when             //
  //       `grainId` is also present. `hostedObject`, when present, is (for now) an empty object.              //
                                                                                                               //
  var results = {};                                                                                            // 324
  var db = this.connection.sandstormDb;                                                                        // 325
                                                                                                               //
  if (descriptorList.length > 0) {                                                                             // 327
    var descriptorMatches = descriptorList.map(function (packedDescriptor) {                                   // 328
      // Decode the descriptor.                                                                                //
      // TODO(now): Also single-segment? Canonical?                                                            //
                                                                                                               //
      // Note: Node's base64 decoder also accepts URL-safe base64, so no need to translate.                    //
      var descriptor = Capnp.parsePacked(Grain.PowerboxDescriptor, new Buffer(packedDescriptor, "base64"));    // 333
                                                                                                               //
      if (!descriptor.tags || descriptor.tags.length === 0) return {};                                         // 336
                                                                                                               //
      // Expand each tag into a match map.                                                                     //
      var tagMatches = descriptor.tags.map(function (tag) {                                                    // 339
        var type = specialCaseTypes[tag.id];                                                                   // 340
        var result = {};                                                                                       // 341
                                                                                                               //
        if (type) {                                                                                            // 343
          type(db, _this.userId, tag.value).forEach(function (option) {                                        // 344
            result[option._id] = option;                                                                       // 345
          });                                                                                                  //
        }                                                                                                      //
                                                                                                               //
        return result;                                                                                         // 349
      });                                                                                                      //
                                                                                                               //
      // Intersect two tags' matches.                                                                          //
      var matches = tagMatches.reduce(function (a, b) {                                                        // 353
        for (var id in babelHelpers.sanitizeForInObject(a)) {                                                  // 354
          if (id in b) {                                                                                       // 355
            if (!a[id].intersect(b[id])) {                                                                     // 356
              // Empty intersection.                                                                           //
              delete a[id];                                                                                    // 358
            }                                                                                                  //
          } else {                                                                                             //
            // This match only exists in a, not b, so delete it.                                               //
            delete a[id];                                                                                      // 362
          }                                                                                                    //
        }                                                                                                      //
                                                                                                               //
        return a;                                                                                              // 366
      });                                                                                                      //
                                                                                                               //
      return { descriptor: descriptor, matches: matches };                                                     // 369
    });                                                                                                        //
                                                                                                               //
    // TODO(someday): The implementation of matchQuality here is not quite right. In theory, we're             //
    //   supposed to compare descriptors to determine which ones are more specific than which                  //
    //   others, and prefer the most-specific match. For now, though, I've implemented a heuristic:            //
    //   consider each descriptor in order. If it is "unacceptable", have it cancel out any                    //
    //   matches seen previously. Otherwise, take the maximum match quality. This will usually                 //
    //   produce the same results as long as "unacceptable" descriptors are placed last in the                 //
    //   list.                                                                                                 //
                                                                                                               //
    var matches = descriptorMatches.reduce(function (finalMatches, clause) {                                   // 380
      if (clause.matchQuality === "unacceptable") {                                                            // 381
        // Remove b's matches from a.                                                                          //
        for (var id in babelHelpers.sanitizeForInObject(clause.matches)) {                                     // 383
          if (id in finalMatches) {                                                                            // 384
            if (!finalMatches[id].subtract(clause.matches[id])) {                                              // 385
              delete finalMatches[id];                                                                         // 386
            }                                                                                                  //
          }                                                                                                    //
        }                                                                                                      //
                                                                                                               //
        return finalMatches;                                                                                   // 391
      } else {                                                                                                 //
        for (var id in babelHelpers.sanitizeForInObject(clause.matches)) {                                     // 393
          if (id in finalMatches) {                                                                            // 394
            finalMatches[id].union(clause.matches[id]);                                                        // 395
          } else {                                                                                             //
            finalMatches[id] = clause.matches[id];                                                             // 397
          }                                                                                                    //
                                                                                                               //
          if (clause.matchQuality === "preferred") {                                                           // 400
            finalMatches[id].matchQuality = "preferred";                                                       // 401
          }                                                                                                    //
        }                                                                                                      //
                                                                                                               //
        return finalMatches;                                                                                   // 405
      }                                                                                                        //
    }, {});                                                                                                    //
                                                                                                               //
    for (var id in babelHelpers.sanitizeForInObject(matches)) {                                                // 409
      if (!matches[id].matchQuality) {                                                                         // 410
        matches[id].matchQuality = "acceptable";                                                               // 411
      }                                                                                                        //
                                                                                                               //
      matches[id].requestId = requestId;                                                                       // 414
      this.added("powerboxOptions", id, matches[id]);                                                          // 415
    }                                                                                                          //
  }                                                                                                            //
                                                                                                               //
  // TODO(someday): Make reactive? Seems annoying.                                                             //
                                                                                                               //
  this.ready();                                                                                                // 421
});                                                                                                            //
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-ui-powerbox'] = {
  SandstormPowerboxRequest: SandstormPowerboxRequest
};

})();

//# sourceMappingURL=sandstorm-ui-powerbox.js.map
